Package: filearray
Type: Package
Title: File-Backed Array for Out-of-Memory Computation
Version: 0.2.2
Language: en-US
Encoding: UTF-8
License: LGPL-3
URL: https://dipterix.org/filearray/,
        https://github.com/dipterix/filearray
BugReports: https://github.com/dipterix/filearray/issues
Authors@R: c(
    person(
        "Zhengjia", "Wang", role = c("aut", "cre", "cph"),
        email = "dipterix.wang@gmail.com"
    ))
Description: Stores large arrays in files to avoid occupying large
    memories. Implemented with super fast gigabyte-level multi-threaded 
    reading/writing via 'OpenMP'. Supports multiple non-character data 
    types (double, float, complex, integer, logical, and raw).
Imports: digest, fastmap (>= 1.1.1), methods, Rcpp, uuid (>= 1.1.0)
Suggests: bit64, knitr, rmarkdown, testthat (>= 3.0.0)
LinkingTo: BH, Rcpp
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-22 19:18:10 UTC; dipterix
Author: Zhengjia Wang [aut, cre, cph]
Maintainer: Zhengjia Wang <dipterix.wang@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-23 05:10:37 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-23 09:23:21 UTC; unix
Archs: filearray.so.dSYM

Package: filearray
Type: Package
Title: File-Backed Array for Out-of-Memory Computation
Version: 0.2.3
Language: en-US
Encoding: UTF-8
License: LGPL-3
URL: https://dipterix.org/filearray/,
        https://github.com/dipterix/filearray
BugReports: https://github.com/dipterix/filearray/issues
Authors@R: c(
    person(
        "Zhengjia", "Wang", role = c("aut", "cre", "cph"),
        email = "dipterix.wang@gmail.com", 
        comment = c(ORCID = "0000-0001-5629-1116")
    ))
Description: Stores large arrays in files to avoid occupying large
    memories. Implemented with super fast gigabyte-level multi-threaded 
    reading/writing via 'OpenMP'. Supports multiple non-character data 
    types (double, float, complex, integer, logical, and raw).
Imports: digest, fastmap (>= 1.1.1), methods, Rcpp, uuid (>= 1.1.0)
Suggests: bit64, knitr, rmarkdown, testthat (>= 3.0.0)
LinkingTo: BH, Rcpp
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-08-20 16:43:08 UTC; dipterix
Author: Zhengjia Wang [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-5629-1116>)
Maintainer: Zhengjia Wang <dipterix.wang@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-21 11:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",FileArray)
S3method("[<-",FileArray)
S3method("[[",FileArray)
S3method("dimnames<-",FileArray)
S3method(as.array,FileArray)
S3method(as_filearray,FileArray)
S3method(as_filearray,FileArrayProxy)
S3method(as_filearray,RAVEFileArray)
S3method(as_filearray,character)
S3method(as_filearray,default)
S3method(as_filearrayproxy,FileArray)
S3method(as_filearrayproxy,FileArrayProxy)
S3method(as_filearrayproxy,default)
S3method(dim,FileArray)
S3method(dimnames,FileArray)
S3method(fwhich,FileArray)
S3method(fwhich,default)
S3method(length,FileArray)
S3method(max,FileArray)
S3method(min,FileArray)
S3method(range,FileArray)
S3method(subset,FileArray)
S3method(sum,FileArray)
export(apply)
export(as_filearray)
export(as_filearrayproxy)
export(filearray_bind)
export(filearray_checkload)
export(filearray_create)
export(filearray_load)
export(filearray_load_or_create)
export(filearray_threads)
export(fmap)
export(fmap2)
export(fmap_element_wise)
export(fwhich)
export(mapreduce)
export(typeof)
exportClasses(FileArray)
exportClasses(FileArrayProxy)
exportMethods(apply)
exportMethods(mapreduce)
importFrom(Rcpp,sourceCpp)
importFrom(methods,new)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,setRefClass)
importFrom(methods,signature)
useDynLib(filearray, .registration = TRUE)

ensure_partition <- function(
    file, partition, dimension, 
    type = c("double", "integer", "logical", "raw"), size = NULL) {
    
    type <- match.arg(type)
    if (is.null(size)) {
        size <- get_elem_size(type)
    } else {
        size <- as.integer(size)
    }
    
    if ( !file.exists(file) ) {
        fid <- file(description = file, open = "w+b")
        write_header(fid, partition, dimension, type, size)
        close(fid)
    }
    
    header <- validate_header(file)
    
    expected_type <- sexp_to_type(header$sexp_type)
    
    if ( type != expected_type ) {
        stop(sprintf("Partition data type mismatch: %s != %s", expected_type, type))
    }
    
    if ( header$partition != partition ) {
        quiet_warning(sprintf("Partition number mismatch: %s != %s", header$partition, partition))
    }
    if ( prod(dimension) != header$partition_size ) {
        quiet_warning(sprintf("Partition size mismatch: %s != %s", header$partition_size, prod(dimension)))
    }
    
    return(header)
    
}

sexp_to_type <- function(sexp) {
    switch(
        as.character(sexp),
        "14" = "double",
        "13" = "integer",
        "10" = "logical",
        "24" = "raw",
        "15" = "complex",
        "26" = "float",
        stop("Unknown SEXP code: ", sexp)
    )
}

type_to_sexp <- function(type) {
    switch(
        type,
        double = 14L,
        integer = 13L,
        logical = 10L,
        raw = 24L,
        complex = 15L,
        float = 26L,
        stop("Unknown data type: ", type)
    )
}

load_partition <- function(file, dim) {
    stopifnot(file.exists(file))
    fid <- file(description = file, open = "rb")
    on.exit({
        close(fid)
    })
    header <- validate_header(fid = fid)
    type <- sexp_to_type(header$sexp_type)
    
    if ( missing(dim) ) {
        dim <- header$partition_dim
    } else {
        stopifnot(prod(header$partition_dim) == prod(dim))
    }
    
    structure(
        readBin(con = fid, what = type, size = header$unit_bytes, 
                n = header$content_length, endian = header$endianness),
        dim = dim
    )
}

#' @title Apply functions over file array margins (extended)
#' @param X a file array
#' @param MARGIN scalar giving the subscripts which the function will be applied over. Current implementation only allows margin size to be one
#' @param FUN the function to be applied
#' @param ... optional arguments to \code{FUN}
#' @param simplify a logical indicating whether results should be simplified if possible
#' @return See Section 'Value' in \code{\link[base]{apply}};
#' @export
setGeneric("apply")

apply_filearray <- function(X, MARGIN, FUN, ..., simplify = TRUE) {
    if (!X$valid()) {
        stop("Invalid file array")
    }
    dim <- X$dimension()
    
    FUN <- match.fun(FUN)
    simplify <- isTRUE(simplify)
    d <- dim(X)
    dl <- length(d)
    dn <- dimnames(X)
    ds <- seq_len(dl)
    if (is.character(MARGIN)) {
        dnn <- names(dn)
        if (is.null(dnn)) 
            stop("'X' must have named dimnames")
        MARGIN <- match(MARGIN, dnn)
        if (anyNA(MARGIN)) 
            stop("not all elements of 'MARGIN' are names of dimensions")
    }
    d.call <- d[-MARGIN]
    d.ans <- d[MARGIN]
    if (anyNA(d.call) || anyNA(d.ans)) {
        stop("'MARGIN' does not match dim(X)")
    }
    s.call <- ds[-MARGIN]
    s.ans <- ds[MARGIN]
    if (length(s.ans) != 1) {
        stop("`apply` on FileArray margin size can only be 1.")
    }
    dn.call <- dn[-MARGIN]
    dn.ans <- dn[MARGIN]
    d2 <- prod(d.ans)
    if (d2 == 0L) {
        newX <- array(vector(typeof(X), 1L), 
                      dim = c(prod(d.call), 1L))
        if (length(d.call) < 2L) {
            tmp <- newX[, 1]
        } else {
            tmp <- array(newX[, 1L], d.call, dn.call)
        }
        ans <- forceAndCall(1, FUN, tmp, ...)
        if (is.null(ans)) {
            return(ans)
        } else if (length(d.ans) < 2L) {
            return(ans[1L][-1L])
        } else {
            return(array(ans, d.ans, dn.ans))
        }
    }
    
    tmp <- rep("", dl)
    tmp[[s.ans]] <- ".__i__."
    f <- sprintf("function(.__i__., ...) { FUN(X[%s], ...) }", paste(tmp, collapse = ","))
    f <- eval(parse(text = f))
    
    sapply(seq_len(d[[s.ans]]), f, ..., simplify = simplify)
}

#' @rdname apply
#' @export
setMethod("apply", signature(X = "FileArray"), apply_filearray)

#' @rdname apply
#' @export
setMethod("apply", signature(X = "FileArrayProxy"), apply_filearray)

# 
# fa_subtract <- function(e1, e2) {
#     call <- match.call()
#     call[[1]] <- quote(`-`)
#     label <- sprintf("Calculating: %s (fa_subtract)", deparse1(call))
#     
#     fa_pairwise_operator(e1, e2, op = "-", label = label)
# }
# 
# 
# 
# setMethod('-', signature(e1 = "FileArray", e2 = "FileArray"), fa_subtract)
# 
# 
# setMethod('-', signature(e1 = "FileArray", e2 = "numeric"), fa_subtract)
# 
# setMethod('-', signature(e1 = "numeric", e2 = "FileArray"), fa_subtract)
# 
# 
# setMethod('-', signature(e1 = "FileArray", e2 = "complex"), fa_subtract)
# 
# setMethod('-', signature(e1 = "complex", e2 = "FileArray"), fa_subtract)
# 
# 
# setMethod('-', signature(e1 = "FileArray", e2 = "logical"), fa_subtract)
# 
# setMethod('-', signature(e1 = "logical", e2 = "FileArray"), fa_subtract)
# 
# 
# setMethod('-', signature(e1 = "FileArray", e2 = "array"), fa_subtract)
# 
# setMethod('-', signature(e1 = "array", e2 = "FileArray"), fa_subtract)

library(testthat)
library(filearray)

# Sys.setenv("FILEARRAY_SKIP_COLLAPSE" = "TRUE")

cat(utils::capture.output({
    print(Sys.getenv())
}), sep = "\n", file = stderr())

test_check("filearray")

filearray:::clear_cache()

#' @title A map-reduce method to iterate blocks of file-array data with little memory usage
#' @param x a file array object
#' @param map mapping function that receives 3 arguments; see 'Details'
#' @param reduce \code{NULL}, or a function that takes a list as input
#' @param buffer_size control how we split the array; see 'Details'
#' @param ... passed to other methods
#' @return If \code{reduce} is \code{NULL}, return mapped results, otherwise
#' return reduced results from \code{reduce} function
#' 
#' @details When handling out-of-memory arrays, it is recommended to load
#' a block of array at a time and execute on block level. See 
#' \code{\link{apply}} for a implementation. When an array is too large,
#' and when there are too many blocks, this operation will become 
#' very slow if computer memory is low. 
#' This is because the R will perform garbage collection frequently. 
#' Implemented in \code{C++}, \code{mapreduce} creates a buffer to store
#' the block data. By reusing the memory over and over again, it is possible
#' to iterate through the array with minimal garbage collections. Many 
#' statistics, including \code{min}, \code{max}, \code{sum}, 
#' \code{mean}, ... These statistics can be calculated in this 
#' way efficiently.
#' 
#' The function \code{map} contains three arguments: \code{data} (mandate), 
#' \code{size} (optional), and \code{first_index} (optional). 
#' The \code{data} is the buffer,
#' whose length is consistent across iterations. \code{size} indicates
#' the effective size of the buffer. If the partition size
#' is not divisible by the buffer size, only first \code{size} elements of
#' the data are from array, and the rest elements will be \code{NA}. 
#' This situation could only occurs when \code{buffer_size} is manually 
#' specified. By default, all of \code{data} should belong to arrays.
#' The last argument \code{first_index} is the index of the first element
#' \code{data[1]} in the whole array. It is useful when positional data 
#' is needed. 
#' 
#' The buffer size, specified by \code{buffer_size} is an 
#' additional optional argument in \code{...}. Its default is \code{NA},
#' and will be calculated automatically. If manually specified, a
#' large buffer size would be desired to speed up the calculation.
#' The default buffer size will not exceed \eqn{nThreads x 2MB}, where 
#' \code{nThreads} is the number of threads set by \code{\link{filearray_threads}}.
#' When partition length cannot be divided by the buffer size, instead of
#' trimming the buffer, \code{NA}s will be filled to the buffer, 
#' passed to \code{map} function; see previous paragraph for treatments.
#' 
#' The function \code{mapreduce} ignores the missing partitions. That means
#' if a partition is missing, its data will not be read nor passed to 
#' \code{map} function. Please run \code{x$initialize_partition()} to make sure
#' partition files exist.
#' 
#' @examples 
#' 
#' 
#' x <- filearray_create(tempfile(), c(100, 100, 10))
#' x[] <- rnorm(1e5)
#' 
#' ## calculate summation
#' # identical to sum(x[]), but is more feasible in large cases
#' 
#' mapreduce(x, map = function(data, size) {
#'     # make sure `data` is all from array
#'     if (length(data) != size) {
#'         data <- data[1:size]
#'     }
#'     sum(data)
#' }, reduce = function(mapped_list) {
#'     do.call(sum, mapped_list)
#' })
#' 
#' 
#' ## Find elements are less than -3
#' positions <- mapreduce(
#'     x,
#'     map = function(data, size, first_index) {
#'         if (length(data) != size) {
#'             data <- data[1:size]
#'         }
#'         which(data < -3) + (first_index - 1)
#'     },
#'     reduce = function(mapped_list) {
#'         do.call(c, mapped_list)
#'     }
#' )
#' 
#' if(length(positions)) {
#'     x[[positions[1]]]
#' }
#' 
#' 
#' @export
setGeneric("mapreduce", function(x, map, reduce, ...) {
    standardGeneric("mapreduce")
})

buffer_mapreduce <- function(x, map, reduce = NULL, buffer_size = NA) {
    # TODO: edit for proxy arrays
    if (!x$valid()) {
        stop("Invalid file array")
    }
    if ( is_fileproxy(x) ) {
        x <- fa_eval_ops(x)
    }
    
    current_bsz <- get_buffer_size()
    on.exit({
        set_buffer_size(current_bsz)
    })
    set_buffer_size(max_buffer_size())
    
    argnames <- names(formals(map))
    if (length(argnames) < 3 && !"..." %in% argnames) {
        if ( length(argnames) == 1 ) {
            map_ <- function(data, size, idx) {
                map(data)
            }
        } else {
            map_ <- function(data, size, idx) {
                map(data, size)
            }
        }
    } else {
        map_ <- function(data, size, idx) {
            map(data, size, idx)
        }
    }
    
    if (is.function(reduce) && !length(formals(reduce))) {
        stop("Reduce function must contain at least one argument")
    }
    
    dim <- x$dimension()
    cum_partlen <- x$.partition_info[, 3]
    sexp_type <- x$sexp_type()
    
    if (is.na(buffer_size)) {
        elem_size <- get_elem_size(typeof(x))
        mbsz <- max_buffer_size() * getThreads(FALSE) / elem_size
        
        sel <- cumprod(dim) <= mbsz
        if (any(sel)) {
            buffer_size <- prod(dim[sel])
        } else {
            buffer_size <- dim[[1]]
        }
    }
    
    if (buffer_size < 1) {
        buffer_size <- 1
    }
    
    # filebase <- paste0(x$.filebase, x$.sep)
    # FARR_buffer_mapreduce(filebase, map_, reduce, dim, 
    #                       cum_partlen, buffer_size, sexp_type)
    FARR_buffer_mapreduce(
        x$.filebase,
        map_, reduce,
        buffer_size
    )
}

#' @rdname mapreduce
#' @export
setMethod(
    mapreduce, signature(x = "FileArray", reduce = "function"),
    function(x, map, reduce, buffer_size = NA, ...) {
        buffer_mapreduce(x, map, reduce, buffer_size)
    }
)

#' @rdname mapreduce
#' @export
setMethod(
    mapreduce, signature(x = "FileArray", reduce = "NULL"),
    function(x, map, reduce, buffer_size = NA, ...) {
        buffer_mapreduce(x, map, NULL, buffer_size, ...)
    }
)

#' @rdname mapreduce
#' @export
setMethod(
    mapreduce, signature(x = "FileArray", reduce = "missing"),
    function(x, map, reduce, buffer_size = NA, ...) {
        buffer_mapreduce(x, map, NULL, buffer_size, ...)
    }
)

## S3

#' @describeIn S3-filearray get element by index
#' @export
`[[.FileArray` <- function(x, i) {
    
    if (length(i) > 1) {
        stop("attempt to select more than one element")
    } else if (length(i) < 1) {
        stop("attempt to select less than one element")
    }
    
    dim <- x$dimension()
    
    locs <- arrayInd(i, dim)
    args <- c(
        list(
            x = quote(x),
            drop = TRUE
        ),
        lapply(seq_len(ncol(locs)), function(ii) {
            locs[, ii, drop = TRUE]
        })
    )
    
    do.call("[", args)
}

#' @describeIn S3-filearray converts file array to native array in R
#' @export
as.array.FileArray <- function(x, reshape = NULL, drop = FALSE, ...) {
    x[reshape = reshape, drop = drop]
}

#' @describeIn S3-filearray get dimensions
#' @export
dim.FileArray <- function(x) {
    x$dimension()
}

#' @describeIn S3-filearray get dimension names
#' @export
dimnames.FileArray <- function(x) {
    x$dimnames()
}

#' @describeIn S3-filearray set dimension names
#' @export
`dimnames<-.FileArray` <- function(x, value) {
    x$dimnames(value)
    invisible(x)
}


#' @describeIn S3-filearray get array length
#' @export
length.FileArray <- function(x) {
    prod(x$dimension())
}

#' @describeIn S3-filearray get max value
#' @export
max.FileArray <- function(x, na.rm = FALSE, ...) {
    reduce <- function(x) {
        do.call("max", x)
    }
    if ( na.rm ) {
        mapreduce(
            x = x, map = .max_mapper1,
            reduce = reduce,
            ...
        )
    } else {
        x$initialize_partition()
        mapreduce(
            x = x, map = .max_mapper2,
            reduce = reduce,
            ...
        )
    }
    
}

#' @describeIn S3-filearray get min value
#' @export
min.FileArray <- function(x, na.rm = FALSE, ...) {
    reduce <- function(x) {
        do.call("min", x)
    }
    
    if ( na.rm ) {
        mapreduce(
            x = x, map = .min_mapper1,
            reduce = reduce, ...
        )
    } else {
        x$initialize_partition()
        mapreduce(
            x = x, map = .min_mapper2,
            reduce = reduce, ...
        )
    }
    
}

#' @describeIn S3-filearray get value range
#' @export
range.FileArray <- function(x, na.rm = FALSE, ...) {
    red_f <- function(x) {
        x <- do.call("cbind", x)
        c(min(x[1, ]), max(x[2, ]))
    }
    
    if ( na.rm ) {
        mapreduce( x = x, map = .range_mapper1,
                   reduce = red_f, ...)
    } else {
        x$initialize_partition()
        mapreduce( x = x, map = .range_mapper2,
                   reduce = red_f, ...)
    }
    
}

#' @describeIn S3-filearray get summation
#' @export
sum.FileArray <- function(x, na.rm = FALSE, ...) {
    red_f <- function(x) {
        do.call("sum", x)
    }
    
    if ( na.rm ) {
        mapreduce( x = x, map = .sum_mapper1,
                   reduce = red_f, ...)
    } else {
        x$initialize_partition()
        mapreduce( x = x, map = .sum_mapper2,
                   reduce = red_f, ...)
    }
}

#' @describeIn S3-filearray get subset file array with formulae
#' @export
subset.FileArray <- function(x, ..., drop = FALSE, .env = parent.frame()) {
    if (...length() == 0) {
        stop("No filters to subset")
    }
    dnames <- dimnames(x)
    nms <- names(dnames)
    filters <- list(...)
    criteria <- list()
    for (ftr in filters) {
        if (!is.call(ftr)) {
            stop("All filters must be formula object, e.g `Var ~ Var < 10`")
        }
        name <- as.character(ftr[[2]])
        if (!name %in% nms) {
            stop("Cannot find name `", name, "` in x")
        }
        expr <- ftr[[3]]
        if (length(criteria[[name]])) {
            expr <- as.call(list(`&`, criteria[[name]], expr))
        }
        criteria[[name]] <- expr
    }
    dim <- dim(x)
    locs <- lapply(seq_along(dim), function(ii) {
        if (ii > length(nms)) {
            return(seq_len(dim[[ii]]))
        }
        name <- nms[[ii]]
        if (is.null(criteria[[name]])) {
            return(seq_len(dim[[ii]]))
        }
        eval(criteria[[name]], envir = dnames, enclos = .env)
    })
    do.call(`[`, c(list(x = quote(x), drop = drop), locs))
}

#' @title A generic function of \code{which} that is \code{'FileArray'} compatible
#' @param x any R vector, matrix, array or file-array
#' @param val values to find, or a function taking one argument (a slice of 
#' data vector) and returns either logical vector with the same length as the 
#' slice or index of the slice; see 'Examples'
#' @param arr.ind logical; should array indices be 
#' returned when \code{x} is an array?
#' @param ret.values whether to return the values of corresponding indices as 
#' an attributes; default is false
#' @param ... passed to \code{val} if \code{val} is a function
#' @return The indices of \code{x} elements that are listed in \code{val}.
#' @examples 
#' 
#' 
#' # ---- Default case ------------------------------------
#' x <- array(1:27 + 2, rep(3,3))
#' 
#' # find index of `x` equal to either 4 or 5
#' fwhich(x, c(4,5))
#' res <- fwhich(x, c(4,5), ret.values = TRUE)
#' res
#' attr(res, "values")
#' 
#' # ---- file-array case --------------------------------
#' arr <- filearray_create(tempfile(), dim(x))
#' arr[] <- x
#' fwhich(arr, c(4,5))
#' fwhich(arr, c(4,5), arr.ind = TRUE, ret.values = TRUE)
#' 
#' arr[2:3, 1, 1]
#' 
#' # Clean up this example
#' arr$delete()
#' 
#' # ---- `val` is a function ----------------------------
#' x <- as_filearray(c(sample(15), 15), dimension = c(4,4))
#' 
#' ret <- fwhich(x, val = which.max, 
#'               ret.values = TRUE, arr.ind = FALSE)
#' 
#' # ret is the index
#' ret == which.max(x[])
#' 
#' # attr(ret, "values") is the max value
#' max(x[]) == attr(ret, "values")
#' 
#' # customize `val`
#' fwhich(x, ret.values = TRUE, arr.ind = FALSE,
#'        val = function( slice ) {
#'            slice > 10 # or which(slice > 10)
#'        })
#' 
#' 
#' @export
fwhich <- function(x, val, arr.ind = FALSE, ret.values = FALSE, ...) {
    UseMethod("fwhich")
}

#' @rdname fwhich
#' @export
fwhich.default <- function(x, val, arr.ind = FALSE, ret.values = FALSE, ...) {
    if (is.function(val)) {
        idx <- val(x, ...)
        if (is.logical(idx)) {
            idx <- which(idx, useNames = FALSE)
        }
    } else {
        idx <- which(x %in% val, arr.ind = FALSE, useNames = FALSE)
    }
    if ( arr.ind && length(dim(x)) > 0 ) {
        re <- arrayInd(idx, .dim = dim(x), .dimnames = dimnames(x))
    } else {
        re <- idx
    }
    if ( ret.values ) {
        attr(re, "values") <- x[idx]
    }
    return(re)
}

#' @rdname fwhich
#' @export
fwhich.FileArray <- function(x, val, arr.ind = FALSE, ret.values = FALSE, ...) {
    if (is.function(val)) {
        args <- list(1, ...)
        impl <- function(y) {
            args[[1]] <- y
            do.call(val, args)
        }
    } else {
        impl <- function(y) {
            which(y %in% val, arr.ind = FALSE)
        }
    }
    mapreduce(
        x, 
        map = function(data, size, start_index) {
            idx <- impl(data[seq_len(size)])
            if (is.logical(idx)) { idx <- which(idx, arr.ind = FALSE, useNames = FALSE) }
            if (!length(idx)) { return(NULL) }
            data.frame(
                index = start_index + idx - 1,
                value = data[idx]
            )
        },
        reduce = function(mapped_list) {
            mapped_data <- do.call("rbind", mapped_list)
            if (!length(mapped_data)) {
                idx <- numeric(0L)
                val <- raw(0L)
                storage.mode(val) <- storage.mode(x$.na)
            } else {
                mapped_data <- mapped_data[impl( mapped_data$val ), , drop = FALSE]
                idx <- mapped_data$index
                val <- mapped_data$value
            }
            
            if ( arr.ind ) {
                idx <- arrayInd(idx, .dim = dim(x), .dimnames = dimnames(x))
            }
            if ( ret.values ) {
                attr(idx, "values") <- val
            }
            idx
        }
    )
}

#' @title Merge and bind homogeneous file arrays
#' @description The file arrays to be merged must be homogeneous:
#' same data type, partition size, and partition length
#' @param ...,.list file array instances
#' @param filebase where to create merged array
#' @param symlink whether to use \code{\link[base]{file.symlink}}; if true,
#' then partition files will be symbolic-linked to the original arrays,
#' otherwise the partition files will be copied over. If you want your data
#' to be portable, do not use symbolic-links. The default value is \code{FALSE}
#' @param overwrite whether to overwrite when \code{filebase} already exists;
#' default is false, which raises errors
#' @param cache_ok see 'Details', only used if \code{overwrite} is true.
#' 
#' @return A bound array in \code{'FileArray'} class.
#' 
#' @details The input arrays must share the same data type and partition size.
#' The dimension for each partition should also be the same. For example
#' an array \code{x1} has dimension \eqn{100x20x30} with partition size 
#' \code{1}, then each partition dimension is \eqn{100x20x1}, and there are 
#' \code{30} partitions. \code{x1} can bind with another array of the same 
#' partition size. This means if \code{x2} has dimension
#' \eqn{100x20x40} and each partition size is \code{1}, then \code{x1} and 
#' \code{x2} can be merged.
#' 
#' If \code{filebase} exists and \code{overwrite} is \code{FALSE}, an error will 
#' always raise. If \code{overwrite=TRUE} and \code{cache_ok=FALSE}, then
#' the existing \code{filebase} will be erased and any data stored within will
#' be lost. 
#' If both \code{overwrite} and \code{cache_ok} are \code{TRUE}, then 
#' , before erasing \code{filebase}, the function validates the existing
#' array header and compare the header signatures. If the existing header
#' signature is the same as the array to be created, then the existing array 
#' will be returned. This \code{cache_ok} could be extremely useful when
#' binding large arrays with \code{symlink=FALSE} as the cache might avoid
#' moving files around. However, \code{cache_ok} should be enabled with caution.
#' This is because only the header information will be compared, but the 
#' partition data will not be compared. If the existing array was generated from
#' an old versions of the source arrays, but the data from the source arrays
#' has been altered, then the \code{cache_ok=TRUE} is rarely proper as the cache
#' is outdated.
#' 
#' The \code{symlink} option should be used with extra caution. Creating 
#' symbolic links is definitely faster than copying partition files. However, 
#' since the partition files are simply linked to the original partition files, 
#' changing to the input arrays will also affect the merged arrays, and 
#' vice versa; see 'Examples'. Also for arrays created from symbolic links, if 
#' the original 
#' arrays are deleted, while the merged arrays will not be invalidated, 
#' the corresponding partitions will no longer be accessible. Attempts to 
#' set deleted partitions will likely result in failure. Therefore
#' \code{symlink} should be set to true when creating merged arrays are
#' temporary for read-only purpose, and when speed and disk space is in
#' consideration. For extended reading, please check \code{\link[base]{files}} 
#' for details.
#' 
#' @examples 
#' partition_size <- 1
#' type <- "double"
#' x1 <- filearray_create(
#'     tempfile(), c(2,2), type = type,
#'     partition_size = partition_size)
#' x1[] <- 1:4
#' x2 <- filearray_create(
#'     tempfile(), c(2,1), type = type,
#'     partition_size = partition_size)
#' x2[] <- 5:6
#' 
#' y1 <- filearray_bind(x1, x2, symlink = FALSE)
#' y2 <- filearray_bind(x1, x2)
#' 
#' # y1 copies partition files, and y2 simply creates links 
#' # if symlink is supported
#' 
#' y1[] - y2[]
#' 
#' # change x1
#' x1[1,1] <- NA
#' 
#' # y1 is not affected
#' y1[]
#' 
#' # y2 changes 
#' y2[]
#' 
#' 
#' @export
filearray_bind <- function(
    ..., .list = list(), filebase = tempfile(), 
    symlink = FALSE, overwrite = FALSE, cache_ok = FALSE
) {
    stopifnot(length(filebase) == 1)
    symlink <- as.logical(symlink)[[1]]
    # options("filearray.symlink_enabled" = TRUE)
    if (symlink && !getOption("filearray.symlink_enabled", symlink_enabled())) {
        symlink <- FALSE
        quiet_warning("Symbolic link is disabled. Force `symlink` to be FALSE")
    }
    arrays <- c(list(...), .list)
    if (!length(arrays)) {
        stop("`filearray_bind`: No file arrays to bind")
    }
    if (!all(sapply(arrays, is_filearray))) {
        stop("`filearray_bind`: All inputs must be file arrays")
    }
    # TODO: This is quick fix, should change this
    arrays <- lapply(arrays, fa_eval_ops)
    type <- typeof(arrays[[1]])
    part_size <- arrays[[1]]$partition_size()
    dim <- arrays[[1]]$dimension()
    dim[[length(dim)]] <- part_size
    
    last_margin <- sapply(arrays, function(x) {
        if (!x$valid()) {
            stop("`filearray_bind`: some input arrays are invalid")
        }
        if (x$type() != type) {
            stop("`filearray_bind`: All arrays must be the same type")
        }
        if (x$partition_size() != part_size) {
            stop("`filearray_bind`: All arrays must share the same `partition_size`")
        }
        xdim <- x$dimension()
        xdim[[length(xdim)]] <- part_size
        
        if (!identical(as.double(xdim), as.double(dim))) {
            stop("`filearray_bind`: The partition size mismatch: they must be the same for all arrays")
        }
        x$initialize_partition()
        nrow(x$.partition_info)
    })
    
    last_margins <- sapply(arrays, dim)
    last_margins <- last_margins[nrow(last_margins), ]
    
    if (!all(last_margins %% part_size == 0)) {
        quiet_warning("One or more arrays have last margin size that cannot be devided by `partition_size`. This will cause binding arrays to be mis-aligned. ")
    }
    
    dim[[length(dim)]] <- sum(last_margin) * part_size
    
    # Dry-run
    start <- 1
    end <- 1
    
    bind_info <- list(
        is_bound = TRUE,
        dimension = as.double(dim), 
        type = type, 
        partition_size = as.double(part_size),
        partition_header_signatures = sapply(arrays, function(y) { y$header_signature(include_path = TRUE) })
    )
    bind_signature <- digest::digest(bind_info, algo = "sha256")
    
    source_info <- list()
    for (ii in seq_along(last_margin)) {
        arr <- arrays[[ii]]
        end <- start - 1 + last_margin[[ii]]
        idx <- seq.int(start, end)
        source_info[[ii]] <- arr$partition_path(seq_len(last_margin[[ii]]))
        start <- end + 1
    }
    bind_info$source_info <- source_info
    bind_info$symlink  <- symlink
    
    if (file.exists(filebase)) {
        
        if (!overwrite) {
            stop("Array has already existed at: ", filebase)
        } 
        
        if (cache_ok) {
            check <- tryCatch({
                check <- FALSE
                if (file.exists(file.path(filebase, "meta"))) {
                    header <- load_meta(filebase)
                    if (
                        identical(header$filearray_bind_signature, bind_signature) &&
                        identical(header$filearray_bind$symlink, symlink)
                    ) {
                        check <- TRUE
                    }
                }
                check
            }, error = function(e) {
                FALSE
            })
            if (check) {
                re <- filearray_load(filebase = filebase, mode = "readonly")
                attr(re, "cached_bind") <- TRUE
                return(re)
            }
        }
        
        unlink(filebase, recursive = TRUE)
    }
    re <- filearray_create(filebase = filebase, dimension = dim, type = type, partition_size = part_size)
    
    start <- 1
    end <- 1
    
    for (ii in seq_along(last_margin)) {
        arr <- arrays[[ii]]
        end <- start - 1 + last_margin[[ii]]
        idx <- seq.int(start, end)
        if ( symlink ) {
            file.symlink(
                arr$partition_path(seq_len(last_margin[[ii]])),
                re$partition_path(idx)
            )
        } else {
            file.copy(
                arr$partition_path(seq_len(last_margin[[ii]])),
                re$partition_path(idx)
            )
        }
        start <- end + 1
    }
    
    re$.header$filearray_bind <- bind_info
    re$.header$filearray_bind_signature <- bind_signature
    re$.save_header()
    
    if (symlink) {
        re$.mode <- "readonly"
    } 
    re
    
}

fa_subsetAssign1 <- function(x, ..., value) {
    if (!x$valid()) {
        stop("Invalid file array")
    }
    if (isTRUE(x$.mode == "readonly")) {
        stop("File array is read-only")
    }
    
    buf_bytes <- get_buffer_size()
    on.exit({
        set_buffer_size(buf_bytes)
    })
    
    # parse ...
    dim <- x$dimension()
    arglen <- ...length()
    
    
    locs <- list()
    if (arglen <= 1) {
        if ( arglen == 1 ) {
            missing_args <- check_missing_dots(environment())
            if (!missing_args) {
                stop("SubsetAssign FileArray only allows x[] <- v or x[i,j,...] <- v (single index not allowed)")
            }
        }
        if (!length(value) %in% c(1, prod(dim))) {
            stop("SubsetAssign FileArray `value` length mismatch: `value` length must be either 1 or the same length of the subset.")
        }
        target_dim <- dim
        x$initialize_partition(x$.partition_info[, 1])
    } else if (arglen > 1) {
        if (arglen != length(dim)) {
            stop("SubsetAssign FileArray dimension mismatch.")
        }
        
        missing_args <- check_missing_dots(environment())
        for (ii in seq_len(arglen)) {
            if (missing_args[[ii]]) {
                locs[[ii]] <- seq_len(dim[[ii]])
            } else {
                tmp <- ...elt(ii)
                if ( !length(tmp) ) {
                    return(x)
                }
                if (any(is.na(tmp))) {
                    stop("SubsetAssign cannot contain duplicated or invalid indices.")
                }
                if (is.logical(tmp)) {
                    if (length(tmp) > dim[[ii]]) {
                        stop("(subscript) logical subscript too long")
                    }
                    tmp <- rep(tmp, ceiling(dim[[ii]] / length(tmp)))
                    tmp <- which(tmp[seq_len(dim[[ii]])])
                } else if (any(tmp > dim[[ii]]) ||
                          any(tmp <= 0) ||
                          any(duplicated(tmp))) {
                    stop("SubsetAssign cannot contain duplicated or invalid indices.")
                }
                locs[[ii]] <- tmp
            }
        }
        
        target_dim <- sapply(locs, length)
        if (!length(value) %in% c(1, prod(target_dim))) {
            stop("SubsetAssign FileArray `value` length mismatch: `value` length must be either 1 or the same length of the subset.")
        }
        
        # make sure partitions exist
        tmp <- locs[[length(locs)]]
        sapply(tmp, function(i) {
            sel <- x$.partition_info[, 3] <= i
            if (any(sel)) {
                sel <- max(x$.partition_info[sel, 1])
                if (x$.partition_info[sel, 3] < i) {
                    sel <- sel + 1
                }
            } else {
                sel <- 1
            }
            x$initialize_partition(sel)
        })
    }
    
    if (prod(target_dim) == 0) {
        return(invisible(x))
    }
    
    # decide split_dim
    buffer_sz <- buf_bytes / x$element_size()
    cprod <- cumprod(dim)
    if (length(locs) == length(dim)) {
        tmp <- sapply(locs, function(x) {
            if (!length(x) || all(is.na(x))) { return(1L) }
            rg <- range(x, na.rm = TRUE)
            return(rg[2] - rg[1] + 1L)
        })
        cprod <- cprod / dim * tmp
    }
    cprod <- cprod[-length(cprod)]
    if (all(cprod > buffer_sz)) {
        split_dim <- 1
    } else {
        split_dim <- max(which(cprod <= buffer_sz))
    }
    
    # filebase <- paste0(x$.filebase, x$.sep)
    # FARR_subset_assign(
    #     filebase,
    #     listOrEnv = locs,
    #     dim = x$dimension(),
    #     cum_part_sizes = x$.partition_info[, 3],
    #     split_dim = split_dim,
    #     type = x$sexp_type(),
    #     value_ = value
    # )
    FARR_subset_assign2(
        filebase = x$.filebase, 
        value = value,
        listOrEnv = locs,
        split_dim = split_dim,
        thread_buffer = buf_bytes
    ) 
    invisible(x)
}

# setMethod("[<-",
#     signature(x = "Filearray"),
#     function (x, i, j, ..., value) 
#     {
#         stop("need a definition for the method here")
#     }
# )

# DIPSAUS DEBUG START
# verbose <- TRUE
# z = filearray_create(temp_path(), c(2,3,4))
# z[] <- 1:24
# x <- z+1
# i <- as_filearray(z[] < 5)

# `i` is a filearray
fa_subsetAssign2 <- function(x, i, value, label = "subset-assign (lazy)") {
    stopifnot(is_filearray(x) && is_filearray(i))
    
    if (typeof(i) != "logical") {
        stop("`fa_subsetAssign2`: subset index filearray must be logical")
    }
    
    e1 <- as_filearrayproxy(x)
    e2 <- as_filearrayproxy(i)
    uuid1 <- e1$uuid()
    uuid2 <- e2$uuid()
    e1$link_proxy( e2 )
    
    # check input types 
    out_type <- typeof(e1)
    
    globals <- fastmap::fastmap()
    globals$set("starting_idx", 0L)
    value_len <- length(value)
    
    op_func <- function(value_list, ...) {
        data <- value_list[[ uuid1 ]]
        
        idx <- value_list[[ uuid2 ]]
        n_assigned <- sum(idx)
        if (n_assigned == 0) { return(data) }
        
        if ( value_len == 1L ) {
            data[ idx ] <- value
        } else {
            starting_idx <- globals$get("starting_idx")
            data[ idx ] <- value[starting_idx + seq_len(n_assigned)]
            globals$set("starting_idx", starting_idx + n_assigned)
        }
        
        return(data)
    }
    
    e1$add_operator(op_func,
                    out_type = out_type,
                    context = "array_subset_assign",
                    label = label)
    
    return( e1 )
}


#' @describeIn S3-filearray subset assign array
#' @export
`[<-.FileArray` <- function(x, i, ..., lazy = FALSE, value) {
    if (!x$valid()) {
        stop("Invalid file array")
    }
    
    if (missing(i)) {
        if (is_fileproxy(x) && length(x$.ops) > 0) {
            x <- fa_eval_ops(x)
            x$.mode <- "readwrite"
        }
        return(fa_subsetAssign1(x, , ..., value = value))
    } else {
        if (is_filearray(i) || is.array(i) || 
           ( ...length() == 0 && !is.logical(i) )) {
            re <- fa_subsetAssign2(x, i, value = value)
            if (!lazy) {
                re <- fa_eval_ops(re)
                re$.mode <- "readwrite"
            }
            return(re)
        } else {
            if (is_fileproxy(x) && length(x$.ops) > 0) {
                x <- fa_eval_ops(x)
                x$.mode <- "readwrite"
            }
            return(fa_subsetAssign1(x, i, ..., value = value))
        }
    }
}

# ---- UUID generator -------------------------------------------------------
new_uuid <- function(prefix = 0L) {
    sprintf("%04d-%s", prefix, uuid::UUIDgenerate(output = "string"))
}

# ---- Index helpers --------------------------------------------------------

# Check if two arrays share the same dimension
is_same_dim <- function(x, y) {
    dx <- dim(x)
    if (length(dx) >= 2L) {
        dy <- dim(y)
        if (length(dx) != length(dy)) { return(FALSE) }
        # valid # of margins and dimensions are consistent
        if (all(dx == dy)) { return(TRUE) }
    } else if ( length(x) == length(y) ) {
        # dx dy might not exist, check length
        return(TRUE)
    }
    return(FALSE)
}

# Guess partition size from given dimensions and element size.
# This function is used to estimate a decent partition size when creating 
# arrays
guess_partition <- function(dim, elem_size) {
    last_margin <- dim[[length(dim)]]
    unit_size <- prod(dim) / last_margin * elem_size
    
    # 1: partition size cannot go beyond 1GB
    max_ <- floor(2^30 / unit_size)
    if (max_ <= 1L) {
        return(1L)
    }
    # 2: n partitions <= 100
    if (last_margin <= 100) {
        return(1L)
    }
    # 3: at most max_ units, at least fct units
    fct <- ceiling(last_margin / max_)
    if (fct > 50) {
        return(max_)
    }
    while (fct <= 50) {
        max_ <- max_ - 1L
        fct <- ceiling(last_margin / max_)
        if (max_ <= 1L) {
            return(1L)
        }
    }
    return(max_)
}


guess_fmap_buffer_size <- function(dim, element_size = 16L) {
    buffer_len <- get_buffer_size() / element_size
    buffer_large <- get_buffer_size() * 16L # default is 8.5MB, 1024 x 1024
    
    
    dimcprod <- cumprod(dim)
    len <- dimcprod[[length(dimcprod)]]
    min_partition_size <- dimcprod[[length(dimcprod) - 1]]
    
    # tiny array
    if ( buffer_len >= len ) { return( as.integer(len) ) }
    
    # small array
    if ( buffer_large >= min_partition_size ) { return( as.integer(min_partition_size) ) }
    
    # large array
    dimcprod <- dimcprod[dimcprod <= buffer_large]
    if (!length(dimcprod)) { return( dim[[1]] * dim[[2]] ) }
    
    # mid array
    unit_size <- dimcprod[[length(dimcprod)]]
    fct <- floor(buffer_large / unit_size)
    if (fct < 1) { fct <- 1 }
    return( as.integer(unit_size * fct) )
}

common_fmap_buffer_count <- function(..., .list = NULL) {
    stopifnot(...length() + length(.list) > 0)
    buffer_large <- get_buffer_size() * 16L
    sizes <- vapply(c(list(...), .list), function(x) {
        as.double(prod(x))
    }, FUN.VALUE = 0.0)
    if (length(sizes) == 1L) { return( sizes ) }
    if (all(sizes <= buffer_large)) { return(1) }
    
    gcd <- Reduce(function(a, b) {
        if ( a <= 0 || b <= 0 ) {
            stop("Cannot find proper input buffer size when data has zero length")
        }
        rem <- a %% b
        if ( rem == 0 ) { return( b ) }
        return( Recall(b, rem) )
    }, sizes)
    max_buffer <- max(sizes / gcd)
    
    max_mult <- floor(buffer_large / max_buffer)
    if (max_mult <= 1) { return(gcd) }
    
    guess_factors <- seq_len(min(max_mult, gcd))
    guess_factors <- guess_factors[floor(gcd / guess_factors) * guess_factors == gcd]
    return(gcd / max(guess_factors))
}

validate_fmap_buffer_count <- function(count, input_lens) {
    count <- count[[1]]
    if ( is.na(count) || count != round(count) ) {
        return(FALSE)
    }
    
    sizes <- input_lens[input_lens > 0]
    if ( !length(sizes) ) { return(TRUE) }
    
    # # elems in buffer is input_lens / count
    buffer_nelems <- input_lens / count
    
    if (any( buffer_nelems != round(buffer_nelems))) {
        return(FALSE)
    }
    return(structure(TRUE, buffer_nelems = buffer_nelems))
}

# ---- data type helpers ----------------------------------------------------

is_filearray <- function(object, proxy_ok = TRUE) {
    if (!isS4(object)) { return(FALSE) }
    if ( !proxy_ok && is_fileproxy(object) ) {
        return( FALSE )
    }
    return(inherits(object, c("FileArray", "FileArrayProxy")))
}

is_fileproxy <- function(object) {
    if (!isS4(object)) { return(FALSE) }
    return(inherits(object, "FileArrayProxy"))
}


# Guess output data types for meth operators such as \code{+-*/}
operation_output_type <- function(
        type1, type2, 
        logical = c("integer", "logical"), raw = c("error", "integer"),
        float = getOption("filearray.operator.precision", "double")
) {
    
    raw <- match.arg(raw)
    has_raw <- identical(type1, "raw") || identical(type2, "raw")
    
    if ( has_raw && identical(raw, "error") ) {
        stop("non-numeric argument to binary operator")
    }
    
    if ( identical(type1, "complex") ) { return("complex") }
    if ( identical(type2, "complex") ) { return("complex") }
    
    if ( identical(type1, "double") ) { return(float) }
    if ( identical(type2, "double") ) { return(float) }
    
    if ( identical(type1, "float") ) { return(float) }
    if ( identical(type2, "float") ) { return(float) }
    
    if ( identical(type1, "integer") ) { return("integer") }
    if ( identical(type2, "integer") ) { return("integer") }
    
    if ( identical(type1, "raw") ) { return("integer") }
    if ( identical(type2, "raw") ) { return("integer") }
    
    logical <- match.arg(logical)
    if ( identical(type1, "logical") ) { return(logical) }
    if ( identical(type2, "logical") ) { return(logical) }
    
    stop("unrecognized combination of types: ", type1, ", ", type2)
}



# ---- function calls ---------------------------------------------------
parent_call <- function(def, deparse = FALSE, env = parent.frame()) {
    tryCatch({
        call <- with(env, {match.call(expand.dots = FALSE)})
        if (!missing(def)) {
            def <- substitute(def)
            call[[1]] <- def
        }
        if (deparse) {
            call <- deparse1(call)
        }
        call
    }, error = function(e) { NULL })
}

test_that("is_same_dim", {
    # sanity check
    testthat::expect_true(is_same_dim(array(1:12, c(3, 4)), matrix(NA, nrow = 3, ncol = 4)))
    testthat::expect_false(is_same_dim(array(1:12, c(3, 4)), matrix(NA, nrow = 3, ncol = 3)))
    testthat::expect_false(is_same_dim(array(1:12, c(3, 4)), array(NA, c(3, 4, 1))))
    testthat::expect_false(is_same_dim(array(1:12, c(3, 4)), array(NA, c(4, 3))))
    
    testthat::expect_false(is_same_dim(array(NA, c(3, 4)), list()))
    testthat::expect_false(is_same_dim(array(NA, c(3, 4)), data.frame()))
    testthat::expect_false(is_same_dim(array(NA, c(3, 4)), NULL))
    testthat::expect_false(is_same_dim(array(NA, c(3, 4)), 1))
    
    # number of margins <= 1
    testthat::expect_true(is_same_dim(1, 1))
    testthat::expect_false(is_same_dim(1, 1:3))
    testthat::expect_false(is_same_dim(1, 1:3))
    testthat::expect_true(is_same_dim(NULL, numeric()))
    
    # margin =1 but with dim
    x <- 1:5
    dim(x) <- 5
    testthat::expect_true(is_same_dim(x, 1:5))
    
})

test_that("operation_output_type", {
    double_type <- "float"
    options("filearray.operator.precision" = double_type)
    on.exit({
        options("filearray.operator.precision" = NULL)
    })
    
    # complex
    type1 <- "complex"
    for (type2 in c("complex", "double", "float", "integer", "logical")) {
        testthat::expect_equal(operation_output_type(type1, type2), type1)
    }
    testthat::expect_equal(operation_output_type(type1, "raw", raw = "integer"), type1)
    testthat::expect_error(operation_output_type(type1, "raw"))
    
    # double/float
    type1 <- "double"
    for (type2 in c("double", "float", "integer", "logical")) {
        testthat::expect_equal(operation_output_type(type1, type2), double_type)
    }
    testthat::expect_equal(operation_output_type(type1, "raw", raw = "integer"),
                           double_type)
    testthat::expect_error(operation_output_type(type1, "raw"))
    
    type1 <- "float"
    for (type2 in c("double", "float", "integer", "logical")) {
        testthat::expect_equal(operation_output_type(type1, type2), double_type)
    }
    testthat::expect_equal(operation_output_type(type1, "raw", raw = "integer"),
                           double_type)
    testthat::expect_error(operation_output_type(type1, "raw"))
    
    # integer
    type1 <- "integer"
    for (type2 in c("integer", "logical")) {
        testthat::expect_equal(operation_output_type(type1, type2), type1)
    }
    testthat::expect_equal(operation_output_type(type1, "raw", raw = "integer"), type1)
    testthat::expect_error(operation_output_type(type1, "raw"))
    
    # logical
    testthat::expect_equal(operation_output_type("logical", "logical"), "integer")
    testthat::expect_equal(operation_output_type("logical", "logical", logical = "logical"),
                           "logical")
    testthat::expect_equal(operation_output_type("logical", "raw", raw = "integer"),
                           "integer")
    testthat::expect_error(operation_output_type("logical", "raw"))
    
    # raw
    testthat::expect_equal(operation_output_type("raw", "raw", raw = "integer"),
                           "integer")
    testthat::expect_error(operation_output_type("raw", "raw"))
    
})

test_that("fmap_buffer", {
    buffer_large <- get_buffer_size() * 16
    
    testthat::expect_equal(common_fmap_buffer_count(84, 12, 96), 1)
    testthat::expect_equal(common_fmap_buffer_count(84, 12, 97), 1)
    
    testthat::expect_equal(common_fmap_buffer_count(buffer_large * 2, 12, 1024), 2)
    testthat::expect_equal(common_fmap_buffer_count(buffer_large * 2 + 2, 12, 1024), 2)
    testthat::expect_equal(common_fmap_buffer_count(buffer_large * 4096, buffer_large, 1024),
                           1024)
    testthat::expect_equal(common_fmap_buffer_count(buffer_large * 2, 13, 1024), 1)
    
    expect_true(validate_fmap_buffer_count(12, input_lens = c(84, 12, 96)))
    expect_true(validate_fmap_buffer_count(1, input_lens = c(84, 12, 96)))
    expect_false(validate_fmap_buffer_count(15, input_lens = c(84, 12, 96)))
    expect_true(validate_fmap_buffer_count(1, input_lens = c(84, 12, 97)))
    
})

in_rcmdcheck <- function(...) {
    evidences <- list()
    args <- commandArgs()
    is_vanilla <- is.element("--vanilla", args)
    if (!is_vanilla) {
        return(FALSE)
    }
    pwd <- getwd()
    dirname <- basename(pwd)
    parent <- basename(dirname(pwd))
    pattern <- ".+[.]Rcheck$"
    in_test <- (grepl(pattern, parent) && grepl("^tests(|_.*)$", dirname))
    if (!(in_test || grepl(pattern, dirname))) {
        return(FALSE)
    }
    in_examples <- is.element("CheckExEnv", search())
    in_win_builder <- (.Platform$OS.type == "windows" && grepl("Rterm[.]exe$", args[1]))
    if (in_win_builder) {
        n <- length(args)
        if (!all(c("--no-save", "--no-restore", "--no-site-file", 
                  "--no-init-file") %in% args)) {
            return(FALSE)
        }
        if (!grepl(pattern, parent)) {
            return(FALSE)
        }
    }
    if (in_test) {
        return(structure(TRUE, status = "tests"))
    }
    if (in_examples) {
        return(structure(TRUE, status = "examples"))
    }
    return(FALSE)
}

symlink_enabled <- local({
    enabled <- NA
    function() {
        if (!is.na(enabled)) { return(enabled) }
        tempdir(check = TRUE)
        f1 <- temp_path(pattern = "filearray_simlink_test_from")
        f2 <- temp_path(pattern = "filearray_simlink_test_to")
        on.exit({
            if (file.exists(f1)) {
                unlink(f1)
            }
            if (file.exists(f2)) {
                unlink(f2)
            }
        }, add = FALSE)
        s <- paste(sample(LETTERS), collapse = "")
        writeLines(s, con = f1)
        file.symlink(f1, to = f2)
        en <- tryCatch({
            if (identical(readLines(f2), s)) {
                TRUE
            } else {
                FALSE
            }
        }, error = function(e) {
            FALSE
        }, warning = function(e) {
            FALSE
        })
        enabled <<- en
        
        if (file.exists(f1)) {
            unlink(f1)
        }
        if (file.exists(f2)) {
            unlink(f2)
        }
        on.exit({}, add = FALSE)
        
        return(enabled)
    }
})


.onLoad <- function(libname, pkgname) {
    # Check if in R CMD check mode
    if (Sys.getenv("_R_CHECK_LIMIT_CORES_") == "TRUE") {
        #  R CMD check with --as-cran
        n <- 2L
    } else if (in_rcmdcheck()) {
        #  R CMD check (without CRAN)
        n <- 2L
    } else {
        n <- filearray_threads(-1)
        if (n > 8L) {
            n <- 8L
        }
    }
    
    filearray_threads(n)
    ns <- asNamespace(pkgname)
    ns$NA_float_ <- get_float_na()
}

.onAttach <- function(libname, pkgname) {
    if (Sys.getenv("_R_CHECK_LIMIT_CORES_") == "TRUE") {
        packageStartupMessage(
            "Found environment variable `_R_CHECK_LIMIT_CORES_`=TRUE. Using ",
            filearray_threads(), " threads.")
    } else if (in_rcmdcheck()) {
        packageStartupMessage("R CMD check mode. Using ", 
                              filearray_threads(), " threads.")
    } else {
        packageStartupMessage(
            "Filearray is using ", filearray_threads(), " threads.")
    }
}

#' @title Set or get file array threads
#' @description Will enable/disable multi-threaded reading or writing
#' at \code{C++} level. 
#' @param n number of threads to set. If \code{n} is negative,
#' then default to the number of cores that computer has. 
#' @param ... internally used
#' @return An integer of current number of threads
#' @export
filearray_threads <- function(n, ...) {
    
    if (!missing(n)) {
        setThreads(n, ...)
    }
    
    return(getThreads(FALSE))
}

setThreads <- function(n = "auto", stack_size = "auto", ...) {
    if (identical(n, "auto")) {
        n <- -1L
    } else {
        n <- as.integer(n)
        if (length(n) != 1 || is.na(n) || !is.numeric(n)) {
            stop("n must be an integer")
        }
    }
    if (identical(stack_size, "auto")) {
        stack_size <- 0L
    } else if (!is.numeric(stack_size)) {
        stop("stack_size must be an integer")
    } else {
        stack_size <- as.integer(stack_size)
    }
    if (n == -1L) {
        Sys.unsetenv("FILEARRAY_NUM_THREADS")
    } else {
        Sys.setenv(FILEARRAY_NUM_THREADS = n)
    }
    if (stack_size == 0L) {
        Sys.unsetenv("FILEARRAY_STACK_SIZE")
    } else {
        Sys.setenv(FILEARRAY_STACK_SIZE = stack_size)
    }
    invisible()
}

#' @title Create or load existing file arrays
#' @name filearray
#' @author Zhengjia Wang
#' @param filebase a directory path to store arrays in the local file 
#' system. When creating an array, the path must not exist.
#' @param x R object such as array, file array proxy, or character that can be 
#' transformed into file array
#' @param dimension dimension of the array, at least length of 2
#' @param type storage type of the array; default is \code{'double'}. Other
#' options include \code{'integer'}, \code{'logical'}, and \code{'raw'}.
#' @param mode whether allows writing to the file; choices are 
#' \code{'readwrite'} and \code{'readonly'}.
#' @param partition_size positive partition size for the last margin, or
#' \code{NA} to automatically guess; see 'Details'.
#' @param initialize whether to initialize partition files; default is false
#' for performance considerations. However, if the array is dense, it is 
#' recommended to set to true
#' @param ... additional headers to check used by \code{filearray_checkload}
#' (see 'Details'). This argument is ignored by \code{filearray_create}, 
#' reserved for future compatibility.
#' @param symlink_ok whether arrays with symbolic-link partitions can pass 
#' the test; this is usually used on bound arrays with symbolic-links; see 
#' \code{\link{filearray_bind}};
#' @param verbose whether to print out some debug messages
#' @param on_missing function to handle file array (such as initialization)
#' when a new array is created; must take only one argument, the array object
#' @param auto_set_headers whether to automatically set headers if array is 
#' missing or to be created; default is true
#' @return A \code{\link{FileArray-class}} instance.
#' 
#' @details The file arrays partition out-of-memory array objects and store them 
#' separately in local file systems. Since R stores matrices/arrays 
#' in column-major style, file array uses the slowest margin (the 
#' last margin) to slice the partitions. This helps to align the elements
#' within the files with the corresponding memory order. An array with 
#' dimension \code{100x200x300x400} has 4 margins. The length of the 
#' last margin is 400, which is also the maximum number of potential
#' partitions. The number of partitions are determined by the last margin
#' size divided by \code{partition_size}. For example, if the partition
#' size is 1, then there will be 400 partitions. If the partition size 
#' if 3, there will be 134 partitions. The default partition sizes 
#' are determined internally following these priorities:
#' \describe{
#' \item{1. }{the file size of each partition does not exceed \code{1GB}}
#' \item{2. }{the number of partitions do not exceed 100}
#' }
#' These two rules are not hard requirements. The goal is to reduce the
#' numbers of partitions as much as possible. 
#' 
#' The arguments \code{...} in \code{filearray_checkload} should be named
#' arguments that provide additional checks for the header information. 
#' The check will fail if at least one header is not identical. For example,
#' if an array contains header key-signature pair, one can use 
#' \code{filearray_checkload(..., key = signature)} to validate the signature.
#' Note the comparison will be rigid, meaning the storage type of the headers 
#' will be considered as well. If the signature stored in the array is an 
#' integer while provided is a double, then the check will result in failure.
#' 
#' @examples 
#' 
#' 
#' # Prepare
#' library(filearray)
#' filebase <- tempfile()
#' if(file.exists(filebase)) { unlink(filebase, TRUE) }
#' 
#' # create array
#' x <- filearray_create(filebase, dimension = c(200, 30, 8))
#' print(x)
#' 
#' # Assign values
#' x[] <- rnorm(48000)
#' 
#' # Subset
#' x[1,2,]
#' 
#' # load existing array
#' filearray_load(filebase)
#' 
#' x$set_header("signature", "tom")
#' filearray_checkload(filebase, signature = "tom")
#' 
#' \dontrun{
#' # Trying to load with wrong signature
#' filearray_checkload(filebase, signature = "jerry")
#' }
#' 
#' 
#' # check-load, and create a new array if fail
#' x <- filearray_load_or_create(
#'     filebase = filebase, dimension = c(200, 30, 8),
#'     verbose = FALSE, signature = "henry"
#' )
#' x$get_header("signature")
#' 
#' # check-load with initialization
#' x <- filearray_load_or_create(
#'     filebase = filebase,
#'     dimension = c(3, 4, 5),
#'     verbose = FALSE, mode = "readonly",
#'     on_missing = function(array) {
#'         array[] <- seq_len(60)
#'     }
#' )
#' 
#' x[1:3,1,1]
#' 
#' # Clean up
#' unlink(filebase, recursive = TRUE)
#' 
NULL

#' @rdname filearray
#' @export
filearray_create <- function(
    filebase, dimension, 
    type = c("double", "float", "integer", "logical", "raw", "complex"), 
    partition_size = NA,
    initialize = FALSE,
    ...)
{
    type <- match.arg(type)
    
    if (length(dimension) < 2 || any(dimension == 0)) {
        stop("Invalid dimension: FileArray dimension must not contain 0. Its length must be at least 2.")
    }
    
    size <- get_elem_size(type)
    if (is.na(partition_size)) {
        partition_size <- guess_partition(dimension, size)
    } else {
        partition_size <- round(partition_size)
        if (partition_size <= 0) {
            stop("Invalid `partition_size`: ", partition_size)
        }
    }
    
    arr <- new("FileArray")
    arr$create(
        filebase = filebase,
        dimension = dimension,
        type = type,
        partition_size = partition_size
    )
    if (initialize) {
        arr$initialize_partition()
    }
    arr
}

#' @rdname filearray
#' @export
filearray_load <- function(filebase, mode = c("readwrite", "readonly")) {
    mode <- match.arg(mode)
    arr <- new("FileArray")
    arr$load(
        filebase = filebase,
        mode = mode
    )
    arr
}

#' @rdname filearray
#' @export
filearray_checkload <- function(
    filebase, mode = c("readonly", "readwrite"), ..., symlink_ok = TRUE
) {
    mode <- match.arg(mode)
    if (!dir.exists(filebase)) {
        stop("Filearray does not exists.")
    }
    arr <- filearray::filearray_load(filebase, mode = mode)
    
    if (!symlink_ok) {
        bind_info <- arr$get_header("filearray_bind", list())
        if (isTRUE(bind_info$symlink)) {
            stop("The array partition files contain symlinks")
        }
    }
    
    if (...length() == 0) { return(arr) }
    header_list <- list(...)
    header_names <- names(header_list)
    
    for (nm in header_names) {
        expected <- arr$get_header(nm)
        given <- header_list[[nm]]
        if (nm != "" && !identical(given, expected)) {
            stop("The header `", nm, "` (", deparse1(expected), ") is not identical with given values (", deparse1(given), ").")
        }
    }
    return(arr)
}


#' @rdname filearray
#' @export
filearray_load_or_create <- function(
    filebase, dimension, on_missing = NULL, type = NA, 
    ..., mode = c("readonly", "readwrite"), symlink_ok = TRUE,
    initialize = FALSE, partition_size = NA, verbose = FALSE, 
    auto_set_headers = TRUE
) {
    mode <- match.arg(mode)
    filebase <- normalizePath(filebase, mustWork = FALSE, winslash = "/")
    if (length(filebase) != 1 || grepl("(^|^[A-Za-z]:)/$", filebase)) {
        stop("Invalid filebase to store a file array.")
    }
    
    
    dimension <- as.integer(dimension)
    if (length(dimension) < 2 || any(is.na(dimension) | dimension < 0)) {
        stop("Incorrect dimension for a file array: `dimension` must a valid positive integer vector with length of two or above.")
    }
    
    if (!is.null(on_missing)) {
        if (!is.function(on_missing) || !length(formals(on_missing))) {
            stop("`filearray_load_or_create`: `on_missing` must be a function with one argument (i.e. the file array)")
        }
    }
    
    
    additional_headers <- list(...)
    add_header_names <- names(additional_headers)
    if (length(additional_headers)) {
        if (!length(add_header_names) || "" %in% trimws(additional_headers)) {
            stop("`filearray_load_or_create`: additional parameters must be named.")
        }
    }
    
    
    arr <- tryCatch(
        expr = {
            # try to load existing array
            arr <- filearray_checkload(
                filebase = filebase, mode = mode, 
                symlink_ok = symlink_ok, ...
            )
            
            # If no error raised, the array has been loaded
            
            if (!is.na(type)) {
                if (!identical(arr$type(), type)) {
                    stop("`filearray_load_or_create`: Requested array type does not match with existing array.")
                }
            } else {
                # in case the array needs to be reconstructed, assuming the type
                # remain the same
                type <- arr$type()
            }
            
            # Now check the dimension
            arr_dim <- as.integer(arr$dimension())
            if (!identical(arr_dim, dimension)) {
                stop("`filearray_load_or_create`: Requested dimension does not match with existing array.")
            }
            
            arr
        },
        error = function(e) {
            if (verbose) {
                message("`filearray_load_or_create`: cannot load the existing file array: ", e$message, "\nTrying creating a new one. If the array already exists, its file path will be removed.")
            }
            if (file.exists(filebase)) {
                unlink(filebase, recursive = TRUE, force = TRUE)
            }
            pdir <- dirname(filebase)
            if (!dir.exists(pdir)) {
                dir.create(pdir, showWarnings = FALSE, recursive = TRUE)
            }
            # create the array
            if (is.na(type)) { type <- "double" }
            arr <- filearray_create(
                filebase = filebase,
                dimension = dimension,
                type = type,
                partition_size = partition_size,
                initialize =  initialize
            )
            # seal the header
            if ( auto_set_headers ) {
                for (nm in add_header_names) {
                    arr$set_header(key = nm, value = additional_headers[[nm]], save = FALSE)
                }
            }
            # run on_missing if the function exists
            if (is.function(on_missing)) {
                arr$.mode <- "readwrite"
                on_missing(arr)
            }
            arr$.save_header()
            arr
        }
    )
    
    # set mode
    arr$.mode <- mode
    arr
}

test_that("bind (FileArray)", {
    x <- array(rnorm(120), 2:5)
    y <- filearray_create(tempfile(),
                          dimension = c(2, 3, 4, 10),
                          partition_size = 3L)
    z <- filearray_create(tempfile(),
                          dimension = c(2, 3, 4, 10),
                          partition_size = 3L)
    
    options("filearray.quiet" = FALSE)
    on.exit({
        options("filearray.quiet" = FALSE)
        y$delete(force = TRUE)
        z$delete(force = TRUE)
    }, add = TRUE)
    lapply(1:10, function(ii) {
        if (ii %% 2 == 0) {
            y[, , , ii] <- x[, , , ii / 2]
            z[, , , ii] <- x[, , , ii / 2]
        }
    })
    
    testthat::expect_warning({
        w <- filearray_bind(y, z, symlink = FALSE)
        w$delete()
    }, regexp = "^One or more arrays have last margin size.+")
    
    
    
    options("filearray.quiet" = TRUE)
    y$expand(n = 12)
    z$expand(n = 12)
    w <- filearray_bind(y, z, symlink = TRUE)
    l <- filearray_load(w$.filebase, mode = "readonly")
    
    on.exit({
        w$.mode <- "readwrite"
        w$delete()
        l$.mode <- "readwrite"
        l$delete()
    }, add = TRUE)
    
    expect_null({
        filearray_checkload(filebase = w$.filebase,
                            symlink_ok = TRUE)
        filearray_checkload(filebase = w$.filebase, partition = 3)
        NULL
    })
    
    if (w$.header$filearray_bind$symlink) {
        expect_error({
            filearray_checkload(filebase = w$.filebase,
                                symlink_ok = FALSE)
        })
    }
    
    expect_identical(w[], l[])
    expect_identical(w[, , , seq(2, 10, 2)], x)
    expect_identical(w[, , , seq(2, 10, 2) + 12], x)
    expect_identical(l[, , , seq(2, 10, 2)], x)
    expect_identical(l[, , , seq(2, 10, 2) + 12], x)
    
    expect_identical(l[, , , seq(2, 10, 2) + c(0, 12, 12, 0, NA)], x[, , , c(1:4, NA)])
    
    expect_equal(l$collapse(
        keep = c(2, 3),
        method = "sum",
        na.rm = TRUE
    ),
    apply(x, c(2, 3), sum) * 2)
    
    # Check if cached bind works
    l <- filearray_bind(
        y,
        z,
        filebase = w$.filebase,
        symlink = w$.header$filearray_bind$symlink,
        overwrite = TRUE,
        cache_ok = TRUE
    )
    
    expect_true(attr(l, "cached_bind"))
    
})


test_that("bind (FileArrayProxy)", {
    y <- filearray_create(
        tempfile(),
        dimension = c(2, 3, 4, 10),
        partition_size = 3L,
        type = "integer"
    )
    z <- filearray_create(
        tempfile(),
        dimension = c(2, 3, 4, 10),
        partition_size = 3L,
        type = "double"
    )
    
    options("filearray.quiet" = FALSE)
    on.exit({
        options("filearray.quiet" = FALSE)
        y$delete(force = TRUE)
        z$delete(force = TRUE)
    }, add = TRUE)
    
    y[] <- rep(0L, 240)
    z[] <- rep(0L, 240)
    x <- array(rnorm(240), c(2, 3, 4, 10))
    
    y <- y + x
    z <- z + x
    
    testthat::expect_error(y$expand(n = 12))
    
    options("filearray.quiet" = TRUE)
    w <- filearray_bind(y, z, symlink = TRUE)
    l <- filearray_load(w$.filebase, mode = "readonly")
    
    on.exit({
        w$.mode <- "readwrite"
        w$delete()
        l$.mode <- "readwrite"
        l$delete()
    }, add = TRUE)
    
    expect_null({
        filearray_checkload(filebase = w$.filebase,
                            symlink_ok = TRUE)
        filearray_checkload(filebase = w$.filebase, partition = 3)
        NULL
    })
    
    if (w$.header$filearray_bind$symlink) {
        expect_error({
            filearray_checkload(filebase = w$.filebase,
                                symlink_ok = FALSE)
        })
    }
    
    expect_identical(w[], l[])
    expect_identical(w[, , , 1:10, dimnames = NULL], x)
    expect_identical(w[, , , 1:10 + 12, dimnames = NULL], x)
    expect_identical(l[, , , 1:10, dimnames = NULL], x)
    expect_identical(l[, , , 1:10 + 12, dimnames = NULL], x)
    
    expect_identical(l[, , , seq(2, 10, 2) + c(0, 12, 12, 0, NA), dimnames = NULL], x[, , , c(1:4 *
                                                                                                  2, NA)])
    
    expect_equal(l$collapse(
        keep = c(2, 3),
        method = "sum",
        na.rm = TRUE
    ),
    apply(x, c(2, 3), sum) * 2)
    
    expect_equal(y$collapse(
        keep = c(2, 3),
        method = "sum",
        na.rm = TRUE
    ),
    apply(x, c(2, 3), sum))
    expect_equal(z$collapse(
        keep = c(2, 3),
        method = "sum",
        na.rm = TRUE
    ),
    apply(x, c(2, 3), sum))
    
    
    # Check if cached bind works
    # l <- filearray_bind(y, z, filebase = w$.filebase, symlink = w$.header$filearray_bind$symlink, overwrite = TRUE, cache_ok = TRUE)
    #
    # expect_true(attr(l, "cached_bind"))
    
})

## usethis namespace: start
#' @importFrom Rcpp sourceCpp
#' @useDynLib filearray, .registration = TRUE
## usethis namespace: end
NULL

#' @title 'S3' methods for 'FileArray'
#' @name S3-filearray
#' @description These are 'S3' methods for 'FileArray'
#' @param x a file array
#' @param drop whether to drop dimensions; see topic \code{\link[base]{Extract}}
#' @param reshape a new dimension to set before returning subset results; default is \code{NULL} (use default dimensions)
#' @param strict whether to allow indices to exceed bound; currently only accept \code{TRUE}
#' @param dimnames whether to preserve \code{\link[base]{dimnames}}
#' @param value value to substitute or set
#' @param na.rm whether to remove \code{NA} values during the calculation
#' @param split_dim internally used; split dimension and calculate indices to
#' manually speed up the subset; value ranged from 0 to size of dimension minus
#' one.
#' @param lazy whether to lazy-evaluate the method, only works when assigning
#' arrays with logical array index
#' @param i,... index set, or passed to other methods
#' @param .env environment to evaluate formula when evaluating subset margin indices.
NULL

#' @name S4-filearray
#' @title 'S4' methods for \code{FileArray}
#' @param x,z,e1,e2 \code{FileArray} or compatible data
#' @param base,digits,... passed to other methods
#' @returns See \code{\link[methods]{S4groupGeneric}}
NULL

# DIPSAUS DEBUG START
# verbose <- TRUE
# z = filearray_create(temp_path(), c(2,3,4))
# z[] <- 1:24
# aaa <- z+1
# print(aaa[,,4])
# idx <- as_filearray(z[] > 2)
# 
# print(z[idx])
# print(aaa[idx])

# Normal subset via indices
fa_subset1 <- function(x, ..., drop = TRUE, reshape = NULL, strict = TRUE, dimnames = TRUE, split_dim = 0) {
    if (!x$valid()) {
        stop("Invalid file array")
    }
    drop <- isTRUE(drop)
    # file <- temp_path(); x <- filearray_create(file, c(300, 400, 100, 1))
    arglen <- ...length()
    elem_size <- x$element_size()
    dim <- x$dimension()
    
    listOrEnv <- list()
    if (arglen == 1) {
        missing_args <- check_missing_dots(environment())
        if (length(missing_args) > 0 && !missing_args[[1]]) {
            tmp <- ...elt(1)
            if (length(tmp)) {
                stop("Subset FileArray only allows x[], x[i:j] or x[i,j,...] (single logical index like x[c(TRUE, ...)] is not allowed)")
            }
        }
        
        
    } else if (arglen > 1) {
        if (arglen != length(dim)) {
            stop("Subset FileArray dimension mismatch.")
        }
        missing_args <- check_missing_dots(environment())
        
        for (ii in seq_len(arglen)) {
            if ( missing_args[[ii]] ) {
                listOrEnv[[ii]] <- seq_len(dim[[ii]])
            } else {
                tmp <- ...elt(ii)
                if (!length(tmp)) {
                    tmp <- integer(0L)
                } else if (is.logical(tmp)) {
                    if (length(tmp) > dim[[ii]]) {
                        stop("(subscript) logical subscript too long")
                    }
                    tmp <- rep(tmp, ceiling(dim[[ii]] / length(tmp)))
                    tmp <- tmp[seq_len(dim[[ii]])]
                    tmp <- seq_along(tmp)[tmp]
                }
                listOrEnv[[ii]] <- tmp
            }
        }
    }
    
    # guess split dim
    max_buffer <- get_buffer_size() / elem_size
    
    if (length(listOrEnv) == length(dim)) {
        idxrange <- sapply(listOrEnv, function(x) {
            if (!length(x) || all(is.na(x))) { return(1L) }
            rg <- range(x, na.rm = TRUE)
            return(rg[2] - rg[1] + 1)
        })
    } else {
        idxrange <- dim
    }
    split_dim <- as.integer(split_dim)
    if (is.na(split_dim) || split_dim <= 0 || split_dim >= length(dim)) {
        # worst-case time-complexity
        time_complexity <-
            sapply(seq_len(length(dim) - 1), function(split_dim) {
                dim[[length(dim)]] <- 1
                idx1dim <- dim[seq_len(split_dim)]
                idx1dim[[split_dim]] <- idxrange[[split_dim]]
                idx1len <- prod(idx1dim)
                idx2len <- prod(dim[-seq_len(split_dim)])
                buffer_sz <-
                    ifelse(idx1len > max_buffer, max_buffer, idx1len)
                nloops <- ceiling(idx1len / buffer_sz)
                (idx1len * nloops + idx2len) * idx2len
            })
        split_dim <- which.min(time_complexity)
        split_dim <- split_dim[[length(split_dim)]]
    }
    
    
    FARR_subset2(
        filebase = x$.filebase,
        listOrEnv = listOrEnv,
        reshape = reshape,
        drop = drop,
        use_dimnames = isTRUE(dimnames),
        thread_buffer = get_buffer_size(), 
        split_dim = split_dim,
        strict = isTRUE(strict)
    )
    
}

# Special subset using another array
fa_subset2 <- function(x, i, ...) {
    
    single_index <- FALSE
    if (!is_filearray(i)) {
        if (is.array(i) && is.logical(i)) {
            if (!is_same_dim(x, i)) {
                stop("x[i]: `x` and `i` must share the same dimension when i is a logical array.")
            }
            i <- as_filearray(i)
        } else if (is.numeric(i) && is.vector(i)) {
            single_index <- TRUE
        } else {
            stop("x[i]: Cannot determine subset mode from `i`: must be a logical array or integer vector")
        }
    }
    
    if ( single_index ) {
        stop("Single vector indexing (e.g. x[c(1,2,3,4,...)]) hasn't been implemented yet.")
        # re <- fastmap::fastqueue()
        # idx <- fastmap::fastmap()
        # idx$set("start_idx", 0L)
        # idx$set("idx_left", length(i))
        # 
        # 
        # fa_eval_ops(x, addon = function(env, data, uuid) {
        #     idx_left <- idx$get("idx_left")
        #     
        #     if ( idx_left < 0 ) {
        #         return(NULL)
        #     }
        #     start_idx <- idx$get("start_idx")
        #     
        #     vec <- env[[ uuid ]]
        #     
        #     end_idx <- start_idx + length(vec)
        #     sel <- i[ i > start_idx & i <= end_idx] - start_idx
        #     if (length(sel)) {
        #         if (idx_left <= length(sel)) {
        #             sel <- sel[seq_len(idx_left)]
        #             re$add(vec[sel])
        #             idx_left <- 0L
        #         } else {
        #             re$add(vec[sel])
        #         }
        #         idx$set("idx_left", idx_left - length(sel))
        #     }
        #     idx$set("start_idx", start_idx)
        # })
        # re <- unlist(re$as_list())
        # 
    } else {
        re <- unlist(
            fmap2(
                x = list(x, i),
                fun = function(v) {
                    v[[1]][v[[2]]]
                }, 
                .simplify = FALSE
            )
        )
    }
    
    expected_mode <- operation_output_type(
        typeof(x), typeof(x), float = "double", logical = "logical")
    if (!identical(expected_mode, mode(re))) {
        mode(re) <- expected_mode
    }
    return(re)
    
}

#' @describeIn S3-filearray subset array
#' @export
`[.FileArray` <- function(x, i, ..., drop = TRUE, reshape = NULL, strict = TRUE, dimnames = TRUE, split_dim = 0) {
    x <- fa_eval_ops(x)
    if (missing(i)) {
        return(fa_subset1(x, , ..., drop = drop, reshape = reshape, strict = strict, dimnames = dimnames, split_dim = split_dim))
    } else {
        if (is_filearray(i) || is.array(i) || 
           ( ...length() == 0 && !is.logical(i) )) {
            return(fa_subset2(x, i))
        } else {
            return(fa_subset1(x, i, ..., drop = drop, reshape = reshape, strict = strict, dimnames = dimnames, split_dim = split_dim))
        }
    }
}

#' Position (size): description - example value
#' 0 (8): determine endianness - 1.0 (double)
#' 8 (4 x 3): version number - [1, 0, 0] (int[3])
#' 20 (4): SEXP type - 14L (SEXPTYPE, data are double type)
#' 24 (4): element size - 8L (int, data element size is 8)
#' 28 (8): number of last dimensions per partition - 1 (double)
#' 36 (8): total number of elements - 24 (double)
#' 44 (4): number of dimensions (ndims) - 3 (int)
#' 48 (8 x ndims): dimensions - [2, 3, 4] (double)
#' 
#' ...
#' 
#' 1012 (4): size of the header - 72 (int)
#' 1016 (8): number of elements in the file - (double)
#' 
#' @noRd
NULL

write_header <- function(fid, partition, dimension, type, size) {
    seek(con = fid, where = 0, rw = "write")
    
    # 0 (8): determine endianness - 1.0 (double)
    writeBin(con = fid, object = 1.0, size = 8L, endian = ENDIANNESS)
    
    # 8 (4 x 3): version number - [1, 0, 0] (int[3])
    writeBin(con = fid, object = FILE_VER, size = 4L, endian = ENDIANNESS)
    
    sexp_type <- type_to_sexp(type)
    
    # 20 (4): SEXP type - 14L (SEXPTYPE, data are double type)
    writeBin(con = fid, object = sexp_type, size = 4L,
             endian = ENDIANNESS)
    
    # 24 (4): element size - 8L (int, data element size is 8)
    writeBin(con = fid, object = as.integer(size), size = 4L,
             endian = ENDIANNESS)
    
    # partition number
    partition <- as.double(partition)
    
    # 28 (8): number of last dimensions per partition - 1 (double)
    writeBin(con = fid, object = partition, size = 8L, endian = ENDIANNESS)
    
    ## # write length of dimension
    dimension <- as.double(dimension)
    len <- prod(dimension)
    
    # 36 (8): total number of elements - 24 (double)
    writeBin(con = fid, object = len, size = 8L, endian = ENDIANNESS)
    
    # 44 (4): number of dimensions (ndims) - 3 (int)
    writeBin(con = fid, object = length(dimension), size = 4L, 
             endian = ENDIANNESS)
    
    # 48 (8 x ndims): dimensions - [2, 3, 4] (double)
    writeBin(con = fid, object = dimension, size = 8L, 
             endian = ENDIANNESS)
    header_len <- 48L + 8L * length(dimension)
    writeBin(con = fid, object = rep(0L, (HEADER_SIZE - header_len) / 4 - 3),
             size = 4L, endian = ENDIANNESS)
    writeBin(con = fid, object = header_len, size = 4L, endian = ENDIANNESS)
    writeBin(con = fid, object = 0.0, size = 8L, endian = ENDIANNESS)
    seek(con = fid, where = 0, rw = "write")
}

read_header <- function(fid) {
    seek(con = fid, where = 0, rw = "read")
    
    one <- readBin(fid, n = 1, size = 8L, what = "double", endian = ENDIANNESS)
    native <- one == 1
    
    if ( native ) {
        endian <- ENDIANNESS
    } else {
        if ( ENDIANNESS == "little" ) {
            endian <- "big"
            # TODO: support big-endian file? maybe not
            stop("The file endianess is not little?")
        } else {
            endian <- "little"
        }
    }
    
    version <- readBin(fid, what = "int", n = 3, size = 4L, endian = endian)
    
    # type, size
    sexp_type <- readBin(fid, what = "int", size = 4L, endian = endian)
    vsize <- readBin(fid, what = "int", size = 4L, endian = endian)
    
    # partition number
    partition <- readBin(con = fid, what = "double", size = 8L, endian = endian)
    
    # dimension
    size <- readBin(con = fid, what = "double", size = 8L, endian = endian)
    ndims <- readBin(con = fid, what = "int", size = 4L, endian = endian)
    dim <- readBin(con = fid, what = "double", size = 8L, n = ndims, 
                   endian = endian)
    
    seek(con = fid, where = HEADER_SIZE - 12L, rw = "read")
    header_bytes <- readBin(con = fid, what = "int", size = 4L, 
                            endian = endian)
    
    content_length <- readBin(con = fid, what = "double", size = 8L, 
                              endian = endian)
    list(
        endianness = endian,
        version = version,
        sexp_type = sexp_type,
        unit_bytes = vsize,
        partition = partition,
        partition_size = size,
        partition_dim = dim,
        header_bytes = header_bytes,
        content_length = content_length
    )
}

validate_header <- function(file, fid) {
    fz <- -1
    if (!missing(file)) {
        if (!file.exists(file)) {
            stop("File is missing")
        }
        fz <- file.size(file)
        if (fz < HEADER_SIZE) {
            # Might be on windows and partition files are symlinked
            if (get_os() != "windows" || fz != 0) {
                stop("Invalid `filearray` partition. File size too small:\n  ", file)
            }
        }
        fid <- file(description = file, open = "rb")
        on.exit({
            close(fid)
        })
    }
    
    header <- read_header(fid)
    if ( header$header_bytes != 48 + 8 * length(header$partition_dim) ) {
        stop("Filearray partition header is corrupted.")
    }
    if ( fz > 0 && header$content_length + HEADER_SIZE > fz ) {
        stop("Filearray data is corrupted")
    }
    
    return(header)
}

set_meta_content <- function(meta_file, data) {
    stopifnot(file.exists(meta_file))
    data <- as.list(data)
    data$header_version <- HEADER_VER
    
    keys <- names(data)
    names(data) <- sprintf("__%s__", keys)
    
    conn <- rawConnection(raw(), "w+b")
    saveRDS(file = conn, data, ascii = FALSE)
    v <- rawConnectionValue(conn)
    close(conn)
    
    fid <- file(meta_file, "r+b")
    on.exit({ close(fid) })
    seek(con = fid, where = HEADER_SIZE, origin = "start", rw = "write")
    writeBin(object = v, con = fid, endian = ENDIANNESS)
    
    seek(con = fid, where = HEADER_SIZE - 8L, origin = "start", rw = "write")
    writeBin(con = fid, object = as.double(length(v)), size = 8L, endian = ENDIANNESS)
    seek(con = fid, where = 0, rw = "write")
}

# Get internal UUIDs for proxy
get_inital_uuids <- function(arrays) {
    return(vapply(arrays, function(a) { a$.uuid }, "", USE.NAMES = FALSE))
}

# Get output UUIDs for proxy
get_uuids <- function(arrays) {
    return(vapply(arrays, function(a) { a$uuid() }, "", USE.NAMES = FALSE))
}

#' @rdname filearray
#' @export
FileArrayProxy <- setRefClass(
    Class = "FileArrayProxy",
    contains = "FileArray",
    fields = list(
        .uuid = "character",
        .ops = "list",
        .linked = "list",
        persist = "logical"
    ),
    methods = list(
        .mature = function(env, data, verbose = FALSE) {
            # make sure the elements of .linked are matured
            lapply(.self$.linked, function(arr) { 
                arr$.mature(env = env, data = data, verbose = verbose) 
            })
            
            lapply(.self$.ops, function(op) {
                op$op_func(env = env, data = data, verbose = verbose)
                return(NULL)
            })
            return(invisible())
        },
        initialize = function() {
            # callNextMethod()
            if (length(.self$.filebase) == 1) {
                .self$.uuid <- sprintf("0000-%s", .self$.filebase)
            } else {
                .self$.uuid <- new_uuid()
            }
        },
        uuid = function() {
            if (length(.self$.ops)) {
                return(.self$.ops[[length(.self$.ops)]]$uuid)
            }
            return(.self$.uuid)
        },
        get_all_uuids = function() {
            return(c(.self$.uuid, vapply(.self$.ops, "[[", "uuid", FUN.VALUE = "")))
        },
        
        # Only get unique arrays
        linked_arrays = function(exclude_uuids = NULL) {
            re <- list()
            if (!.self$.uuid %in% exclude_uuids) {
                re[[.self$.uuid]] <- .self
                exclude_uuids <- c(exclude_uuids, .self$get_all_uuids())
            }
            for (proxy in .self$.linked) {
                linked_subs <- proxy$linked_arrays(exclude_uuids = exclude_uuids)
                if (length(linked_subs)) {
                    uuids <- get_inital_uuids(linked_subs)
                    uuids <- uuids[!uuids %in% exclude_uuids]
                    re[uuids] <- linked_subs[uuids]
                }
            }
            return(re)
        },
        link_proxy = function(x) {
            .self$.linked[[x$uuid()]] <- x
            return(invisible())
        },
        add_operator = function(
            operator, context = "element_wise", label = "Unknown operator",
            out_type = c("double", "complex", "float", "integer", "raw", "logical")
        ) {
            out_type <- match.arg(out_type)
            
            op_idx <- length(.self$.ops) + 1L
            output_uuid <- new_uuid(op_idx)
            
            out_mode <- ifelse(out_type == "float", "double", out_type)
            
            .self$.ops[[op_idx]] <- list(
                uuid = output_uuid, 
                label = label,
                
                # Output type
                output_type = out_type,
                
                # operation type
                context = context,
                
                # operation function
                op_func = function(env, data, verbose = FALSE) {
                    if (is.null(env[[ output_uuid ]])) {
                        if ( verbose ) {
                            message(label)
                        }
                        tmp <- operator(env, data)
                        if (!identical(typeof(tmp), out_mode)) {
                            storage.mode(tmp) <- out_mode
                        }
                        env[[ output_uuid ]] <- tmp
                    }
                }
            )
        },
        collapse = function(
            keep, method = c("mean", "sum"),
            transform = c("asis", "10log10", "square", "sqrt", "normalize"), 
            na.rm = FALSE
        ) {
            # TODO This is a quick fix
            method <- match.arg(method)
            transform <- match.arg(transform)
            fa_eval_ops(.self)$collapse(
                keep = keep, method = method,
                transform = transform, 
                na.rm = na.rm
            )
        },
        expand = function(n) {
            stop("Cannot expand a proxy array. Please mature the array first before expanding.")
        },
        show = function() {
            if (.self$valid()) {
                cat("Reference class object of class \"FileArrayProxy\"\n")
                cat("Mode:", .self$.mode, "\n")
                tryCatch({
                    cat(sprintf("UUID: %s (depth=%d)\n", .self$uuid(), length(.self$.ops)))
                    cat("Dimension:", paste(.self$dimension(), collapse = "x"), "\n")
                    cat("Partition count:", nrow(.self$.partition_info), "\n")
                    cat("Partition size:", .self$partition_size(), "\n")
                    cat("Data type:", typeof(.self), "\n")
                    cat("Internal type:", .self$type(), "\n")
                }, error = function(e) {
                    quiet_warning("Partition information is unavailable: might be broken or improperly set.", immediate. = FALSE)
                })
                cat("Location:", .self$.filebase, "\n")
            } else {
                if (.self$.valid) {
                    cat("The FileArray is invalid (it has not been initialized).")
                } else {
                    cat("The FileArray is invalid (unable to locate the array in the file system).")
                }
            }
            
        },
        mature = function(...) {
            return(fa_eval_ops(.self, ...))
        }
    )
)

#' @rdname filearray
#' @export
as_filearrayproxy <- function(x, ...) {
    UseMethod("as_filearrayproxy")
}

#' @export
as_filearrayproxy.FileArray <- function(x, ...) {
    proxy <- new("FileArrayProxy")
    proxy$load(x$.filebase)
    if (length(x$.filebase) == 1) {
        proxy$.uuid <- sprintf("0000-%s", x$.filebase)
    }
    return(proxy)
}

#' @export
as_filearrayproxy.FileArrayProxy <- function(x, ...) {
    proxy <- x$copy()
    proxy$.uuid <- x$.uuid
    return(proxy)
    
}

#' @export
as_filearrayproxy.default <- function(x, ...) {
    return(as_filearrayproxy.FileArray(
        as_filearray(x, ...)
    ))
}

fa_pairwise_operator <- function(e1, e2, op, out_type = NULL, label = NULL) {
    stopifnot(is_filearray(e1) || is_filearray(e2))
    if (is.null(out_type)) {
        out_type <- operation_output_type(typeof(e1), typeof(e2))
    }
    if (is.null(label)) {
        if (is.character(op)) {
            label <- sprintf("e1 %s e2", op)
        } else {
            label <- deparse1(op)
        }
    }
    
    # check if e1 is scalar
    if (length(e1) == 1) {
        e2 <- as_filearrayproxy(e2)
        uuid2 <- e2$uuid()
        op_func <- function(v, ...) {
            # v is an env with names: uuids
            return(do.call(op, list(e1, v[[ uuid2 ]])))
        }
        e2$add_operator( op_func, out_type = out_type, context = "scalar", label = label )
        return(e2)
    }
    # check if e2 is scalar
    if (length(e2) == 1) {
        e1 <- as_filearrayproxy(e1)
        uuid1 <- e1$uuid()
        op_func <- function(v, ...) {
            # v is an env with names: uuids
            return(do.call(op, list(v[[ uuid1 ]], e2)))
            return(v[[ uuid1 ]] + e2)
        }
        e1$add_operator( op_func, out_type = out_type, context = "scalar", label = label )
        return(e1)
    }
    
    # check if e1 or e2 is numerical
    if (!is_filearray(e1)) {
        if (!is_same_dim(e1, e2)) {
            stop("non-conformable arrays")
        }
        e1 <- as_filearray(e1, dimension = dim(e2))
    }
    
    if (!is_filearray(e2)) {
        if (!is_same_dim(e2, e1)) {
            stop("non-conformable arrays")
        }
        e2 <- as_filearray(e2, dimension = dim(e1))
    }
    
    # e1 and e2 must be filearray or filearray proxy
    if (!is_same_dim(e1, e2)) {
        stop("non-conformable arrays")
    }
    
    e1 <- as_filearrayproxy(e1)
    e2 <- as_filearrayproxy(e2)
    
    
    uuid1 <- e1$uuid()
    uuid2 <- e2$uuid()
    e1$link_proxy( e2 )
    
    op_func <- function(v, ...) {
        return(do.call(op, list(v[[ uuid1 ]], v[[ uuid2 ]])))
    }
    
    e1$add_operator( op_func, out_type = out_type, context = "array", label = label )
    
    return( e1 )
    
}

fa_operator <- function(x, op, ..., out_type = NULL, label = NULL) {
    stopifnot(is_filearray(x))
    if (is.null(out_type)) {
        out_type <- typeof(x)
    }
    if (is.null(label)) {
        if (is.null(label)) {
            if (is.character(op)) {
                label <- sprintf("%s(x)", op)
            } else {
                label <- deparse1(op)
            }
        }
    }
    
    x <- as_filearrayproxy(x)
    uuid <- x$uuid()
    v <- NULL
    args <- list(quote(v[[ uuid ]]), ...)
    
    op_func <- function(v, ...) {
        # v is an env with names: uuids
        return(do.call(op, args))
    }
    x$add_operator( op_func, out_type = out_type, context = "scalar", label = label )
    return(x)
}

# Be careful when using addon, must be deterministic, or signature will be invalid
fa_eval_ops <- function(
        x, addon = NULL, verbose = FALSE, input_size = NA_integer_, 
        filebase = NULL, use_cache = is.null(filebase)) {
    
    # DIPSAUS DEBUG START
    # verbose <- TRUE
    # addon <- NULL
    # input_size <- NA
    # filebase <- NULL
    # z = filearray_create(tempfile(), c(2,3,4))
    # z[] <- 1:24
    # y <- (z + 1 + (z - 1))
    # x <- y + (1+y)
    # result <- fa_eval_ops(x, verbose = TRUE)
    # print(result[])
    
    
    has_addon <- is.function(addon)
    if ( has_addon ) {
        fml <- formals(addon)
        if (length(fml) < 3 && !"..." %in% names(fml)) {
            stop("`fa_eval_ops`: addon function must take 3 parameters (environment, list of data, output UUID)")
        }
    }
    
    if (!inherits(x, "FileArrayProxy") ) {
        if ( has_addon ) {
            x <- as_filearrayproxy(x)
        } else {
            return(x)
        }
    }
    if (!length(x$.ops) && !has_addon) { return(x) }
    
    if ( !has_addon ) {
        addon_func <- function(...) {}
    } else {
        addon_func <- addon
    }
    
    arrays <- x$linked_arrays()
    # the first uuid is x's
    
    uuids <- get_inital_uuids(arrays)
    is_unique <- !duplicated(uuids)
    
    uuids <- uuids[is_unique]
    arrays <- arrays[is_unique]
    
    uuid_x <- x$uuid()
    
    dm_x <- dim(x)
    
    verbose <- as.logical(verbose)
    
    signature <- digest::digest(list(
        x$.uuid,
        x$get_all_uuids(),
        addon = addon
    ))
    if (length(filebase) != 1 || !is.character(filebase) || is.na(filebase)) {
        fbase <- file.path(tempdir(check = TRUE), sprintf("filearrayproxy_%s", signature))
    } else {
        fbase <- filebase
    }
    
    
    if (verbose) {
        message("Results will be saved to: ", fbase)
    }
    
    re <- filearray_load_or_create(
        filebase = fbase, mode = "readwrite", 
        dimension = dm_x,
        type = typeof(x),
        partition_size = x$partition_size(),
        signature = signature,
        verbose = verbose,
        initialize = FALSE,
        on_missing = function(arr) {
            headers <- x$.header
            arr$.header <- headers[!names(headers) %in% RESERVED_HEADERS]
            arr$.header$dimnames <- x$.header$dimnames
            # x header might have matured, force unset
            arr$.header$matured <- FALSE
        }
    )
    
    if (!use_cache || !isTRUE(re$get_header("matured")) || has_addon) {
        
        if ( length(input_size) == 1 && !is.na(input_size) ) {
            nruns <- length(x) / input_size
            if (!isTRUE(nruns == round(nruns))) {
                input_size <- NULL
            }
        } 
        if ( length(input_size) != 1 || is.na(input_size) ) {
            input_size <- guess_fmap_buffer_size(dim(x), x$element_size())
        }
        
        matured <- tryCatch({
            fmap_element_wise_internal(arrays, function(data) {
                env <- new.env(parent = emptyenv(), hash = TRUE)
                lapply(seq_along(uuids), function(i) {
                    env[[ uuids[[i]] ]] <- data[[ i ]]
                    return(NULL)
                })
                x$.mature(env, data = data, verbose = verbose)
                
                # tmp <- as.list(env)
                # print(tmp[order(names(tmp))])
                
                if (has_addon) {
                    addon_func(env, data, uuid_x)
                }
                
                env[[ uuid_x ]]
            }, .y = re, .input_size = input_size)
            TRUE
        }, error = function(e) {
            stop(e)
            FALSE
        })
        
        re$set_header(key = "matured", value = matured, save = TRUE)
    }
    re$.mode <- "readonly"
    return(re)
}


# DIPSAUS DEBUG START
# z = filearray_create(tempfile(), c(2,3,4), type = "integer")
# z[] <- 1:24
# e1 = as_filearrayproxy(z)
# e1$.ops |> length()
# e2 <- fa_add_scalar(e1,1)
# e2$.ops |> length()
# x = fa_add_filearray(e1, e2)
# #x <- e2
# x$.ops |> length()
# .self = x
# exclude_uuids=NULL
# data <- list(1:24)
# op = .self$.ops[[1]]
# verbose <- TRUE

test_that("Simple operators", {
  
    # logical
    x <- as_filearrayproxy(1:24, dimension = c(4, 6), type = "double")
    x[3:4, 3] <- rep(NA, 2)
    y <- !is.na(x)
    testthat::expect_s4_class(y, "FileArrayProxy")
    testthat::expect_equal(y[], !is.na(x[]))
    
    # math to double
    x0 <- x[]
    suppressWarnings({
        for (op in FILEARRAY_SIMPLE_OPS$math) {
            y <- do.call(op, list(x))
            testthat::expect_s4_class(y, "FileArrayProxy")
            testthat::expect_equal(y[], do.call(op, list(x0)))
        }
    })
    
    # math to complex
    x0 <- array(rnorm(20) + rnorm(20) * 1i, c(5, 4))
    x <- as_filearrayproxy(x0)
    suppressWarnings({
        for (op in FILEARRAY_SIMPLE_OPS$math) {
            valid_op <- tryCatch({
                y0 <- do.call(op, list(x0))
                TRUE
            }, error = function(e) {
                FALSE
            })
            
            if ( valid_op ) {
                y <- do.call(op, list(x))
                testthat::expect_s4_class(y, "FileArrayProxy")
                dif <- y[dimnames = NULL] - y0
                mag <- Mod(y[dimnames = NULL] + y0)
                mag[mag < 1] <- 1
                dif <- dif / Mod(mag)
                testthat::expect_lt(max(abs(Re(dif))), 1e-5, label = sprintf("Operator `%s` on complex", op))
                testthat::expect_lt(max(abs(Im(dif))), 1e-5, label = sprintf("Operator `%s` on complex", op))
            } else {
                testthat::expect_error(do.call(op, list(x)), label = sprintf("Operator `%s` on complex should fail, but", op))
            }
            
            
        }
    })

    # complex
    for (op in FILEARRAY_SIMPLE_OPS$complex) {
        y0 <- do.call(op, list(x0))
        y <- do.call(op, list(x))
        testthat::expect_s4_class(y, "FileArrayProxy")
        dif <- y[dimnames = NULL] - y0
        mag <- Mod(y[dimnames = NULL] + y0)
        mag[mag < 1] <- 1
        dif <- dif / Mod(mag)
        testthat::expect_lt(max(abs(Re(dif))), 1e-5, label = sprintf("Operator `%s` on complex", op))
        testthat::expect_lt(max(abs(Im(dif))), 1e-5, label = sprintf("Operator `%s` on complex", op))
    }
    
})

test_that("method: addition", {
    on.exit({
        options("filearray.operator.precision" = NULL)
        clear_cache()
    })
    
    dm <- c(2, 3, 4, 10)
    len <- prod(dm)
    
    x_dbl <- array(rnorm(len), dim = dm)
    x_int <- array(seq_len(len), dim = dm)
    x_lgl <- x_dbl > 0
    
    arr_dbl <- as_filearray(x_dbl, type = "double", partition_size = 2)
    arr_int <- as_filearray(x_int, type = "integer", partition_size = 3)
    arr_lgl <- as_filearray(x_lgl, type = "logical", partition_size = 4)
    
    
    proxy_dbl <- as_filearrayproxy(arr_dbl)
    proxy_int <- as_filearrayproxy(arr_int)
    proxy_lgl <- as_filearrayproxy(arr_lgl)
    
    # sanity
    expect_equal(arr_dbl[dimnames = NULL], x_dbl, tolerance = 1e-5)
    expect_equal(arr_int[dimnames = NULL], x_int)
    expect_equal(arr_lgl[dimnames = NULL], x_lgl)
    
    expect_equal(proxy_dbl[dimnames = NULL], x_dbl, tolerance = 1e-5)
    expect_equal(proxy_int[dimnames = NULL], x_int)
    expect_equal(proxy_lgl[dimnames = NULL], x_lgl)
    
    check_add <- function(e1, e2, d1, d2, type, tolerance = .Machine$double.eps) {
        y <- e1 - e2
        res <- y[dimnames = NULL]
        
        if (!missing(type)) {
            expect_identical(typeof(y), type)
        }
        
        expect_equal(res, d1 - d2, tolerance = tolerance)
    }
    
    options("filearray.operator.precision" = "double")
    
    # filearrayproxy - Double
    check_add(proxy_dbl, proxy_dbl, x_dbl, x_dbl, "double", 1e-5)
    check_add(proxy_dbl, proxy_int, x_dbl, x_int, "double", 1e-5)
    check_add(proxy_dbl, proxy_lgl, x_dbl, x_lgl, "double", 1e-5)
    
    check_add(proxy_dbl, arr_dbl, x_dbl, x_dbl, "double", 1e-5)
    check_add(proxy_dbl, arr_int, x_dbl, x_int, "double", 1e-5)
    check_add(proxy_dbl, arr_lgl, x_dbl, x_lgl, "double", 1e-5)
    
    check_add(proxy_dbl, 1.0, x_dbl, 1.0, "double", 1e-5)
    check_add(proxy_dbl, 1L, x_dbl, 1L, "double", 1e-5)
    check_add(proxy_dbl, TRUE, x_dbl, TRUE, "double", 1e-5)
    
    # filearrayproxy - Int
    check_add(proxy_int, proxy_dbl, x_int, x_dbl, "double", 1e-5)
    check_add(proxy_int, proxy_int, x_int, x_int, "integer", 1e-5)
    check_add(proxy_int, proxy_lgl, x_int, x_lgl, "integer", 1e-5)
    
    check_add(proxy_int, arr_dbl, x_int, x_dbl, "double", 1e-5)
    check_add(proxy_int, arr_int, x_int, x_int, "integer", 1e-5)
    check_add(proxy_int, arr_lgl, x_int, x_lgl, "integer", 1e-5)
    
    check_add(proxy_int, 1.0, x_int, 1.0, "double", 1e-5)
    check_add(proxy_int, 1L, x_int, 1L, "integer", 1e-5)
    check_add(proxy_int, TRUE, x_int, TRUE, "integer", 1e-5)
    
    # filearrayproxy - Logical
    check_add(proxy_lgl, proxy_dbl, x_lgl, x_dbl, "double", 1e-5)
    check_add(proxy_lgl, proxy_int, x_lgl, x_int, "integer", 1e-5)
    check_add(proxy_lgl, proxy_lgl, x_lgl, x_lgl, "integer", 1e-5)
    
    check_add(proxy_lgl, arr_dbl, x_lgl, x_dbl, "double", 1e-5)
    check_add(proxy_lgl, arr_int, x_lgl, x_int, "integer", 1e-5)
    check_add(proxy_lgl, arr_lgl, x_lgl, x_lgl, "integer", 1e-5)
    
    check_add(proxy_lgl, 1.0, x_lgl, 1.0, "double", 1e-5)
    check_add(proxy_lgl, 1L, x_lgl, 1L, "integer", 1e-5)
    check_add(proxy_lgl, TRUE, x_lgl, TRUE, "integer", 1e-5)
    
    # filearray - Double
    check_add(arr_dbl, proxy_dbl, x_dbl, x_dbl, "double", 1e-5)
    check_add(arr_dbl, proxy_int, x_dbl, x_int, "double", 1e-5)
    check_add(arr_dbl, proxy_lgl, x_dbl, x_lgl, "double", 1e-5)
    
    check_add(arr_dbl, arr_dbl, x_dbl, x_dbl, "double", 1e-5)
    check_add(arr_dbl, arr_int, x_dbl, x_int, "double", 1e-5)
    check_add(arr_dbl, arr_lgl, x_dbl, x_lgl, "double", 1e-5)
    
    check_add(arr_dbl, 1.0, x_dbl, 1.0, "double", 1e-5)
    check_add(arr_dbl, 1L, x_dbl, 1L, "double", 1e-5)
    check_add(arr_dbl, TRUE, x_dbl, TRUE, "double", 1e-5)
    
    # filearray - Int
    check_add(arr_int, proxy_dbl, x_int, x_dbl, "double", 1e-5)
    check_add(arr_int, proxy_int, x_int, x_int, "integer", 1e-5)
    check_add(arr_int, proxy_lgl, x_int, x_lgl, "integer", 1e-5)
    
    check_add(arr_int, arr_dbl, x_int, x_dbl, "double", 1e-5)
    check_add(arr_int, arr_int, x_int, x_int, "integer", 1e-5)
    check_add(arr_int, arr_lgl, x_int, x_lgl, "integer", 1e-5)
    
    check_add(arr_int, 1.0, x_int, 1.0, "double", 1e-5)
    check_add(arr_int, 1L, x_int, 1L, "integer", 1e-5)
    check_add(arr_int, TRUE, x_int, TRUE, "integer", 1e-5)
    
    # filearray - Logical
    check_add(arr_lgl, proxy_dbl, x_lgl, x_dbl, "double", 1e-5)
    check_add(arr_lgl, proxy_int, x_lgl, x_int, "integer", 1e-5)
    check_add(arr_lgl, proxy_lgl, x_lgl, x_lgl, "integer", 1e-5)
    
    check_add(arr_lgl, arr_dbl, x_lgl, x_dbl, "double", 1e-5)
    check_add(arr_lgl, arr_int, x_lgl, x_int, "integer", 1e-5)
    check_add(arr_lgl, arr_lgl, x_lgl, x_lgl, "integer", 1e-5)
    
    check_add(arr_lgl, 1.0, x_lgl, 1.0, "double", 1e-5)
    check_add(arr_lgl, 1L, x_lgl, 1L, "integer", 1e-5)
    check_add(arr_lgl, TRUE, x_lgl, TRUE, "integer", 1e-5)
    
    
    # scalar - double
    check_add(1.0, proxy_dbl, 1.0, x_dbl, "double", 1e-5)
    check_add(1.0, proxy_int, 1.0, x_int, "double", 1e-5)
    check_add(1.0, proxy_lgl, 1.0, x_lgl, "double", 1e-5)
    
    check_add(1.0, arr_dbl, 1.0, x_dbl, "double", 1e-5)
    check_add(1.0, arr_int, 1.0, x_int, "double", 1e-5)
    check_add(1.0, arr_lgl, 1.0, x_lgl, "double", 1e-5)
    
    # scalar - int
    check_add(1L, proxy_dbl, 1L, x_dbl, "double", 1e-5)
    check_add(1L, proxy_int, 1L, x_int, "integer", 1e-5)
    check_add(1L, proxy_lgl, 1L, x_lgl, "integer", 1e-5)
    
    check_add(1L, arr_dbl, 1L, x_dbl, "double", 1e-5)
    check_add(1L, arr_int, 1L, x_int, "integer", 1e-5)
    check_add(1L, arr_lgl, 1L, x_lgl, "integer", 1e-5)
    
    # scalar - lgl
    check_add(FALSE, proxy_dbl, FALSE, x_dbl, "double", 1e-5)
    check_add(FALSE, proxy_int, FALSE, x_int, "integer", 1e-5)
    check_add(FALSE, proxy_lgl, FALSE, x_lgl, "integer", 1e-5)
    
    check_add(FALSE, arr_dbl, FALSE, x_dbl, "double", 1e-5)
    check_add(FALSE, arr_int, FALSE, x_int, "integer", 1e-5)
    check_add(FALSE, arr_lgl, FALSE, x_lgl, "integer", 1e-5)
    
    # wrong dimensions
    testthat::expect_error(arr_int - array(0L, c(10, 20, 1)))
    testthat::expect_error(arr_int - 1:10)
    tmp <- arr_int[]
    dm <- dim(arr_int)
    dm <- c(dm[1] * dm[2], dm[-c(1, 2)])
    dim(tmp) <- dm
    tmp_arr <- as_filearray(tmp)
    testthat::expect_error( arr_int - tmp )
    testthat::expect_error( tmp - arr_int )
    testthat::expect_error( arr_int - tmp_arr )
    testthat::expect_error( tmp_arr - arr_int )
    testthat::expect_equal(
        (arr_int - seq_along(arr_int))[dimnames = NULL],
        arr_int[dimnames = NULL] - seq_along(arr_int)
    )
    testthat::expect_equal(
        (seq_along(arr_int) - arr_int)[dimnames = NULL],
        seq_along(arr_int) - arr_int[dimnames = NULL]
    )
    
})

#' @importFrom methods new
#' @importFrom methods signature
#' @importFrom methods setGeneric
#' @importFrom methods setRefClass
#' @importFrom methods setMethod
NULL
HEADER_SIZE <- 1024
FILE_VER <- c( 1L, 1L, 0L )
HEADER_VER <- 1L
RESERVED_HEADERS <- c("endianness", "version", "sexp_type", 
                      "unit_bytes", "partition", "partition_size", 
                      "partition_dim", "header_bytes", "header_version",
                      "dimnames", "content_length")

# The saved files are always little endian
ENDIANNESS <- "little"

max_buffer_size <- local({
    # By default, maximum of 2MB buffer size
    size <- 2097152
    function(v) {
        if (!missing(v)) {
            if (v < 64) {
                stop("Maximum buffer size is too small.")
            }
            v <- 2^ceiling(log2(v))
            if (v > 2^30) {
                stop("Maximum buffer size is too large.")
            }
            size <<- v
        }
        return(size)
    }
})

quiet_warning <- function(..., call. = FALSE) {
    if (!getOption("filearray.quiet", FALSE)) {
        warning(..., '\n\n* To suppress this message, set `options("filearray.quiet" = TRUE)`', call. = call.)
    }
}

get_os <- function() {
    if ("windows" %in% tolower(.Platform$OS.type)) {
        return("windows")
    }
    os <- tolower(R.version$os)
    if (startsWith(os, "darwin")) {
        return("darwin")
    }
    if (startsWith(os, "linux")) {
        return("linux")
    }
    if (startsWith(os, "solaris")) {
        return("solaris")
    }
    if (startsWith(os, "win")) {
        return("windows")
    }
    return("unknown")
}


deparse1 <- function(expr, collapse = " ") {
    paste(deparse(expr), collapse = collapse)
}

temp_dir <- function(check = FALSE) {
    re <- file.path(getOption("filearray.temporary.path", tempdir()), "_filearray_tempdir")
    if (check && !dir.exists(re)) {
        dir.create(re, showWarnings = FALSE, recursive = TRUE)
    }
    re
}

temp_path <- function(pattern = "tmpfilearray", fileext = ".farr", check = FALSE) {
    tempfile(pattern = pattern, tmpdir = temp_dir(check = check), fileext = fileext)
}

clear_cache <- function() {
    tdir <- temp_dir()
    if (dir.exists(tdir)) {
        unlink(tdir, recursive = TRUE)
    }
}

#' @title Definition of file array
#' @name FileArray-class
#' @description \code{S4} class definition of \code{FileArray}. Please 
#' use \code{\link{filearray_create}} and \code{\link{filearray_load}}
#' to create instances.
#' @section Public Methods:
#' \describe{
#' \item{\code{get_header(key, default = NULL)}}{Get header information; returns \code{default} if \code{key} is missing}
#' \item{\code{set_header(key, value)}}{Set header information; the extra headers will be stored in meta file. Please do not store large headers as they will be loaded into memory frequently.}
#' \item{\code{can_write()}}{Whether the array data can be altered}
#' \item{\code{create(filebase, dimension, type = "double", partition_size = 1)}}{Create a file array instance}
#' \item{\code{delete(force = FALSE)}}{Remove array from local file system and reset}
#' \item{\code{dimension()}}{Get dimension vector}
#' \item{\code{dimnames(v)}}{Set/get dimension names}
#' \item{\code{element_size()}}{Internal storage: bytes per element}
#' \item{\code{fill_partition(part, value)}}{Fill a partition with given scalar}
#' \item{\code{get_partition(part, reshape = NULL)}}{Get partition data, and reshape (if not null) to desired dimension}
#' \item{\code{expand(n)}}{Expand array along the last margin; returns true if expanded; if the \code{dimnames} have been assigned prior to expansion, the last dimension names will be filled with \code{NA}}
#' \item{\code{initialize_partition()}}{Make sure a partition file exists; if not, create one and fill with \code{NA}s or 0 (\code{type='raw'})}
#' \item{\code{load(filebase, mode = c("readwrite", "readonly"))}}{Load file array from existing directory}
#' \item{\code{partition_path(part)}}{Get partition file path}
#' \item{\code{partition_size()}}{Get partition size; see \code{\link{filearray}}}
#' \item{\code{set_partition(part, value, ..., strict = TRUE)}}{Set partition value}
#' \item{\code{sexp_type()}}{Get data \code{SEXP} type; see R internal manuals}
#' \item{\code{show()}}{Print information}
#' \item{\code{type()}}{Get data type}
#' \item{\code{valid()}}{Check if the array is valid.}
#' }
#' @seealso \code{\link{filearray}}
#' @exportClass FileArray
NULL

initialize_filearray <- function(path, dimension, partition_size, type) {
    stopifnot(!dir.exists(path))
    dir.create(path, showWarnings = FALSE)
    meta <- file.path(path, "meta")
    conn <- file(meta, "w+b")
    on.exit({
        try({close(conn)}, silent = TRUE)
    })
    size <- get_elem_size(type = type)
    write_header(conn, partition = partition_size, dimension = dimension, 
                 type = type, size = size)
    close(conn)
}

load_meta <- function(path) {
    meta <- file.path(path, "meta")
    stopifnot(file.exists(meta))
    header <- validate_header(meta)
    
    if (header$content_length > 0) {
        # load dimnames
        fid <- file(meta, "rb")
        on.exit({ close(fid) })
        seek(con = fid, where = HEADER_SIZE, origin = "start", rw = "read")
        v <- readBin(con = fid, what = "raw", size = 1L, 
                     n = header$content_length)
        conn <- rawConnection(v, open = "rb")
        
        # .self$.header[c("dimnames", "symlinks")]
        extra_header <- readRDS(conn)
        close(conn)
        if (is.list(extra_header)) {
            extra_header_names <- names(extra_header)
            if ("__header_version__" %in% extra_header_names) {
                extra_header_names <- sub("^__", "", extra_header_names)
                extra_header_names <- sub("__$", "", extra_header_names)
                extra_header_names <- extra_header_names[!extra_header_names %in% RESERVED_HEADERS]
                extra_header_names <- c(extra_header_names, "dimnames", "header_version")
                for (nm in extra_header_names) {
                    header[[nm]] <- extra_header[[sprintf("__%s__", nm)]]
                }
            } else {
                # old format
                header$dimnames <- extra_header
                header$header_version <- 0
            }
        } else {
            header$header_version <- 0
        }
    } else {
        header$header_version <- 0
    }
    header
}

get_elem_size <- function(type) {
    switch(
        type,
        double = 8L,
        float = 4L,
        integer = 4L,
        logical = 1L,
        raw = 1L,
        complex = 8L,
        stop("Unknown data type: ", type)
    )
}

FileArray <- setRefClass(
    "FileArray",
    fields = list(
        .mode = "character",
        .sep = "character",
        .filebase = "character",
        .header = "list",
        .partition_info = "ANY",
        .na = "ANY",
        .valid = "logical"
    ),
    methods = list(
        initialize = function() {
            .self$.mode <- "readonly"
            .self$.sep <- .Platform$file.sep
            .self$.filebase <- "."
            .self$.header <- list()
            .self$.partition_info <- NULL
            .self$.na <- NA_real_
            .self$.valid <- FALSE
        },
        create = function(filebase, dimension, type = "double",
                          partition_size = 1) {
            dir <- dirname(filebase)
            if (!dir.exists(dir)) {
                stop("Path not exists: ", dir)
            }
            partition_size <- as.integer(partition_size)
            if (is.na(partition_size) || partition_size < 1) {
                partition_size <- 1
            }
            
            if (dir.exists(filebase)) {
                stop("To create a file array, the path must be empty.")
            }
            stopifnot(length(dimension) >= 2)
            
            # create
            initialize_filearray(
                path = filebase,
                dimension = dimension,
                partition_size = partition_size,
                type = type
            )
            .self$load(filebase, "readwrite")
        },
        dimension = function() {
            .self$.header$partition_dim
        },
        dimnames = function(v) {
            if (!missing(v)) {
                dim <- .self$.header$partition_dim
                stopifnot(is.list(v) || length(v) <= length(dim))
                for (ii in seq_along(v)) {
                    if (length(v[[ii]]) && length(v[[ii]]) != dim[[ii]]) {
                        stop("Dimension ", ii, " length mismatch")
                    }
                }
                nms <- names(v)
                if (!is.null(nms) && length(nms) < length(dim)) {
                    nms <- c(nms, rep("", length(dim) - length(nms)))
                }
                v <- structure(lapply(seq_along(dim), function(ii) {
                    if (ii > length(v)) { return(NULL) }
                    v[[ii]]
                }), names = nms)
                # set dimnames
                meta <- file.path(.self$.filebase, "meta")
                .self$.header$dimnames <- v
                .self$.save_header()
                .self$.header$dimnames
            }
            .self$.header$dimnames
        },
        type = function() {
            sexp_to_type(.self$.header$sexp_type)
        },
        sexp_type = function() {
            .self$.header$sexp_type
        },
        element_size = function() {
            .self$.header$unit_bytes
        },
        partition_size = function() {
            .self$.header$partition
        },
        .save_header = function() {
            if (.self$.mode != "readwrite") {
                quiet_warning("Cannot save extra headers because the array has no write access.")
                return(invisible())
            }
            # get header version
            keys <- names(.self$.header)
            keys <- keys[!keys %in% RESERVED_HEADERS | keys %in% c("dimnames")]
            extra_header <- .self$.header[keys]
            meta <- file.path(.self$.filebase, "meta")
            set_meta_content(meta, extra_header)
            if (!identical(as.integer(.self$.header$header_version), HEADER_VER)) {
                # old header
                .self$.header <- load_meta(.self$.filebase)
            }
            return(invisible())
        },
        get_header = function(key, default = NULL) {
            if (key %in% names(.self$.header)) {
                return(.self$.header[[key]])
            }
            return(default)
        },
        set_header = function(key, value, save = TRUE) {
            force(value)
            if (key %in% RESERVED_HEADERS) {
                stop("Key `", key, "` is preserved and should be read-only or altered via other methods.")
            }
            .self$.header[[key]] <- value
            if ( save ) {
                .self$.save_header()
            }
            invisible(value)
        },
        header_signature = function(include_path = TRUE) {
            header_sig <- digest::digest(.self$.header, algo = "sha256")
            if ( include_path ) {
                path <- normalizePath(.self$.filebase)
                header_sig <- digest::digest(c(
                    header_sig, path
                ), algo = "sha256")
            }
            header_sig
        },
        load = function(filebase, mode = c("readwrite", "readonly")) {
            mode <- match.arg(mode)
            
            filebase <- normalizePath(filebase, mustWork = TRUE)
            if (endsWith(filebase, "/|\\\\")) {
                filebase <- gsub("[/\\\\]+$", "", x = filebase)
            }
            
            # set members
            .self$.mode <- mode
            .self$.sep <- .Platform$file.sep
            .self$.filebase <- filebase
            .self$.valid <- TRUE
            
            # meta
            header <- load_meta(filebase)
            .self$.header <- header
            
            
            .self$.na <- switch(
                .self$type(),
                double = NA_real_,
                integer = NA_integer_,
                logical = FALSE,
                raw = as.raw(0),
                complex = NA_complex_,
                float = NA_real_,
                stop("Unknown data type: ", .self$type())
            )
            
            # load partition information
            files <- list.files(filebase, pattern = "[0-9]+.farr", 
                                recursive = FALSE, all.files = FALSE, 
                                full.names = FALSE, ignore.case = TRUE, 
                                include.dirs = FALSE)
            dimension <- .self$dimension()
            partition_size <- .self$partition_size()
            margin <- dimension[[length(dimension)]]
            nparts <- ceiling(margin / partition_size)
            if (length(files)) {
                partition_info <- t(sapply(files, function(f) {
                    header <- validate_header(file = file.path(filebase, f))
                    c(header$partition, header$sexp_type, header$partition_dim)
                }))
                sel <- partition_info[, 2] != .self$sexp_type()
                if (any(sel)) {
                    stop(
                        sprintf(
                            "Expected data type: [%s], but the following partition files disagree [%s]",
                            .self$type(),
                            paste(partition_info[sel, 1], collapse = ", ")
                        )
                    )
                }
                expected_part <- as.integer(gsub("[^0-9]+", "", files)) + 1
                if (!all(expected_part == partition_info[, 1])) {
                    # check if this array is a bound array
                    bind_info <- .self$.header$filearray_bind
                    if (is.list(bind_info) && isTRUE(bind_info$is_bound)) {
                        partition_info[, 1] <- seq_along(partition_info[, 1])
                        if (!all(expected_part == partition_info[, 1])) {
                            stop("Partition filenames mismatch with partition headers.")
                        }
                        if (bind_info$symlink &&
                            .self$.mode != "readonly") {
                            quiet_warning(
                                "Partition filenames mismatch with partition headers. This happens when the array partitions are symlinked from other arrays. For safety reasons, switched to read-only mode."
                            )
                            .self$.mode <- "readonly"
                        }
                    } else {
                        stop("Partition filenames mismatch with partition headers.")
                    }
                    
                }
                
                nparts <- max(c(nparts, partition_info[, 1]))
                pinfo <- cbind(seq_len(nparts), partition_size)
                pinfo[partition_info[, 1], 2] <- partition_info[, ncol(partition_info)]
                cs <- cumsum(pinfo[, 2])
                idx <- max(which(cs <= margin))
                if (cs[idx] < margin) { idx <- idx + 1 }
                pinfo <- pinfo[seq_len(idx), , drop = FALSE]
            } else {
                pinfo <- cbind(seq_len(nparts), partition_size)
            }
            # need to make sure cumpart-size is correct
            cpl <- cumsum(pinfo[, 2])
            nr <- sum(cpl <= margin)
            if (cpl[[nr]] < margin) {
                nr <- nr + 1
            }
            cpl <- cpl[seq_len(nr)]
            cpl[[nr]] <- margin
            pinfo <- cbind(pinfo[seq_len(nr), , drop = FALSE], cpl)
            .self$.partition_info <- pinfo
            return(.self)
        },
        show = function() {
            if (.self$valid()) {
                cat("Reference class object of class \"FileArray\"\n")
                cat("Mode:", .self$.mode, "\n")
                tryCatch({
                    cat("Dimension:", paste(.self$dimension(), collapse = "x"), "\n")
                    cat("Partition count:", nrow(.self$.partition_info), "\n")
                    cat("Partition size:", .self$partition_size(), "\n")
                    cat("Storage type: ", .self$type(), " (internal size: ", get_elem_size(.self$type()), ")\n", sep = "")
                    
                }, error = function(e) {
                    quiet_warning("Partition information is unavailable: might be broken or improperly set.", immediate. = FALSE)
                })
                cat("Location:", .self$.filebase, "\n")
            } else {
                if (.self$.valid) {
                    cat("The FileArray is invalid (it has not been initialized).")
                } else {
                    cat("The FileArray is invalid (unable to locate the array in the file system).")
                }
            }
            
        },
        partition_path = function(part) {
            part <- as.integer(part)
            file.path(.self$.filebase, sprintf("%d.farr", part - 1))
        },
        set_partition = function(part, value, ..., strict = TRUE) {
            if (isTRUE(.self$.mode == "readonly")) {
                stop("File array is read-only")
            }
            part <- as.integer(part)
            path <- .self$partition_path(part)
            dim <- .self$dimension()
            lastm <- dim[[length(dim)]]
            if ( is.na(part) || part < 1 || part > lastm ) {
                if ( strict ) {
                    stop("Invalid partition: ", part)
                } else {
                    return(FALSE)
                }
            }
            
            arglen <- ...length()
            dim[[length(dim)]] <- .self$partition_size()
            if ( arglen > 1 ) {
                if (file.exists(path)) {
                    x <- load_partition(path, dim)
                } else {
                    x <- array(.self$.na, dim = dim)
                }
                x[...] <- value
                value <- x
            }
            
            # Edit start
            end <- .self$.partition_info[part, 3]
            if (part == 1) {
                start <- 1
            } else {
                start <- .self$.partition_info[part - 1, 3] + 1
            }
            
            tmp <- rep("", length(dim))
            tmp[[length(tmp)]] <- "seq.int(start, end)"
            expr <- sprintf(".self[%s] <- value", paste(tmp, collapse = ","))
            
            eval(parse(text = expr))
            # Edit end
            # Old implementation
            # write_partition(file = path, partition = part, dimension = dim, value = value, type = .self$type())
            
            return(TRUE)
        },
        get_partition = function(part, reshape = NULL) {
            if (is.null(reshape)) {
                reshape <- .self$dimension()
                reshape[[length(reshape)]] <- .self$partition_size()
            }
            return(load_partition(.self$partition_path(part), dim = reshape))
        },
        expand = function(n) {
            if (isTRUE(.self$.mode == "readonly")) {
                stop("File array is read-only")
            }
            
            # get the last margin size
            dim <- .self$dimension()
            ndims <- length(dim)
            last_dim <- dim[[ndims]]
            if (last_dim >= n) {
                return(FALSE)
            }
            dnames <- .self$.header$dimnames
            if (is.list(dnames) && length(dnames) >= ndims) {
                ldn <- tryCatch({
                    ldn <- dnames[[ndims]]
                    c(ldn, rep(NA, n - length(ldn)))
                }, error = function(e) {
                    NULL
                })
                dnames[[ndims]] <- ldn
            }
            meta <- file.path(.self$.filebase, "meta")
            fid <- file(meta, "r+b")
            on.exit({
                try({close(fid)}, silent = TRUE)
            }, add = TRUE)
            
            
            dim[[ndims]] <- n
            type <- sexp_to_type(.self$.header$sexp_type)
            write_header(fid = fid, partition = .self$.header$partition, 
                         dimension = dim, type = type, 
                         size = get_elem_size(type))
            close(fid)
            
            .self$load(.self$.filebase, mode = .self$.mode)
            .self$dimnames(dnames)
            return(TRUE)
            
        },
        fill_partition = function(part, value) {
            if (isTRUE(.self$.mode == "readonly")) {
                stop("File array is read-only")
            }
            if (!.self$valid()) {
                stop("Invalid file array")
            }
            part <- as.integer(part)
            if (length(part) != 1) {
                stop("`fill_partition` only allows one partition at a time")
            }
            if (is.na(part) || part <= 0) {
                stop("NA or non-positive partition are invalid")
            }
            if (length(value) > 1) {
                quiet_warning("`fill_partition` value length coerced to first value")
            }
            
            value <- value[[1]]
            
            type <- .self$type()
            switch(
                type,
                "complex" = {
                    value <- cplxToReal2(as.complex(value))
                },
                "float" = {
                    value <- realToFloat2(as.double(value))
                },
                {
                    storage.mode(value) <- type
                }
            )
            size <- get_elem_size(type)
            file <- .self$partition_path(part)
            fid <- file(file, "wb")
            on.exit({
                close(fid)
            })
            
            if ( part <= nrow(.self$.partition_info)) {
                partition_size <- .self$.partition_info[part, 2]
            } else {
                partition_size <- .self$partition_size()
            }
            dimension <- .self$dimension()
            dimension[[length(dimension)]] <- partition_size
            write_header(
                fid = fid,
                partition = part,
                dimension = dimension,
                type = type,
                size = size
            )
            seek(con = fid, where = HEADER_SIZE, rw = "write")
            part_len <- prod(dimension)
            buffer_len <- get_buffer_size() / size
            if (buffer_len > part_len) {
                buffer_len <- part_len
            }
            buf <- writeBin(con = raw(), object = rep(value, buffer_len), 
                           size = size, endian = ENDIANNESS)
            nloop <- floor(part_len / buffer_len)
            replicate(nloop, {
                writeBin(con = fid, object = buf)
                NULL
            })
            rest <- part_len - buffer_len * nloop
            if ( rest > 0 ) {
                writeBin(con = fid, object = buf[seq_len(rest * size)])
            }
            seek(con = fid, where = HEADER_SIZE - 8L, rw = "write")
            writeBin(con = fid, object = part_len, 
                     size = 8L, endian = ENDIANNESS)
            seek(con = fid, where = 0, rw = "write")
            invisible()
        },
        initialize_partition = function(parts) {
            if (!.self$valid()) {
                stop("Invalid file array")
            }
            if (isTRUE(.self$.mode == "readonly")) {
                .self$.mode <- "readwrite"
                on.exit({
                    .self$.mode <- "readonly"
                })
            }
            if (missing(parts)) {
                parts <- .self$.partition_info[, 1]
            }
            parts <- parts[!is.na(parts)]
            for (part in parts) {
                file <- .self$partition_path(part)
                if (!file.exists(file)) {
                    .self$fill_partition(part, NA)
                }
            }
        },
        can_write = function() {
            if (isTRUE(.self$.mode == "readonly")) {
                return(FALSE)
            }
            return(TRUE)
        },
        delete = function(force = FALSE) {
            if (!.self$valid()) {
                .self$.valid <- FALSE
                return(invisible())
            }
            if (isTRUE(.self$.mode == "readonly")) {
                if (force) {
                    quiet_warning("File array is read-only, but deleted with `force=TRUE`")
                } else {
                    quiet_warning("File array is read-only.")
                }
            }
            filebase <- .self$.filebase
            if (dir.exists(filebase)) {
                unlink(filebase, recursive = TRUE, force = force)
            }
            .self$.valid <- FALSE
            invisible()
        },
        valid = function() {
            return(.self$.valid && dir.exists(.self$.filebase))
        },
        collapse = function(
            keep, method = c("mean", "sum"),
            transform = c("asis", "10log10", "square", "sqrt", "normalize"), 
            na.rm = FALSE
        ) {
            method <- match.arg(method)
            transform <- match.arg(transform)
            keep <- as.integer(keep)
            dim <- .self$dimension()
            if (any(is.na(keep)) || !all(keep %in% seq_along(dim))) {
                stop("`keep` must be valid margin numbers")
            }
            filebase <- paste0(.self$.filebase, .self$.sep)
            
            dim1 <- dim
            transform1 <- which(c("asis", "10log10", "square", "sqrt", "normalize") == transform)
            if ( method == "sum" ) {
                scale <- 1
            } else {
                scale <- 1 / prod(dim[-keep])
            }
            # make sure last margin is the last one
            dim <- .self$dimension()
            ndims <- length(dim)
            perm <- FALSE
            keep1 <- keep
            nkeeps <- length(keep)
            if (nkeeps > 1 && 
               ndims %in% keep && 
               keep[[nkeeps]] != ndims)
            {
                keep_ldidx <- which(keep == ndims)
                
                keep_ord <- seq_along(keep)
                keep_ord <- c(
                    keep_ord[-seq.int(keep_ldidx, nkeeps)], 
                    nkeeps,
                    keep_ord[-seq.int(1, keep_ldidx)] - 1
                )
                
                keep1 <- c(keep[-keep_ldidx], keep[keep_ldidx])
                perm <- TRUE
            }
            
            
            expr <- quote(FARR_collapse(
                filebase = filebase,
                dim = dim1,
                keep = keep1,
                cum_part = .self$.partition_info[, 3],
                remove_na = na.rm,
                method = transform1,
                scale = scale
            ))
            if (.self$type() == "complex") {
                expr[[1]] <- quote(FARR_collapse_complex)
            } else {
                expr[["array_type"]] <- .self$sexp_type()
            }
                
            re <- eval(expr)
            
            dnames <- .self$dimnames()
            if (length(dnames) == length(dim)) {
                dnames <- dnames[keep]
                if (length(keep) == 1) {
                    re <- structure(re, names = dnames[[1]])
                } else {
                    re <- structure(re, dimnames = dnames)
                }
            }
            
            if (perm) {
                re <- aperm(re, keep_ord)
            }
            
            return(re)
            
        }
    )
)


#' @rdname filearray
#' @export
as_filearray <- function(x, ...) {
    UseMethod("as_filearray")
}

#' @export
as_filearray.default <- function(x, filebase = NULL, type = NULL, dimension = dim(x), ...) {
    if (!length(filebase)) {
        filebase <- temp_path(check = TRUE)
        while (file.exists(filebase)) {
            filebase <- temp_path(check = FALSE)
        }
    }
    x <- as.array(x)
    
    if (length(type) == 1) {
        expected_type <- type
    } else {
        expected_type <- typeof(x)
        if (expected_type %in% c("double", "float")) {
            expected_type <- getOption("filearray.operator.precision", "double")
        }
    }
    
    
    re <- filearray_create(filebase = filebase, dimension = dimension, type = expected_type, ...)
    re[] <- x
    
    dimnames(re) <- dimnames(x)
    return(re)
}

#' @export
as_filearray.character <- function(x, mode = c("readonly", "readwrite"), ...) {
    mode <- match.arg(mode)
    filearray_load(filebase = x, mode = mode)
}

#' @export
as_filearray.FileArray <- function(x, ...) {
    x
}

#' @export
as_filearray.FileArrayProxy <- function(x, ...) {
    x
}

#' @export
as_filearray.RAVEFileArray <- function(x, ...) {
    # x <- ravepipeline::RAVEFileArray$new(filearray::as_filearray(array(1:16, c(4,4))))
    if (inherits(x, c("FileArray", "FileArrayProxy"))) {
        return(x)
    }
    if (inherits(x$`@impl`, c("FileArray", "FileArrayProxy"))) {
        return(x$`@impl`)
    }
    stop("`x` is not a valid `RAVEFileArray` object.")
}

test_that("Subset with dimnames", {
    x <- filearray_create(tempfile(), dimension = c(3, 4, 5, 6))
    on.exit({
        x$delete()
    })
    y <- array(as.double(1:240), c(3, 4, 5, 6))
    x[] <- y
    expect_identical(dimnames(x[]), dimnames(y))
    
    dnames1 <- list(A = 1:3,
                    B = 1:4,
                    NULL,
                    D = 1:6)
    dnames2 <- list(A = 1:3)
    dnames3 <- list(
        A = 1:3,
        B = 1:4,
        C = NULL,
        D = NULL
    )
    dnames4 <- list(
        A = 1:3,
        B = 1:4,
        C = 1:5,
        D = 1:6
    )
    
    expect_error({
        dimnames(x) <- list(A = 1:4)
    })
    
    
    dimnames(y) <- dnames1
    dimnames(x) <- dnames1
    
    expect_identical(dimnames(x[]), dimnames(y))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4]), dimnames(y[1, 1:2, 2:3, 1:4]))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4, drop = FALSE]), dimnames(y[1, 1:2, 2:3, 1:4, drop =
                                                                                 FALSE]))
    expect_identical(names(x[2, 1, 1, 1]), names(y[2, 1, 1, 1]))
    
    dimnames(y) <- dnames2
    dimnames(x) <- dnames2
    
    expect_identical(dimnames(x[]), dimnames(y))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4]), dimnames(y[1, 1:2, 2:3, 1:4]))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4, drop = FALSE]), dimnames(y[1, 1:2, 2:3, 1:4, drop =
                                                                                 FALSE]))
    expect_identical(names(x[2, 1, 1, 1]), names(y[2, 1, 1, 1]))
    
    dimnames(y) <- dnames3
    dimnames(x) <- dnames3
    
    expect_identical(dimnames(x[]), dimnames(y))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4]), dimnames(y[1, 1:2, 2:3, 1:4]))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4, drop = FALSE]), dimnames(y[1, 1:2, 2:3, 1:4, drop =
                                                                                 FALSE]))
    expect_identical(names(x[2, 1, 1, 1]), names(y[2, 1, 1, 1]))
    
    
    dimnames(y) <- dnames4
    dimnames(x) <- dnames4
    
    expect_identical(dimnames(x[]), dimnames(y))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4]), dimnames(y[1, 1:2, 2:3, 1:4]))
    expect_identical(dimnames(x[1, 1:2, 2:3, 1:4, drop = FALSE]), dimnames(y[1, 1:2, 2:3, 1:4, drop =
                                                                                 FALSE]))
    expect_identical(names(x[2, 1, 1, 1]), names(y[2, 1, 1, 1]))
    
    
    # expand
    dimnames(x) <- dnames4
    x$expand(10)
    expect_equal(dimnames(x)$D, c(dnames4$D, rep(NA_integer_, 4)))
    
})

test_that("R: FileArray-class", {
    set.seed(2)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    dim <- 3:5
    x <- filearray_create(file, dim, partition_size = 3, initialize = FALSE)
    
    expect_equal(x$dimension(), dim)
    expect_true(x$can_write())
    expect_equal(x$type(), "double")
    expect_equal(x$element_size(), get_elem_size(x$type()))
    expect_error(x$fill_partition(-1, 1))
    expect_error(x$fill_partition(NA, 1))
    expect_error(x$fill_partition(0, 1))
    expect_warning(x$fill_partition(1, 1:10))
    x$fill_partition(2, 2)
    expect_equal(x[[60]], 2)
    
    unlink(file, recursive = TRUE)
    x <- filearray_create(file, dim, partition_size = 1, initialize = FALSE)
    expect_equal(x$.mode, "readwrite")
    x$fill_partition(1, 2)
    
    x <- filearray_load(file, mode = "readonly")
    expect_equal(x$.mode, "readonly")
    expect_error(x$fill_partition(1, 1))
    
    x$initialize_partition(1:2)
    expect_equal(x[1, 1, 1:2], c(2, NA))
    expect_equal(file.exists(x$partition_path(1:3)), c(TRUE, TRUE, FALSE))
    expect_equal(x$.mode, "readonly")
    
    expect_error(x[] <- 1:60)
    
    unlink(file, recursive = TRUE)
    # not removed by $delete
    expect_true(x$.valid)
    expect_false(x$valid())
    expect_error(x[])
    expect_error(as.array(x))
    expect_error(x[[1]])
    expect_error(mapreduce(x, I))
    expect_error(typeof(x))
    expect_error(max(x))
    
    x$delete()
    expect_false(x$.valid)
    
    unlink(file, recursive = TRUE)
})


test_that("R: S3 methods", {
    set.seed(3)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    dim <- 3:5
    x <- filearray_create(file, dim, partition_size = 3, initialize = FALSE)
    expect_equal(file.exists(x$partition_path(1:3)), c(FALSE, FALSE, FALSE))
    x[, , 4] <- 1:12
    expect_equal(as.integer(x[, , 4]), 1:12)
    expect_equal(file.exists(x$partition_path(1:3)), c(FALSE, TRUE, FALSE))
    
    expect_equal(typeof(x), "double")
    expect_equal(range(x, na.rm = TRUE), as.double(c(1, 12)))
    expect_equal(file.exists(x$partition_path(1:3)), c(FALSE, TRUE, FALSE))
    expect_equal(range(x), as.double(c(NA, NA)))
    expect_equal(file.exists(x$partition_path(1:3)), c(TRUE, TRUE, FALSE))
    
    unlink(file, recursive = TRUE)
})

library(testthat)


test_that("map with proxy", {
    set.seed(10)
    
    # A large array example
    x1 <- filearray_create(
        temp_path(check = TRUE),
        dimension = c(28, 100, 3, 4),
        initialize = FALSE,
        partition_size = 3L
    )
    x1[] <- rnorm(33600)
    x2 <- x1 + 1
    x3 <- x1 + x2
    x4 <- as_filearray(matrix(1:12, nrow = 4))
    
    x <- list(x1, x2, x3, x4)
    # check common input size
    
    bc <- 12
    re <- fmap2(x, function(input) {
        testthat::expect_length(input, 4)
        testthat::expect_length(input[[1]], length(x1) / bc)
        testthat::expect_length(input[[2]], length(x2) / bc)
        testthat::expect_length(input[[3]], length(x3) / bc)
        testthat::expect_length(input[[4]], length(x4) / bc)
        testthat::expect_equal(input[[1]] + 1, input[[2]])
        testthat::expect_equal(input[[3]], input[[2]] + input[[1]])
        
        sum(input[[4]]) + sum(input[[3]] - input[[2]] - input[[1]])
    }, .buffer_count = bc)
    expect_equal(re, colSums(matrix(x4[], ncol = bc)))
    
    
    bc <- 4
    re <- fmap2(x, function(input) {
        testthat::expect_length(input, 4)
        testthat::expect_length(input[[1]], length(x1) / bc)
        testthat::expect_length(input[[2]], length(x2) / bc)
        testthat::expect_length(input[[3]], length(x3) / bc)
        testthat::expect_length(input[[4]], length(x4) / bc)
        testthat::expect_equal(input[[1]] + 1, input[[2]])
        testthat::expect_equal(input[[3]], input[[2]] + input[[1]])
        
        sum(input[[4]]) + sum(input[[3]] - input[[2]] - input[[1]])
    }, .buffer_count = bc)
    expect_equal(re, colSums(matrix(x4[], ncol = bc)))
    
    bc <- 1
    re <- fmap2(x, function(input) {
        testthat::expect_length(input, 4)
        testthat::expect_length(input[[1]], length(x1) / bc)
        testthat::expect_length(input[[2]], length(x2) / bc)
        testthat::expect_length(input[[3]], length(x3) / bc)
        testthat::expect_length(input[[4]], length(x4) / bc)
        testthat::expect_equal(input[[1]] + 1, input[[2]])
        testthat::expect_equal(input[[3]], input[[2]] + input[[1]])
        
        sum(input[[4]]) + sum(input[[3]] - input[[2]] - input[[1]])
    }, .buffer_count = bc)
    expect_equal(re, colSums(matrix(x4[], ncol = bc)))
    
    
    # check fmap
    bc <- 12
    
    y <- filearray_create(temp_path(), dimension = c(12, 1))
    fmap(x, function(input) {
        testthat::expect_length(input, 4)
        testthat::expect_length(input[[1]], length(x1) / bc)
        testthat::expect_length(input[[2]], length(x2) / bc)
        testthat::expect_length(input[[3]], length(x3) / bc)
        testthat::expect_length(input[[4]], length(x4) / bc)
        testthat::expect_equal(input[[1]] + 1, input[[2]])
        testthat::expect_equal(input[[3]], input[[2]] + input[[1]])
        
        input[[4]] + sum(input[[3]] - input[[2]] - input[[1]])
    }, .buffer_count = bc, .y = y)
    expect_equal(as.vector(y[]), as.vector(x4[]))
    
    clear_cache()
})


test_that("map filearrays", {
    # A large array example
    x <- filearray_create(
        temp_path(check = TRUE),
        dimension = c(28, 100, 301, 4),
        initialize = FALSE,
        partition_size = 3L
    )
    dnames <- list(
        Trial = sample(c("A", "B"), 28, replace = TRUE),
        Marker = 1:100,
        Time = seq(-1, 2, 0.01),
        Location = 1:4
    )
    dimnames(x) <- dnames
    
    expect_equal(dimnames(x), dnames)
    
    y <- array(rnorm(length(x)), dim(x))
    x[] <- y
    
    output <- filearray_create(
        temp_path(check = TRUE),
        dimension = dim(x),
        initialize = FALSE,
        partition_size = 4L
    )
    
    f <- function(input) {
        # get locational data
        if (is.list(input)) {
            location_data <- input[[1]]
        } else {
            location_data <- input
        }
        
        dim(location_data) <- c(28, 100, 301)
        
        # collapse over first 50 time points for
        # each trial, and marker
        baseline <- apply(location_data[, , 1:50], c(1, 2), mean)
        
        # calibrate
        calibrated <- sweep(
            location_data,
            c(1, 2),
            baseline,
            FUN = function(data, bl) {
                (data / bl - 1) * 100
            }
        )
        return(calibrated)
    }
    
    fmap(x, f, .y = output, .buffer_count = 4)
    
    b <- apply(y, 4, f)
    dim(b) <- dim(y)
    
    expect_equal(output[], b)
    
    d <- fmap2(x, f, .buffer_count = 4, .simplify = TRUE)
    expect_equal(d, b)
    
    x$delete()
    output$delete()
    clear_cache()
})


test_that("fwhich", {
    x <- filearray_create(
        temp_path(check = TRUE),
        dimension = c(28, 100, 301, 4),
        initialize = FALSE,
        partition_size = 3L,
        type = "complex"
    )
    dnames <- list(
        Trial = sample(c("A", "B"), 28, replace = TRUE),
        Marker = 1:100,
        Time = seq(-1, 2, 0.01),
        Location = 1:4
    )
    dimnames(x) <- dnames
    
    y <- array(rnorm(length(x)), dim(x))
    x[] <- y
    x[1, c(1, 3), 301, 4] <- c(80 + 10i, 80 + 10i)
    y <- x[]
    
    idx <- fwhich(x,
                  val = 80 + 10i,
                  arr.ind = FALSE,
                  ret.values = FALSE)
    expect_equal(idx, which(y == (80 + 10i)))
    
    idx <- fwhich(x,
                  val = 80 + 10i,
                  arr.ind = FALSE,
                  ret.values = TRUE)
    expect_equal(attr(idx, "values"), rep(80 + 10i, length(idx)))
    
    idx1 <- fwhich(x,
                   val = 80 + 10i,
                   arr.ind = TRUE,
                   ret.values = TRUE)
    idx2 <- fwhich(y,
                   val = 80 + 10i,
                   arr.ind = TRUE,
                   ret.values = TRUE)
    expect_equal(idx1, idx2)
    
    # val is a function
    impl <- function(z) {
        Re(z) > 4
    }
    idx1 <- fwhich(x, impl, arr.ind = FALSE, ret.values = TRUE)
    idx2 <- fwhich(x[], impl, arr.ind = FALSE, ret.values = TRUE)
    expect_equal(idx1, idx2)
    
    impl <- function(z) {
        Im(z) < -100
    }
    idx1 <- fwhich(x, impl, arr.ind = TRUE, ret.values = TRUE)
    idx2 <- fwhich(x[], impl, arr.ind = TRUE, ret.values = TRUE)
    expect_equal(idx1, idx2)
    
    x$delete()
})

test_that("subset-assign filearray-proxy", {
    
    # normal indexing
    x0 <- array(1:120, dim = c(10, 12))
    dimnames(x0) <- list(A = 1:10, B = 1:12)
    x <- as_filearray(x0)
    y <- x + 1L
    y[, 1] <- rep(0, 10)
    y[1, ] <- rep(0, 12)
    z <- y + y
    
    idx1 <- c(3, 4, 5, 2, 1)
    idx2 <- idx1 * 2
    
    expect_equal(x[idx1, idx2], x0[idx1, idx2])
    expect_equal(y[1, 1:12, dimnames = NULL], rep(0L, 12))
    expect_equal(y[1:10, 1, dimnames = NULL], rep(0L, 10))
    expect_equal(y[2:10, 2:10], x0[2:10, 2:10] + 1L)
    expect_equal(z[, idx2], 2 * y[, idx2])
    
    expect_equal(y[2:10, 2:12], x[2:10, 2:12] + 1)
    expect_equal(z[], 2 * y[])
    
    clear_cache()
    
    # Array indexing
    idx1 <- c(3, 4, 5, 2, 1)
    idx2 <- idx1 * 2
    
    i <- array(FALSE, dim(x))
    i[idx1, idx2] <- TRUE
    j <- as_filearray(i)
    expect_equal(typeof(j), "logical")
    
    x0 <- array(1:120, dim = c(10, 12))
    dimnames(x0) <- list(A = 1:10, B = 1:12)
    x <- as_filearray(x0)
    y <- x + 1L
    y[j, lazy = TRUE] <- 0
    z <- y + y
    z[j, lazy = TRUE] <- 1
    
    expect_equal(y$.filebase, x$.filebase)
    expect_equal(z$.filebase, x$.filebase)
    
    # mature z
    expect_equal(z[j], rep(1, sum(j)))
    expect_equal(y[j], rep(0, sum(j)))
    
    expect_equal(x[!i], x0[!i])
    expect_equal(y[!i], x[!i] + 1L)
    expect_equal(z[!i], y[!i] * 2)
    
    expect_equal(x[!j], x0[!i])
    expect_equal(y[!j], x[!i] + 1L)
    expect_equal(z[!j], y[!i] * 2)
    
    clear_cache()
})

test_that("subset filearray-proxy", {

    # normal indexing
    x <- as_filearray(1:120, dimension = c(10, 12))
    dimnames(x) <- list(A = 1:10, B = 1:12)
    y <- x + 1L
    z <- y + y
    
    idx1 <- c(3, 3, 4, 5, 5, 4, 3, 2, 1, 1, 2, 3, 3, 2, 2)
    idx2 <- idx1 * 2
    
    expect_equal(y[idx1, idx2], x[idx1, idx2] + 1)
    expect_equal(z[idx1, idx2], 2 * y[idx1, idx2])
    
    expect_equal(y[, idx2], x[, idx2] + 1)
    expect_equal(z[, idx2], 2 * y[, idx2])
    
    expect_equal(y[idx1, ], x[idx1, ] + 1)
    expect_equal(z[idx1, ], 2 * y[idx1, ])
    
    # filearray as index
    i <- array(FALSE, dim(x))
    i[idx1, idx2] <- TRUE
    j <- as_filearray(i)
    expect_equal(typeof(j), "logical")
    
    expect_equal(x[i], x[][i])
    expect_equal(x[i], x[][j[]])
    
    expect_equal(y[i], x[i] + 1L)
    expect_equal(y[j], y[i])
    
    expect_equal(z[i], y[i] * 2)
    expect_equal(z[j], z[i])
    
    # integer index
    idx <- rep(which(i), 2)
    
    # expect_equal(x[idx], x[][idx])
    # expect_equal(y[idx], y[idx])
    # expect_equal(z[idx], z[idx])
    expect_error(x[idx])
    
})

test_that("subset with empty selection (no matches)", {
    
    # Test case that previously caused segfault:
    # When subset formula evaluates to empty selection on non-first dimension
    x <- as_filearray(matrix(1:16, 4))
    dimnames(x) <- list(A = 1:4, B = 1:4)
    
    # Empty selection on second dimension (was causing segfault)
    result_B <- subset(x, B ~ B == 5)
    expect_equal(dim(result_B), c(4L, 0L))
    expect_equal(length(result_B), 0L)
    
    # Empty selection on first dimension (was working)
    result_A <- subset(x, A ~ A == 5)
    expect_equal(dim(result_A), c(0L, 4L))
    expect_equal(length(result_A), 0L)
    
    # Empty selection on both dimensions
    result_both <- subset(x, A ~ A == 5, B ~ B == 5)
    expect_equal(dim(result_both), c(0L, 0L))
    expect_equal(length(result_both), 0L)
    
    # Direct bracket indexing with empty indices
    expect_equal(dim(x[integer(0), ]), c(0L, 4L))
    expect_equal(dim(x[, integer(0)]), c(4L, 0L))
    expect_equal(dim(x[integer(0), integer(0)]), c(0L, 0L))
    
    # 3D array case
    y <- as_filearray(array(1:24, c(2, 3, 4)))
    dimnames(y) <- list(A = 1:2, B = 1:3, C = 1:4)
    
    # Empty on last dimension
    result_C <- subset(y, C ~ C == 10)
    expect_equal(dim(result_C), c(2L, 3L, 0L))
    
    # Empty on middle dimension  
    result_B3 <- subset(y, B ~ B == 10)
    expect_equal(dim(result_B3), c(2L, 0L, 4L))
    
    # Empty on first dimension
    result_A3 <- subset(y, A ~ A == 10)
    expect_equal(dim(result_A3), c(0L, 3L, 4L))
    
})

test_that("as_filearray", {
    options("filearray.operator.precision" = "float")
    on.exit({
        options("filearray.operator.precision" = NULL)
    }, add = TRUE)
    
    x <- rnorm(24)
    dim(x) <- c(2, 3, 4)
    
    # as_filearray.default
    y <- as_filearray(x)
    testthat::expect_equal(y$type(), getOption("filearray.operator.precision"))
    testthat::expect_equal(y[], x, tolerance = 1e-5, ignore_attr = TRUE)
    testthat::expect_equal(y$.mode, "readwrite")
    
    # as_filearray.character
    y <- as_filearray(y$.filebase)
    testthat::expect_equal(y[dimnames = FALSE], x, tolerance = 1e-5, ignore_attr = TRUE)
    testthat::expect_equal(y$.mode, "readonly")
    
    # as_filearray.FileArray(Proxy)
    z <- as_filearray(y)
    testthat::expect_identical(z, y)
    
    z <- as_filearray(as_filearrayproxy(y))
    testthat::expect_equal(typeof(z), getOption("filearray.operator.precision"))
    testthat::expect_equal(z[], x, tolerance = 1e-5, ignore_attr = TRUE)
    
    y$.mode <- "readwrite"
    y$delete()
    
    testthat::expect_false(z$valid())
    
})


test_that("as_filearrayproxy", {
    options("filearray.operator.precision" = "float")
    on.exit({
        options("filearray.operator.precision" = NULL)
    }, add = TRUE)
    
    x <- rnorm(24)
    dim(x) <- c(2, 3, 4)
    dimnames(x) <- list(A = 1:2, B = 1:3, C = letters[1:4])
    
    # as_filearrayproxy.default
    y <- as_filearrayproxy(x)
    testthat::expect_equal(typeof(y), getOption("filearray.operator.precision"))
    testthat::expect_equal(y[], x, tolerance = 1e-5)
    testthat::expect_equal(y$.mode, "readwrite")
    
    # as_filearrayproxy.character
    y <- as_filearrayproxy(y$.filebase)
    testthat::expect_equal(y[], x, tolerance = 1e-5)
    testthat::expect_equal(y$.mode, "readwrite")
    
    # as_filearrayproxy.FileArrayProxy
    z <- as_filearrayproxy(y)
    testthat::expect_false(identical(z, y))
    testthat::expect_equal(y[], z[])
    
    y$delete()
    
    testthat::expect_false(z$valid())
    
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

FARR_collapse <- function(filebase, dim, keep, cum_part, array_type, method = 1L, remove_na = FALSE, scale = 1.0) {
    .Call(`_filearray_FARR_collapse`, filebase, dim, keep, cum_part, array_type, method, remove_na, scale)
}

FARR_collapse_complex <- function(filebase, dim, keep, cum_part, method = 1L, remove_na = FALSE, scale = 1.0) {
    .Call(`_filearray_FARR_collapse_complex`, filebase, dim, keep, cum_part, method, remove_na, scale)
}

realToInt64 <- function(x, min_, max_, strict) {
    .Call(`_filearray_realToInt64`, x, min_, max_, strict)
}

cplxToReal2 <- function(x) {
    .Call(`_filearray_cplxToReal2`, x)
}

realToCplx2 <- function(x) {
    .Call(`_filearray_realToCplx2`, x)
}

realToFloat2 <- function(x) {
    .Call(`_filearray_realToFloat2`, x)
}

floatToReal2 <- function(x) {
    .Call(`_filearray_floatToReal2`, x)
}

get_float_na <- function() {
    .Call(`_filearray_get_float_na`)
}

set_buffer_size <- function(size) {
    .Call(`_filearray_set_buffer_size`, size)
}

get_buffer_size <- function() {
    .Call(`_filearray_get_buffer_size`)
}

FARR_meta <- function(filebase) {
    .Call(`_filearray_FARR_meta`, filebase)
}

loc2idx <- function(sliceIdx, dim) {
    .Call(`_filearray_loc2idx`, sliceIdx, dim)
}

locationList <- function(listOrEnv, dim, strict) {
    .Call(`_filearray_locationList`, listOrEnv, dim, strict)
}

schedule <- function(listOrEnv, dim, cum_part_sizes, split_dim, strict) {
    .Call(`_filearray_schedule`, listOrEnv, dim, cum_part_sizes, split_dim, strict)
}

filearray_meta <- function(filebase) {
    .Call(`_filearray_filearray_meta`, filebase)
}

filearray_assign <- function(filebase, value, position_indices) {
    .Call(`_filearray_filearray_assign`, filebase, value, position_indices)
}

filearray_subset <- function(filebase, position_indices, drop = TRUE, use_dimnames = TRUE, reshape = NULL) {
    .Call(`_filearray_filearray_subset`, filebase, position_indices, drop, use_dimnames, reshape)
}

FARR_subset_sequential <- function(filebase, unit_partlen, cum_partsizes, array_type, ret, from = 0L, len = 1L) {
    .Call(`_filearray_FARR_subset_sequential`, filebase, unit_partlen, cum_partsizes, array_type, ret, from, len)
}

FARR_subset2 <- function(filebase, listOrEnv, reshape = NULL, drop = FALSE, use_dimnames = TRUE, thread_buffer = 0L, split_dim = 0L, strict = 1L) {
    .Call(`_filearray_FARR_subset2`, filebase, listOrEnv, reshape, drop, use_dimnames, thread_buffer, split_dim, strict)
}

FARR_buffer_map <- function(input_filebases, output_filebase, map, buffer_nelems, result_nelems = 0L) {
    .Call(`_filearray_FARR_buffer_map`, input_filebases, output_filebase, map, buffer_nelems, result_nelems)
}

FARR_buffer_map2 <- function(input_filebases, map, buffer_nelems) {
    .Call(`_filearray_FARR_buffer_map2`, input_filebases, map, buffer_nelems)
}

FARR_buffer_mapreduce <- function(filebase, map, reduce, buffer_nelems) {
    .Call(`_filearray_FARR_buffer_mapreduce`, filebase, map, reduce, buffer_nelems)
}

FARR_subset_assign_sequential <- function(filebase, unit_partlen, cum_partsizes, array_type, value, from) {
    .Call(`_filearray_FARR_subset_assign_sequential`, filebase, unit_partlen, cum_partsizes, array_type, value, from)
}

FARR_subset_assign2 <- function(filebase, value, listOrEnv, thread_buffer = 2097152L, split_dim = 0L) {
    .Call(`_filearray_FARR_subset_assign2`, filebase, value, listOrEnv, thread_buffer, split_dim)
}

getDefaultNumThreads <- function() {
    .Call(`_filearray_getDefaultNumThreads`)
}

getThreads <- function(max) {
    .Call(`_filearray_getThreads`, max)
}

test_farr_findVarInFrame_ <- function(env, sym_name) {
    .Call(`_filearray_test_farr_findVarInFrame_`, env, sym_name)
}

kinda_sorted <- function(idx, min_, buffer_count) {
    .Call(`_filearray_kinda_sorted`, idx, min_, buffer_count)
}

check_missing_dots <- function(env) {
    .Call(`_filearray_check_missing_dots`, env)
}

reshape_or_drop <- function(x, reshape, drop) {
    .Call(`_filearray_reshape_or_drop`, x, reshape, drop)
}

# Tests for farr_findVarInFrame (src/compat.h)
#
# farr_findVarInFrame(rho, symbol) is the compatibility wrapper that replaces
# Rf_findVarInFrame on R >= 4.5.0, where that function became non-API.
#
# The three possible outcomes, all verified below:
#   1. Normal binding  → returns the bound value
#   2. Unbound symbol  → returns R_UnboundValue  (is_unbound = TRUE from helper)
#   3. Missing binding → returns R_MissingArg    (is_missing = TRUE from helper)
#
# On R >= 4.5.0, the wrapper uses R_existsVarInFrame + ...length() + R_getVarEx
# to reproduce the exact Rf_findVarInFrame behavior, including safe handling
# of R_MissingArg (empty dots) without longjmp.
#
# Additionally, the wrapper must NOT walk parent frames (inherits = FALSE).
#
# Proxy function: test_farr_findVarInFrame_(env, "sym") returns a list with
# $result / $is_unbound / $is_missing; see src/utils.cpp.
#
# We also test the real call-sites (check_missing_dots and locationList) to
# verify end-to-end behaviour is identical across R versions.

require(testthat)

# ── helpers ──────────────────────────────────────────────────────────────────

look <- function(env, name) filearray:::test_farr_findVarInFrame_(env, name)

# ── 1. Normal binding ─────────────────────────────────────────────────────────

test_that("farr_findVarInFrame: normal binding returns the value", {
    e <- new.env(parent = emptyenv())
    e$myval <- 42L
    res <- look(e, "myval")
    expect_false(res$is_unbound)
    expect_false(res$is_missing)
    expect_identical(res$result, 42L)

    e$vec <- c(1.1, 2.2, 3.3)
    res2 <- look(e, "vec")
    expect_identical(res2$result, c(1.1, 2.2, 3.3))

    e$txt <- "hello"
    res3 <- look(e, "txt")
    expect_identical(res3$result, "hello")
})

# ── 2. Unbound symbol ─────────────────────────────────────────────────────────

test_that("farr_findVarInFrame: unbound symbol returns R_UnboundValue", {
    e <- new.env(parent = emptyenv())
    res <- look(e, "no_such_var")
    expect_true(res$is_unbound)
    expect_false(res$is_missing)
})

# ── 3. Does NOT search parent frames ─────────────────────────────────────────

test_that("farr_findVarInFrame: does not inherit from parent frames", {
    parent_e <- new.env(parent = emptyenv())
    parent_e$x <- 99L

    child_e <- new.env(parent = parent_e)
    # x is visible from child_e via parent, but NOT in child_e's own frame
    expect_identical(get("x", envir = child_e, inherits = TRUE), 99L)

    res <- look(child_e, "x")
    expect_true(res$is_unbound)   # not found in child frame itself
    expect_false(res$is_missing)
})

# ── 4. R_MissingArg via R_DotsSymbol ─────────────────────────────────────────

test_that("farr_findVarInFrame: R_MissingArg binding (empty '...')", {
    # When a function with '...' is called with no extra arguments, R binds
    # R_DotsSymbol to R_MissingArg inside the call frame.
    # farr_findVarInFrame returns R_MissingArg on all R versions:
    #   - R < 4.5.0: Rf_findVarInFrame returns it directly.
    #   - R >= 4.5.0: R_existsVarInFrame + ...length() detects empty dots.

    f_missing <- function(...) {
        look(environment(), "...")
    }
    res <- f_missing()
    expect_false(res$is_unbound)
    expect_true(res$is_missing)
})

# ── 5. R_DotsSymbol with actual dots ─────────────────────────────────────────

test_that("farr_findVarInFrame: populated '...' binding returns DOTSXP", {
    f_filled <- function(...) {
        look(environment(), "...")
    }
    res <- f_filled(1, 2, 3)
    expect_false(res$is_unbound)
    expect_false(res$is_missing)
    # $result holds the DOTSXP; Rcpp wraps it as a list
    expect_true(is.list(res$result) || !is.null(res$result))
})

# ── 6. '...' not declared → unbound ──────────────────────────────────────────

test_that("farr_findVarInFrame: '...' not declared in function → unbound", {
    # A function without '...' has no binding for R_DotsSymbol at all.
    f_nodots <- function(x) {
        look(environment(), "...")
    }
    res <- f_nodots(1)
    expect_true(res$is_unbound)
    expect_false(res$is_missing)
})

# ── 7. check_missing_dots — the primary call-site ────────────────────────────
#
# check_missing_dots() is the real consumer of farr_findVarInFrame for the
# '...' symbol.  These cases exercise the full code path end-to-end.

test_that("check_missing_dots: empty call", {
    misdot <- function(...) filearray:::check_missing_dots(environment())
    # Empty dots: farr_findVarInFrame returns R_MissingArg, loop exits
    # immediately → zero-length logical vector (same as dipsaus behavior).
    expect_equal(misdot(), logical(0))
})

test_that("check_missing_dots: all-missing args", {
    misdot <- function(...) filearray:::check_missing_dots(environment())
    expect_equal(misdot(, , ), c(TRUE, TRUE, TRUE))
})

test_that("check_missing_dots: mixed missing / present args", {
    misdot <- function(...) filearray:::check_missing_dots(environment())
    expect_equal(misdot(, , 1),    c(TRUE, TRUE, FALSE))
    expect_equal(misdot(1, , 2),   c(FALSE, TRUE, FALSE))
    expect_equal(misdot(1, 2, 3),  c(FALSE, FALSE, FALSE))
})

# ── 8. locationList ENVSXP branch ────────────────────────────────────────────
#
# locationList() uses farr_findVarInFrame when given an environment (to look
# up '...' and iterate the dots), which is the other call-site in core.cpp.

test_that("locationList via environment: full dims → identity", {
    dim <- c(3L, 4L, 5L)

    # Must use '...' so the args are bound to R_DotsSymbol in the environment.
    # locationList ENVSXP branch calls farr_findVarInFrame(env, R_DotsSymbol).
    f <- function(...) filearray:::locationList(environment(), dim, 1L)
    as_int64 <- function(x) filearray:::realToInt64(as.numeric(x), NA_real_, NA_real_, 0L)

    locs <- f(1:3, 1:4, 1:5)
    expect_equal(locs[[1]], as_int64(1:3))
    expect_equal(locs[[2]], as_int64(1:4))
    expect_equal(locs[[3]], as_int64(1:5))
})

test_that("locationList via environment: too many dims → error", {
    dim <- c(3L, 4L)
    # Must use '...' so all args land in R_DotsSymbol (3 dots > 2 dims → error)
    f <- function(...) filearray:::locationList(environment(), dim, 1L)
    expect_error(f(1, 2, 3))
})

# Testing collapse is time consuming, skip if ran
skip_collapse <- Sys.getenv("FILEARRAY_SKIP_COLLAPSE", unset = "") == "TRUE"
testthat::skip_if(skip_collapse)

collapse_real <- function(y,
                          keep,
                          transform = c("asis", "10log10", "square", "sqrt", "normalize")) {
    re <- switch(
        transform,
        "asis" = {
            apply(y, keep, function(x) {
                mean(x)
            })
        },
        "10log10" = {
            apply(y, keep, function(x) {
                mean(10 * log10(x))
            })
        },
        "square" = {
            apply(y, keep, function(x) {
                mean(x^2)
            })
        },
        "sqrt" = {
            apply(y, keep, function(x) {
                mean(sqrt(x))
            })
        },
        {
            stop("wrong transform")
        }
    )
    # if(storage.mode(re) != "double") {
    #     storage.mode(re) <- "double"
    # }
    re
}

collapse_cplx <- function(y,
                          keep,
                          transform = c("asis", "10log10", "square", "sqrt", "normalize")) {
    re <- switch(
        transform,
        "asis" = {
            apply(y, keep, mean)
        },
        "10log10" = {
            apply(y, keep, function(x) {
                mean(20 * log10(Mod(x)))
            })
        },
        "square" = {
            apply(y, keep, function(x) {
                mean(Mod(x)^2)
            })
        },
        "sqrt" = {
            apply(y, keep, function(x) {
                mean(Mod(x))
            })
        },
        "normalize" = {
            apply(y, keep, function(x) {
                mean(x / Mod(x))
            })
        },
        {
            stop("wrong transform")
        }
    )
    # if(storage.mode(re) != "double") {
    #     storage.mode(re) <- "double"
    # }
    re
}

expect_equivalent_cplx <- function(x, y, eps = 1e-6) {
    expect_equal(is.na(x), is.na(y))
    if (is.complex(x)) {
        expect_lte(max(abs(Re(x - y)), na.rm = TRUE), eps)
        expect_lte(max(abs(Im(x - y)), na.rm = TRUE), eps)
    } else {
        expect_lte(max(abs(x - y), na.rm = TRUE), eps)
    }
}

test_that("R/C++ - Collapse", {
    testthat::skip_on_cran()
    bsz <- get_buffer_size()
    on.exit({
        set_buffer_size(bsz)
        max_buffer_size(2097152)
    })
    set_buffer_size(16L)
    max_buffer_size(64L)
    
    # dim <- c(287, 100, 301, 7)
    dim <- c(33:36)
    set.seed(5)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    x <- filearray_create(
        file,
        dim,
        type = "integer",
        partition_size = 2,
        initialize = FALSE
    )
    y <- array(1:(prod(dim)), dim)
    y[[20, 3, 3, 3]] <- NA
    storage.mode(y) <- "integer"
    x[] <- y
    
    # make sure x[] == y
    expect_equal(x[], y)
    
    
    
    # collapse
    keep <- c(1, 2, 3, 4)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    
    keep <- c(1, 4, 3, 2)
    expect_equal(
        x$collapse(
            keep = keep,
            transform = "asis",
            method = "mean"
        ),
        collapse_real(y, keep, transform = "asis")
    )
    
    keep <- c(4, 2, 3, 1)
    expect_equal(
        x$collapse(
            keep = keep,
            transform = "asis",
            method = "mean"
        ),
        collapse_real(y, keep, transform = "asis")
    )
    
    keep <- c(3, 1)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    
    keep <- c(4, 1)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    
    keep <- c(4, 2)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    keep <- c(4, 3)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    keep <- c(4)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        k <- x$collapse(keep = keep,
                        transform = transform,
                        method = "mean")
        s <- collapse_real(y, keep, transform = transform)
        diff <- max(abs(1 - s / k), na.rm = TRUE)
        # cat(transform, diff, "\n")
        expect_lt(diff, 1e-6)
    }
    keep <- c(4, 1, 3)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    keep <- c(3)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        k <- x$collapse(keep = keep,
                        transform = transform,
                        method = "mean")
        s <- collapse_real(y, keep, transform = transform)
        diff <- max(abs(1 - s / k), na.rm = TRUE)
        # cat(transform, diff, "\n")
        expect_lt(diff, 1e-6)
    }
    keep <- c(1)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        k <- x$collapse(keep = keep,
                        transform = transform,
                        method = "mean")
        s <- collapse_real(y, keep, transform = transform)
        diff <- max(abs(1 - s / k), na.rm = TRUE)
        # cat(transform, diff, "\n")
        expect_lt(diff, 1e-6)
    }
    
})

test_that("R/C++ - Float", {
    testthat::skip_on_cran()
    bsz <- get_buffer_size()
    on.exit({
        set_buffer_size(bsz)
        max_buffer_size(2097152)
    })
    set_buffer_size(16L)
    max_buffer_size(64L)
    
    # dim <- c(287, 100, 301, 7)
    dim <- c(33:36)
    set.seed(5)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    x <- filearray_create(
        file,
        dim,
        type = "float",
        partition_size = 2,
        initialize = FALSE
    )
    y <- array(rnorm(length(x))^2, dim)
    y[[20, 3, 3, 3]] <- NA
    x[] <- y
    
    # make sure x[] == y
    eps <- 10^(ceiling(log10(max(abs(
        y
    ), na.rm = TRUE))) - 7)
    expect_equal(x[], y, tolerance = eps)
    y <- x[]
    
    
    
    # collapse
    keep <- c(1, 2, 3, 4)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    
    keep <- c(1, 4, 3, 2)
    expect_equal(
        x$collapse(
            keep = keep,
            transform = "asis",
            method = "mean"
        ),
        collapse_real(y, keep, transform = "asis")
    )
    
    keep <- c(4, 2, 3, 1)
    expect_equal(
        x$collapse(
            keep = keep,
            transform = "asis",
            method = "mean"
        ),
        collapse_real(y, keep, transform = "asis")
    )
    
    keep <- c(3, 1)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    
    keep <- c(4, 1)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    
    keep <- c(4, 2)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    keep <- c(4, 3)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    keep <- c(4)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        k <- x$collapse(keep = keep,
                        transform = transform,
                        method = "mean")
        s <- collapse_real(y, keep, transform = transform)
        diff <- max(abs(1 - s / k), na.rm = TRUE)
        # cat(transform, diff, "\n")
        expect_lt(diff, 1e-6)
    }
    keep <- c(4, 1, 3)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_real(y, keep, transform = transform)
        )
    }
    keep <- c(3)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        k <- x$collapse(keep = keep,
                        transform = transform,
                        method = "mean")
        s <- collapse_real(y, keep, transform = transform)
        diff <- max(abs(1 - s / k), na.rm = TRUE)
        # cat(transform, diff, "\n")
        expect_lt(diff, 1e-6)
    }
    keep <- c(1)
    for (transform in c("asis", "10log10", "square", "sqrt")) {
        k <- x$collapse(keep = keep,
                        transform = transform,
                        method = "mean")
        s <- collapse_real(y, keep, transform = transform)
        diff <- max(abs(1 - s / k), na.rm = TRUE)
        # cat(transform, diff, "\n")
        expect_lt(diff, 1e-6)
    }
    
})

test_that("R/C++ - Collapse (complex)", {
    testthat::skip_on_cran()
    bsz <- get_buffer_size()
    on.exit({
        set_buffer_size(bsz)
        max_buffer_size(2097152)
    })
    set_buffer_size(16L)
    max_buffer_size(64L)
    
    # dim <- c(287, 100, 301, 7)
    dim <- c(33:36)
    set.seed(5)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    x <- filearray_create(
        file,
        dim,
        type = "complex",
        partition_size = 2,
        initialize = FALSE
    )
    y <- array(rnorm(length(x)) + rnorm(length(x)) * 1i, dim)
    y[[20, 3, 3, 3]] <- NA
    x[] <- y
    
    # make sure x[] == y
    expect_equivalent_cplx(x[], y)
    
    y <- x[]
    
    # collapse
    keep <- c(1, 2, 3, 4)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_cplx(y, keep, transform = transform)
        )
    }
    
    keep <- c(1, 4, 3, 2)
    expect_equal(
        x$collapse(
            keep = keep,
            transform = "asis",
            method = "mean"
        ),
        collapse_cplx(y, keep, transform = "asis")
    )
    
    keep <- c(4, 2, 3, 1)
    expect_equal(
        x$collapse(
            keep = keep,
            transform = "asis",
            method = "mean"
        ),
        collapse_cplx(y, keep, transform = "asis")
    )
    
    keep <- c(3, 1)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_cplx(y, keep, transform = transform)
        )
    }
    
    keep <- c(4, 1)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_cplx(y, keep, transform = transform)
        )
    }
    
    keep <- c(4, 2)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_cplx(y, keep, transform = transform)
        )
    }
    keep <- c(4, 3)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_cplx(y, keep, transform = transform)
        )
    }
    keep <- c(4)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(max(abs(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ) -
                collapse_cplx(y, keep, transform = transform)
        ), na.rm = TRUE), 0)
    }
    keep <- c(4, 1, 3)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ),
            collapse_cplx(y, keep, transform = transform)
        )
    }
    keep <- c(3)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(max(abs(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ) -
                collapse_cplx(y, keep, transform = transform)
        ), na.rm = TRUE), 0)
    }
    keep <- c(1)
    for (transform in c("asis", "10log10", "square", "sqrt", "normalize")) {
        expect_equal(max(abs(
            x$collapse(
                keep = keep,
                transform = transform,
                method = "mean"
            ) -
                collapse_cplx(y, keep, transform = transform)
        ), na.rm = TRUE), 0)
    }
    
})


Sys.setenv("FILEARRAY_SKIP_COLLAPSE" = "TRUE")

require(testthat)
as_int64 <- function(x) {
    realToInt64(x, NA_real_, NA_real_, 0L)
}

test_that("C++: Utils", {
    misdot <- function(...) {
        check_missing_dots(environment())
    }
    expect_equal(misdot(, , 1), c(TRUE, TRUE, FALSE))
    expect_equal(misdot(, i = 1, 1), c(TRUE, FALSE, FALSE))
    expect_equal(misdot(), logical(0))
    
    # min buffer size is 16, otherwise round up to power of 2
    expect_equal(set_buffer_size(16), 16L)
    expect_equal(set_buffer_size(2), 16L)
    expect_equal(set_buffer_size(55.4), 64L)
    expect_error(set_buffer_size(0))
    expect_error(set_buffer_size(-1))
    expect_error(set_buffer_size(NA))
    expect_warning(expect_error(set_buffer_size(2^31)))
    
    
    expect_equal(kinda_sorted(as_int64(sample(1:10)), 1, 10), 1L)
    expect_equal(kinda_sorted(as_int64(c(sample(
        1:10
    ), 1)), 1, 10), 1L)
    expect_equal(kinda_sorted(as_int64(c(sample(
        1:10
    ), 1)), 1, 9), 0L)
    
    expect_equal(locationList(list(1:10), 10, 1)[[1]], as_int64(1:10))
    expect_error(locationList(list(1:11), 10, 1))
    expect_equal(locationList(list(1:11), 10, 0)[[1]], as_int64(c(1:10, NA_integer_)))
    
    x <- 1:10
    reshape_or_drop(x, c(2, 5), 1)
    expect_equal(dim(x), c(2, 5))
    
    reshape_or_drop(x, c(1, 10), 1)
    expect_equal(dim(x), c(1, 10))
    
    reshape_or_drop(x, NULL, 0)
    expect_equal(dim(x), c(1, 10))
    
    reshape_or_drop(x, NULL, 1)
    expect_equal(dim(x), NULL)
    
    dim <- c(3, 5, 4)
    x <- array(1:prod(dim), dim)
    locs <- lapply(dim, function(x) {
        sample(c(1:x, NA, NA), size = 20, replace = TRUE)
    })
    re1 <- x[locs[[1]], locs[[2]], locs[[3]]]
    locs <- lapply(locs, as_int64)
    
    v1 <- loc2idx(locs, dim)
    v2 <- as_int64(re1 - 1L)
    expect_identical(v1, v2)
    
})

test_that("C++: IO - subset/assign", {
    bsz <- get_buffer_size()
    on.exit({
        set_buffer_size(bsz)
        max_buffer_size(2097152)
    })
    set_buffer_size(16L)
    max_buffer_size(64L)
    
    set.seed(NULL)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    dim <- 33:35
    x <- filearray_create(file,
                          dim,
                          partition_size = 2,
                          initialize = FALSE)
    
    expect_equal(x[[2]], x$.na)
    
    y <- array(x$.na, dim)
    x[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- 1:prod(dim)
    y[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- 1:prod(dim)
    
    expect_equal(x[], y)
    locs <-
        lapply(dim, function(d) {
            d1 <- sample(c(1:d), 10, replace = TRUE)
            d2 <- c(NA, NA)
            as.double(sample(c(d1, d2)))
        })
    expect_equal(x[locs[[1]], locs[[2]], locs[[3]]], y[locs[[1]], locs[[2]], locs[[3]]])
    expect_equal(x[, , c(TRUE, NA, FALSE)], y[, , c(TRUE, NA, FALSE)])
    
    expect_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
    })
    
    locs <-
        lapply(dim, function(d) {
            d1 <- sample(c(1:d), 10, replace = TRUE)
            d2 <- c(NA, NA)
            as.double(sample(c(d1, d2)))
        })
    expect_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
    })
    
    locs <-
        lapply(dim, function(d) {
            sample(c(1:d), size = d, replace = FALSE)
        })
    expect_no_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1
    })
    expect_equal(unique(as.vector(x[locs[[1]], locs[[2]], locs[[3]]])), 1)
    
    expect_true({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
        TRUE
    })
    y[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
    
    expect_equal(x[], y)
    
    expect_error(x[c(rep(c(TRUE, TRUE, TRUE, TRUE), 8), TRUE, FALSE), , ])
    expect_error(x[c(rep(c(TRUE, TRUE, TRUE, TRUE), 8), TRUE, NA), , ])
    
    unlink(file, recursive = TRUE)
    x <- filearray_create(file,
                          dim,
                          partition_size = 2,
                          initialize = FALSE)
    y <- array(x$.na, dim)
    x[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- 1:prod(dim)
    y[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- 1:prod(dim)
    expect_equal(x[], y)
    
    expect_error({
        x[3:1, , c(TRUE, FALSE, TRUE, TRUE, rep(NA, 29))] <- 1:36
    })
    
    unlink(file, recursive = TRUE)
})

test_that("C++: IO - subset/assign - complex", {
    bsz <- get_buffer_size()
    on.exit({
        set_buffer_size(bsz)
        max_buffer_size(2097152)
    })
    set_buffer_size(16L)
    max_buffer_size(64L)
    
    set.seed(8)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    dim <- 33:35
    x <- filearray_create(
        file,
        dim,
        partition_size = 2,
        type = "complex",
        initialize = FALSE
    )
    
    expect_equal(x[[2]], x$.na)
    
    y <- array(x$.na, dim)
    tmp <- rnorm(prod(dim)) + 1i * runif(prod(dim))
    x[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- tmp
    y[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- tmp
    
    expect_lt(max(Mod(x[] - y)), 1e-6)
    locs <-
        lapply(dim, function(d) {
            d1 <- sample(c(1:d), 10, replace = TRUE)
            d2 <- c(NA, NA)
            as.double(sample(c(d1, d2)))
        })
    max_dif <- max(Mod(x[locs[[1]], locs[[2]], locs[[3]]] -
                           y[locs[[1]], locs[[2]], locs[[3]]]), na.rm = TRUE)
    expect_lt(max_dif, 1e-5)
    if (max_dif > 1e-6) {
        print(max_dif)
    }
    expect_equal(is.na(x[locs[[1]], locs[[2]], locs[[3]]]), is.na(y[locs[[1]], locs[[2]], locs[[3]]]))
    expect_lt(max(Mod(x[, , c(TRUE, NA, FALSE)] -
                          y[, , c(TRUE, NA, FALSE)]), na.rm = TRUE), 1e-6)
    expect_equal(is.na(x[, , c(TRUE, NA, FALSE)]), is.na(y[, , c(TRUE, NA, FALSE)]))
    
    expect_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- tmp[1:prod(sapply(locs, length))]
    })
    
    locs <-
        lapply(dim, function(d) {
            d1 <- sample(c(1:d), 10, replace = TRUE)
            d2 <- c(NA, NA)
            as.double(sample(c(d1, d2)))
        })
    expect_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- tmp[1:prod(sapply(locs, length))]
    })
    
    locs <-
        lapply(dim, function(d) {
            sample(c(1:d), size = d, replace = FALSE)
        })
    expect_no_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1
    })
    
    expect_equal(unique(as.vector(x[locs[[1]], locs[[2]], locs[[3]]])), 1 + 0i)
    
    expect_true({
        x[locs[[1]], locs[[2]], locs[[3]]] <- tmp[1:prod(sapply(locs, length))]
        TRUE
    })
    y[locs[[1]], locs[[2]], locs[[3]]] <- tmp[1:prod(sapply(locs, length))]
    
    expect_lt(max(Mod(x[] - y)), 1e-6)
    
    expect_error(x[c(rep(c(TRUE, TRUE, TRUE, TRUE), 8), TRUE, FALSE), , ])
    expect_error(x[c(rep(c(TRUE, TRUE, TRUE, TRUE), 8), TRUE, NA), , ])
    
    unlink(file, recursive = TRUE)
    x <- filearray_create(
        file,
        dim,
        partition_size = 2,
        type = "complex",
        initialize = FALSE
    )
    y <- array(x$.na, dim)
    x[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- tmp
    y[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- tmp
    expect_lt(max(Mod(x[] - y)), 1e-6)
    
    expect_error({
        x[3:1, , c(TRUE, FALSE, TRUE, TRUE, rep(NA, 29))] <- tmp[1:36]
    })
    
    unlink(file, recursive = TRUE)
})

test_that("C++: IO - subset/assign - float", {
    bsz <- get_buffer_size()
    on.exit({
        set_buffer_size(bsz)
        max_buffer_size(2097152)
    })
    set_buffer_size(16L)
    max_buffer_size(64L)
    
    set.seed(NULL)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    dim <- 33:35
    x <- filearray_create(
        file,
        dim,
        partition_size = 2,
        type = "float",
        initialize = FALSE
    )
    
    expect_equal(x[[2]], x$.na)
    
    y <- array(x$.na, dim)
    tmp <- rnorm(prod(dim))
    x[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- tmp
    y[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- tmp
    
    eps <- 10^(ceiling(log10(max(abs(
        y
    )))) - 7)
    expect_equal(x[], y, tolerance = eps)
    y[] <- x[]
    locs <-
        lapply(dim, function(d) {
            d1 <- sample(c(1:d), 10, replace = TRUE)
            d2 <- c(NA, NA)
            as.double(sample(c(d1, d2)))
        })
    
    a <- x[locs[[1]], locs[[2]], locs[[3]]]
    b <- y[locs[[1]], locs[[2]], locs[[3]]]
    expect_equal(is.na(a), is.na(b))
    
    sel <- !is.na(a - b) & (a - b) > eps
    if (length(a[sel])) {
        # fail the test
        print(a[sel])
        print(b[sel])
        expect_length(object = (a - b)[sel], n = 0)
    }
    
    
    
    a <- x[c(1, 1, 2, 2, 1, 1, 2, 2), 1, c(2, 2)]
    b <- y[c(1, 1, 2, 2, 1, 1, 2, 2), 1, c(2, 2)]
    # a <- x[c(1,1,2,2), c(1,1,2,2), 1]
    # b <- y[c(1,1,2,2), c(1,1,2,2), 1]
    expect_equal(a, b, tolerance = eps)
    
    expect_equal(x[, , c(TRUE, NA, FALSE)], y[, , c(TRUE, NA, FALSE)])
    
    expect_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
    })
    
    locs <-
        lapply(dim, function(d) {
            d1 <- sample(c(1:d), 10, replace = TRUE)
            d2 <- c(NA, NA)
            as.double(sample(c(d1, d2)))
        })
    expect_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
    })
    
    locs <-
        lapply(dim, function(d) {
            sample(c(1:d), size = d, replace = FALSE)
        })
    expect_no_error({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1
    })
    expect_equal(unique(as.vector(x[locs[[1]], locs[[2]], locs[[3]]])), 1)
    
    
    expect_true({
        x[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
        TRUE
    })
    y[locs[[1]], locs[[2]], locs[[3]]] <- 1:prod(sapply(locs, length))
    
    expect_equal(x[], y)
    
    expect_error(x[c(rep(c(TRUE, TRUE, TRUE, TRUE), 8), TRUE, FALSE), , ])
    expect_error(x[c(rep(c(TRUE, TRUE, TRUE, TRUE), 8), TRUE, NA), , ])
    
    unlink(file, recursive = TRUE)
    x <- filearray_create(file,
                          dim,
                          partition_size = 2,
                          initialize = FALSE)
    y <- array(x$.na, dim)
    x[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- 1:prod(dim)
    y[33:1, , c(35, 2, 1, 3, 4, 5:34)] <- 1:prod(dim)
    expect_equal(x[], y)
    
    expect_error({
        x[3:1, , c(TRUE, FALSE, TRUE, TRUE, rep(NA, 29))] <- 1:36
    })
    
    unlink(file, recursive = TRUE)
})

test_that("C++: IO - type conversion", {
    set.seed(1)
    file <- tempfile()
    unlink(file, recursive = TRUE)
    dim <- 3:5
    x <- filearray_create(
        file,
        dim,
        type = "integer",
        partition_size = 2,
        initialize = FALSE
    )
    
    expect_equal(x[[2]], x$.na)
    
    y <- array(x$.na, dim)
    tmp <- rnorm(60) * 100
    x[3:1, , c(5, 2, 1, 3, 4)] <- tmp
    y[3:1, , c(5, 2, 1, 3, 4)] <- as.integer(tmp)
    
    expect_equal(x[], y)
    
    
    unlink(file, recursive = TRUE)
    x <- filearray_create(
        file,
        dim,
        type = "logical",
        partition_size = 2,
        initialize = FALSE
    )
    y <- array(x$.na, dim)
    tmp <- sample(c(1.1, 0, NA), replace = TRUE, size = 60)
    x[] <- tmp
    y[] <- as.logical(tmp)
    expect_equal(x[], y)
    
    unlink(file, recursive = TRUE)
    x <- filearray_create(file, dim, type = "raw", initialize = FALSE)
    y <- array(x$.na, dim)
    tmp <- sample(0:255, replace = TRUE, size = 60)
    tmp[sample(1:60, size = 10)] <- NA
    suppressWarnings({
        x[] <- tmp
        y[] <- as.raw(tmp)
    })
    
    expect_equal(x[], y)
    unlink(file, recursive = TRUE)
})

arith_type_check <- function(x, complex = "complex", ...) {
    switch(
        typeof(x),
        "raw" = { stop("non-numeric argument to mathematical function") },
        "complex" = { complex },
        ...,
        { "double" }
    )
}

# Arith
# "+", "-", "*", "^", "%%", "%/%", "/"
# 
# Compare
# "==", ">", "<", "!=", "<=", ">="
# 
# Logic
# "&", "|".
# 
# Ops
# "Arith", "Compare", "Logic"
# 
# Math
# "abs", "sign", "sqrt", "ceiling", "floor", "trunc", "cummax", "cummin", "cumprod", "cumsum", "log", "log10", "log2", "log1p", "acos", "acosh", "asin", "asinh", "atan", "atanh", "exp", "expm1", "cos", "cosh", "cospi", "sin", "sinh", "sinpi", "tan", "tanh", "tanpi", "gamma", "lgamma", "digamma", "trigamma"
# 
# Math2
# "round", "signif"
# 
# Summary
# "max", "min", "range", "prod", "sum", "any", "all"
# 
# Complex
# "Arg", "Conj", "Im", "Mod", "Re"



# Methods covered in this script: 
# "acos", "acosh", "asin", "asinh", "atan", "atanh", "cos", "cosh", "cospi", "sin", "sinh", "sinpi", "tan", "tanh", "tanpi", 
# "gamma", "lgamma", "digamma", "trigamma"
# 
# Complex
# "Arg", "Conj", "Im", "Mod", "Re"

FILEARRAY_SIMPLE_OPS <- list(
    
    logic = c("!", "is.na"),
    math = c(
        "abs", "sign", "sqrt", "ceiling", "floor", "trunc", "log", "log10", "log2", 
        "log1p", "acos", "acosh", "asin", "asinh", "atan", "atanh", "exp", "expm1", 
        "cos", "cosh", "cospi", "sin", "sinh", "sinpi", "tan", "tanh", "tanpi", 
        "gamma", "lgamma", "digamma", "trigamma", "round", "signif"
    ),
    complex = c("Arg", "Conj", "Im", "Mod", "Re")
    
)


#' @rdname S4-filearray
setMethod("!", signature(x = "FileArray"), function(x) {
    fa_operator(x, op = "!", out_type = "logical", label = parent_call(deparse = TRUE))
})

# ---- exp / log ---------------------------------------------------------------

#' @rdname S4-filearray
setMethod("exp", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "exp", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("expm1", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "expm1", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("log", signature(x = "FileArray"), function(x, base = exp(1)) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "log", out_type = out_type, base = base)
})

#' @rdname S4-filearray
setMethod("log10", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "log10", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("log2", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "log2", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("log1p", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "log1p", out_type = out_type, label = parent_call(deparse = TRUE))
})

# ---- abs / sqrt --------------------------------------------------------------
#' @rdname S4-filearray
setMethod("abs", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = "double")
    fa_operator(x, op = "abs", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("sqrt", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "sqrt", out_type = out_type, label = parent_call(deparse = TRUE))
})

# ---- ceiling / round / floor / trunc / signif / sign -------------------------

#' @rdname S4-filearray
setMethod("sign", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "sign", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("signif", signature(x = "FileArray"), function(x, digits = 6) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "signif", out_type = out_type, label = parent_call(deparse = TRUE), digits = digits)
})

#' @rdname S4-filearray
setMethod("trunc", signature(x = "FileArray"), function(x, ...) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "trunc", out_type = out_type, label = parent_call(deparse = TRUE), ...)
})

#' @rdname S4-filearray
setMethod("floor", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "floor", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("ceiling", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "ceiling", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("round", signature(x = "FileArray"), function(x, digits = 0) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "round", out_type = out_type, label = parent_call(deparse = TRUE), digits = digits)
})

# ---- "acos", "acosh", "asin", "asinh", "atan", "atanh", "cos", "cosh", -------

#' @rdname S4-filearray
setMethod("acos", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "acos", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("acosh", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "acosh", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("asin", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "asin", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("asinh", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "asinh", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("atan", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "atan", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("atanh", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "atanh", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("cos", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "cos", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("cosh", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "cosh", out_type = out_type, label = parent_call(deparse = TRUE))
})

# ---- "cospi", "sin", "sinh", "sinpi", "tan", "tanh", "tanpi", 
#' @rdname S4-filearray
setMethod("cospi", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "cospi", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("sin", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "sin", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("sinh", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "sinh", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("sinpi", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "sinpi", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("tan", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "tan", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("tanh", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x)
    fa_operator(x, op = "tanh", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("tanpi", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "tanpi", out_type = out_type, label = parent_call(deparse = TRUE))
})

# ---- "gamma", "lgamma", "digamma", "trigamma"

#' @rdname S4-filearray
setMethod("gamma", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "gamma", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("lgamma", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "lgamma", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("digamma", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "digamma", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("trigamma", signature(x = "FileArray"), function(x) {
    out_type <- arith_type_check(x, complex = stop("unimplemented complex function"))
    fa_operator(x, op = "trigamma", out_type = out_type, label = parent_call(deparse = TRUE))
})



# ---- "Arg", "Conj", "Im", "Mod", "Re" ----------------------------------------
#' @rdname S4-filearray
setMethod("Arg", signature(z = "FileArray"), function(z) {
    out_type <- arith_type_check(z, complex = "double")
    fa_operator(z, op = "Arg", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("Conj", signature(z = "FileArray"), function(z) {
    out_type <- arith_type_check(z)
    fa_operator(z, op = "Conj", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("Im", signature(z = "FileArray"), function(z) {
    out_type <- arith_type_check(z, complex = "double")
    fa_operator(z, op = "Im", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("Mod", signature(z = "FileArray"), function(z) {
    out_type <- arith_type_check(z, complex = "double")
    fa_operator(z, op = "Mod", out_type = out_type, label = parent_call(deparse = TRUE))
})

#' @rdname S4-filearray
setMethod("Re", signature(z = "FileArray"), function(z) {
    out_type <- arith_type_check(z, complex = "double")
    fa_operator(z, op = "Re", out_type = out_type, label = parent_call(deparse = TRUE))
})

# ---- is.na -------------------------------------------------------------------

#' @rdname S4-filearray
setMethod("is.na", signature(x = "FileArray"), function(x) {
    fa_operator(x, op = "is.na", out_type = "logical", label = parent_call(deparse = TRUE))
})

#' @title The type of a file array (extended)
#' @param x any file array
#' @return A character string. The possible values are \code{"double"},
#' \code{"integer"}, \code{"logical"}, and \code{"raw"}
#' @export
setGeneric("typeof")

#' @rdname typeof
setMethod("typeof", signature(x = "FileArray"), function(x) {
    if (!x$valid()) {
        stop("Invalid file array")
    }
    x$type()
})

#' @rdname typeof
setMethod("typeof", signature(x = "FileArrayProxy"), function(x) {
    if (!x$valid()) {
        stop("Invalid file array")
    }
    if ( length(x$.ops) ) {
        final_op <- x$.ops[[length(x$.ops)]]
        return( final_op$output_type )
    } else {
        x$type()
    }
})

register_double_op <- function(op, counter_types = c("FileArray", "numeric", "complex", "logical", "array"), ...) {
    
    
    func_name <- sprintf("`op_func_%s`", op)
    s <- sprintf("%s <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = '%s', label = parent_call(deparse = TRUE))
}\n", func_name, op)
    
    
    expr_str <- lapply(counter_types, function(ftype) {
        if ( ftype == "FileArray" ) {
            re <- deparse1(bquote(setMethod(.(op), signature(e1 = "FileArray", e2 = "FileArray"), .(str2lang(func_name)))))
        } else {
            re <- c(
                deparse1(bquote(setMethod(.(op), signature(e1 = "FileArray", e2 = .(ftype)), .(str2lang(func_name))))),
                deparse1(bquote(setMethod(.(op), signature(e1 = .(ftype), e2 = "FileArray"), .(str2lang(func_name)))))
            )
        }
        paste0("#' @rdname S4-filearray\n", re, "\n")
    })
    expr_str <- unlist(c(s, expr_str))
    cat(expr_str, "\n", sep  = "\n")
    # op_func <- function(e1, e2) {
    #     fa_pairwise_operator(e1, e2, op = op, label = parent_call(deparse = TRUE), ...)
    # }
    # 
    # for(ftype in counter_types) {
    #     if ( ftype == "FileArray" ) {
    #         setMethod(op, signature(e1 = "FileArray", e2 = "FileArray"), op_func)
    #     } else {
    #         setMethod(op, signature(e1 = "FileArray", e2 = ftype), op_func)
    #         setMethod(op, signature(e1 = ftype, e2 = "FileArray"), op_func)
    #     }
    # }
    
    rstudioapi <- asNamespace("rstudioapi")
    ctx <- rstudioapi$getActiveDocumentContext()
    line <- ctx$selection[[1]]$range$end[[1]]
    rstudioapi$insertText(c(line + 1, 1), paste(c(unlist(expr_str), ""), collapse = "\n"))
    invisible()
}


# Arith
# register_double_op("+")
`op_func_+` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "+", label = parent_call(deparse = TRUE))
}

#' @rdname S4-filearray
setMethod("+", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "FileArray", e2 = "numeric"), `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "numeric", e2 = "FileArray"), `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "FileArray", e2 = "complex"), `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "complex", e2 = "FileArray"), `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "FileArray", e2 = "logical"), `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "logical", e2 = "FileArray"), `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "FileArray", e2 = "array"), `op_func_+`)

#' @rdname S4-filearray
setMethod("+", signature(e1 = "array", e2 = "FileArray"), `op_func_+`)

# register_double_op("-")
`op_func_-` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "-", label = parent_call(deparse = TRUE))
}

#' @rdname S4-filearray
setMethod("-", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "FileArray", e2 = "numeric"), `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "numeric", e2 = "FileArray"), `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "FileArray", e2 = "complex"), `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "complex", e2 = "FileArray"), `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "FileArray", e2 = "logical"), `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "logical", e2 = "FileArray"), `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "FileArray", e2 = "array"), `op_func_-`)

#' @rdname S4-filearray
setMethod("-", signature(e1 = "array", e2 = "FileArray"), `op_func_-`)

# register_double_op("*")
`op_func_*` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "*", label = parent_call(deparse = TRUE))
}

#' @rdname S4-filearray
setMethod("*", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "FileArray", e2 = "numeric"), `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "numeric", e2 = "FileArray"), `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "FileArray", e2 = "complex"), `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "complex", e2 = "FileArray"), `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "FileArray", e2 = "logical"), `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "logical", e2 = "FileArray"), `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "FileArray", e2 = "array"), `op_func_*`)

#' @rdname S4-filearray
setMethod("*", signature(e1 = "array", e2 = "FileArray"), `op_func_*`)

# register_double_op("/")
`op_func_/` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "/", label = parent_call(deparse = TRUE))
}

#' @rdname S4-filearray
setMethod("/", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "FileArray", e2 = "numeric"), `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "numeric", e2 = "FileArray"), `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "FileArray", e2 = "complex"), `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "complex", e2 = "FileArray"), `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "FileArray", e2 = "logical"), `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "logical", e2 = "FileArray"), `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "FileArray", e2 = "array"), `op_func_/`)

#' @rdname S4-filearray
setMethod("/", signature(e1 = "array", e2 = "FileArray"), `op_func_/`)

# register_double_op("^")
`op_func_^` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "^", label = parent_call(deparse = TRUE))
}

#' @rdname S4-filearray
setMethod("^", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "FileArray", e2 = "numeric"), `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "numeric", e2 = "FileArray"), `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "FileArray", e2 = "complex"), `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "complex", e2 = "FileArray"), `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "FileArray", e2 = "logical"), `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "logical", e2 = "FileArray"), `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "FileArray", e2 = "array"), `op_func_^`)

#' @rdname S4-filearray
setMethod("^", signature(e1 = "array", e2 = "FileArray"), `op_func_^`)

# register_double_op("%%")
`op_func_%%` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "%%", label = parent_call(deparse = TRUE))
}

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "FileArray", e2 = "numeric"),      `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "numeric", e2 = "FileArray"),      `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "FileArray", e2 = "complex"),      `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "complex", e2 = "FileArray"),      `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "FileArray", e2 = "logical"),      `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "logical", e2 = "FileArray"),      `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "FileArray", e2 = "array"), `op_func_%%`)

#' @rdname S4-filearray
setMethod("%%", signature(e1 = "array", e2 = "FileArray"), `op_func_%%`)

# register_double_op("%/%")
`op_func_%/%` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "%/%", label = parent_call(deparse = TRUE), out_type = "integer")
}

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "FileArray", e2 = "numeric"),      `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "numeric", e2 = "FileArray"),      `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "FileArray", e2 = "complex"),      `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "complex", e2 = "FileArray"),      `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "FileArray", e2 = "logical"),      `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "logical", e2 = "FileArray"),      `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "FileArray", e2 = "array"), `op_func_%/%`)

#' @rdname S4-filearray
setMethod("%/%", signature(e1 = "array", e2 = "FileArray"), `op_func_%/%`)


# Compare
# register_double_op("==")
`op_func_==` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "==", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod("==", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "FileArray", e2 = "numeric"),      `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "numeric", e2 = "FileArray"),      `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "FileArray", e2 = "complex"),      `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "complex", e2 = "FileArray"),      `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "FileArray", e2 = "logical"),      `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "logical", e2 = "FileArray"),      `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "FileArray", e2 = "array"), `op_func_==`)

#' @rdname S4-filearray
setMethod("==", signature(e1 = "array", e2 = "FileArray"), `op_func_==`)

# register_double_op(">")
`op_func_>` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = ">", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod(">", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "FileArray", e2 = "numeric"), `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "numeric", e2 = "FileArray"), `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "FileArray", e2 = "complex"), `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "complex", e2 = "FileArray"), `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "FileArray", e2 = "logical"), `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "logical", e2 = "FileArray"), `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "FileArray", e2 = "array"), `op_func_>`)

#' @rdname S4-filearray
setMethod(">", signature(e1 = "array", e2 = "FileArray"), `op_func_>`)

# register_double_op("<")
`op_func_<` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "<", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod("<", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "FileArray", e2 = "numeric"), `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "numeric", e2 = "FileArray"), `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "FileArray", e2 = "complex"), `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "complex", e2 = "FileArray"), `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "FileArray", e2 = "logical"), `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "logical", e2 = "FileArray"), `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "FileArray", e2 = "array"), `op_func_<`)

#' @rdname S4-filearray
setMethod("<", signature(e1 = "array", e2 = "FileArray"), `op_func_<`)

# register_double_op("!=")
`op_func_!=` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "!=", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "FileArray", e2 = "numeric"),      `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "numeric", e2 = "FileArray"),      `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "FileArray", e2 = "complex"),      `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "complex", e2 = "FileArray"),      `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "FileArray", e2 = "logical"),      `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "logical", e2 = "FileArray"),      `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "FileArray", e2 = "array"), `op_func_!=`)

#' @rdname S4-filearray
setMethod("!=", signature(e1 = "array", e2 = "FileArray"), `op_func_!=`)

# register_double_op(">=")
`op_func_>=` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = ">=", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "FileArray", e2 = "numeric"),      `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "numeric", e2 = "FileArray"),      `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "FileArray", e2 = "complex"),      `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "complex", e2 = "FileArray"),      `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "FileArray", e2 = "logical"),      `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "logical", e2 = "FileArray"),      `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "FileArray", e2 = "array"), `op_func_>=`)

#' @rdname S4-filearray
setMethod(">=", signature(e1 = "array", e2 = "FileArray"), `op_func_>=`)

# register_double_op("<=")
`op_func_<=` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "<=", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "FileArray", e2 = "numeric"),      `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "numeric", e2 = "FileArray"),      `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "FileArray", e2 = "complex"),      `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "complex", e2 = "FileArray"),      `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "FileArray", e2 = "logical"),      `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "logical", e2 = "FileArray"),      `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "FileArray", e2 = "array"), `op_func_<=`)

#' @rdname S4-filearray
setMethod("<=", signature(e1 = "array", e2 = "FileArray"), `op_func_<=`)


# Logic
# register_double_op("&")
`op_func_&` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "&", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod("&", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "FileArray", e2 = "numeric"), `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "numeric", e2 = "FileArray"), `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "FileArray", e2 = "complex"), `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "complex", e2 = "FileArray"), `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "FileArray", e2 = "logical"), `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "logical", e2 = "FileArray"), `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "FileArray", e2 = "array"), `op_func_&`)

#' @rdname S4-filearray
setMethod("&", signature(e1 = "array", e2 = "FileArray"), `op_func_&`)

# register_double_op("|")
`op_func_|` <- function(e1, e2) {
    fa_pairwise_operator(e1, e2, op = "|", label = parent_call(deparse = TRUE), out_type = "logical")
}

#' @rdname S4-filearray
setMethod("|", signature(e1 = "FileArray", e2 = "FileArray"),      `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "FileArray", e2 = "numeric"), `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "numeric", e2 = "FileArray"), `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "FileArray", e2 = "complex"), `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "complex", e2 = "FileArray"), `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "FileArray", e2 = "logical"), `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "logical", e2 = "FileArray"), `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "FileArray", e2 = "array"), `op_func_|`)

#' @rdname S4-filearray
setMethod("|", signature(e1 = "array", e2 = "FileArray"), `op_func_|`)

#' @name fmap
#' @title Map multiple file arrays and save results
#' @description Advanced mapping function for multiple file arrays. \code{fmap}
#' runs the mapping functions and stores the results in file arrays. 
#' \code{fmap2} stores results in memory. This 
#' feature is experimental. There are several constraints to the input. 
#' Failure to meet these constraints may result in undefined results, or 
#' even crashes. Please read Section 'Details' carefully before using 
#' this function.
#' @param x a list of file arrays to map; each element of \code{x} must 
#' share the same dimensions. 
#' @param fun function that takes one list
#' @param .y a file array object, used to save results
#' @param .input_size number of elements to read from each array of \code{x}
#' @param .output_size \code{fun} output vector length
#' @param .buffer_count number of total buffers (chunks) to run
#' @param .simplify whether to apply \code{\link[base]{simplify2array}} to 
#' the result
#' @param ... other arguments passing to \code{fun}
#' @return File array instance \code{.y}
#' @details 
#' 
#' Denote the first argument of \code{fun} as \code{input}, The length 
#' of \code{input} equals the length of \code{x}. The size of each
#' element of \code{input} is defined by \code{.input_size}, except for the
#' last loop. For example, given dimension of each input array as 
#' \eqn{10x10x10x10}, if \code{.input_size=100}, then 
#' \code{length(input[[1]])=100}. The total number of runs equals to
#' \code{length(x[[1]])/100}. If \code{.input_size=300}, then 
#' \code{length(input[[1]])} will be \code{300} except for the last run. 
#' This is because \eqn{10000} cannot be divided by \code{300}. 
#' The element length of the last run will be \code{100}. 
#' 
#' The returned variable length of \code{fun} will be checked by 
#' \code{.output_size}. If the output length exceed \code{.output_size}, 
#' an error will be raised.
#' 
#' Please make sure that \code{length(.y)/length(x[[1]])} equals to 
#' \code{.output_size/.input_size}.
#' 
#' For \code{fmap_element_wise}, the \code{input[[1]]} and output length 
#' must be the consistent.
#' 
#' @examples 
#' 
#' 
#' set.seed(1)
#' x1 <- filearray_create(tempfile(), dimension = c(100,20,3))
#' x1[] <- rnorm(6000)
#' x2 <- filearray_create(tempfile(), dimension = c(100,20,3))
#' x2[] <- rnorm(6000)
#' 
#' # Add two arrays
#' output <- filearray_create(tempfile(), dimension = c(100,20,3))
#' fmap(list(x1, x2), function(input) {
#'     input[[1]] + input[[2]]
#' }, output)
#' 
#' # check
#' range(output[] - (x1[] + x2[]))
#' 
#' output$delete()
#' 
#' # Calculate the maximum of x1/x2 for every 100 elements
#' # total 60 batches/loops (`.buffer_count`)
#' output <- filearray_create(tempfile(), dimension = c(20,3))
#' fmap(list(x1, x2), function(input) {
#'     max(input[[1]] / input[[2]])
#' }, .y = output, .buffer_count = 60)
#' 
#' # check
#' range(output[] - apply(x1[] / x2[], c(2,3), max))
#' 
#' output$delete()
#' 
#' # A large array example
#' if(interactive()) {
#'     x <- filearray_create(tempfile(), dimension = c(287, 100, 301, 4))
#'     dimnames(x) <- list(
#'         Trial = 1:287,
#'         Marker = 1:100,
#'         Time = 1:301,
#'         Location = 1:4
#'     )
#' 
#'     for (i in 1:4) {
#'         x[,,,i] <- runif(8638700)
#'     }
#'     # Step 1:
#'     # for each location, trial, and marker, calibrate (baseline)
#'     # according to first 50 time-points
#' 
#'     output <- filearray_create(tempfile(), dimension = dim(x))
#' 
#'     # baseline-percentage change
#'     fmap(
#'         list(x),
#'         function(input) {
#'             # get locational data
#'             location_data <- input[[1]]
#'             dim(location_data) <- c(287, 100, 301)
#' 
#'             # collapse over first 50 time points for
#'             # each trial, and marker
#'             baseline <- apply(location_data[,,1:50], c(1,2), mean)
#' 
#'             # calibrate
#'             calibrated <- sweep(location_data, c(1,2), baseline,
#'                                 FUN = function(data, bl) {
#'                                     (data / bl - 1) * 100
#'                                 })
#'             return(calibrated)
#'         },
#' 
#'         .y = output,
#' 
#'         # input dimension is 287 x 100 x 301 for each location
#'         # hence 4 loops in total
#'         .buffer_count = 4
#'     )
#' 
#'     # cleanup
#'     x$delete()
#' 
#' }
#' 
#' # cleanup
#' x1$delete()
#' x2$delete()
#' output$delete()
NULL

# DIPSAUS DEBUG START
# x1 <- filearray_create(tempfile(), dimension = c(2,3,12), partition_size = 5L)
# x1[,,] <- rnorm(prod(dim(x1)[1:3]))
# x2 <- filearray_create(tempfile(), dimension = c(4,3,4), partition_size = 3L)
# x2[] <- rnorm(length(x2))
# x3 <- x1 + x1
# # Add two arrays
# y <- filearray_create(tempfile(), dimension = c(3,4))
# x <- list(x1, x2, x3)
# .input_size <- 12
# fun <- function(input) {
#     nc <- length(input[[1]]) / 6
#     rowSums(sapply(input, function(v) {
#         colSums(matrix(v, ncol = nc))
#     }))
# }
# y <- filearray_map_internal(x, fun, y, .input_size = .input_size)
# expected <- rowSums(sapply(x, function(arr) {
#     colSums(matrix(arr[], ncol = 12))
# }))
# print(y[] - expected)
# print(expected - unlist(filearray_map_internal(x, fun, .output_is_filearray = FALSE)))


filearray_map_internal <- function(x, fun, y = NULL, .buffer_count = NA_integer_, 
                                   .output_is_filearray = TRUE, .simplify = FALSE, ...) {
    if (!length(x)) {
        stop("`x` must be a list of file arrays")
    }
    if (is_filearray(x)) {
        x <- list(x)
    }
    
    x <- lapply(x, function(arr) {
        return(as_filearrayproxy(arr))
    })
    
    # Expand x to get all potential filearrays to load
    arrays <- unlist(lapply(x, function(arr) {
        arr$linked_arrays()
    }))
    uuids <- get_inital_uuids(arrays)
    uuids_x <- get_uuids(x)
    is_unique <- !duplicated(uuids)
    
    uuids <- uuids[is_unique]
    arrays <- arrays[is_unique]
    
    fbases <- unname(vapply(arrays, function(el) {
        # el$initialize_partition()
        el$.filebase
    }, "", USE.NAMES = FALSE))
    
    # prepare mapping function
    args <- list(quote(data), ...)
    mapping_func <- function(data) {
        env <- new.env(parent = emptyenv())
        lapply(seq_along(uuids), function(i) {
            env[[ uuids[[i]] ]] <- data[[ i ]]
            return(NULL)
        })
        lapply(x, function(arr) {
            arr$.mature(env, data = data)
        })
        data <- lapply(uuids_x, function(nm) { env[[nm]] })
        return(do.call(fun, args))
    }
    
    if ( .output_is_filearray ) {
        input_types <- vapply(x, typeof, "")
        output_type <- Reduce(operation_output_type, input_types)
        if (missing(y) || is.null(y)) {
            y <- filearray_create(filebase = temp_path(check = TRUE), dimension = dim(x[[1]]), type = output_type)
        } else {
            if (!is_filearray(y)) {
                stop("Output `y` must be a filearray or a proxy array")
            }
        }
        
        all_lens <- c(vapply(arrays, length, 0L), length(y))
        
        if (!validate_fmap_buffer_count(.buffer_count, all_lens)) {
            .buffer_count <- common_fmap_buffer_count(dim(y), .list = lapply(x, dim))
        }
        buffer_count_is_valid <- validate_fmap_buffer_count(.buffer_count, all_lens)
        if (!buffer_count_is_valid) {
            stop("Input & output arrays have inconsistent lengths. Cannot map functions")
        }
        buffer_nelems <- all_lens / .buffer_count
        
        
        y$initialize_partition()
        
        FARR_buffer_map(
            input_filebases = fbases,
            output_filebase = y$.filebase,
            map = mapping_func,
            buffer_nelems = buffer_nelems[-length(buffer_nelems)],
            result_nelems = buffer_nelems[[length(buffer_nelems)]]
        )
        return(y)
    } else {
        all_lens <- vapply(arrays, length, 0L)
        
        if (!validate_fmap_buffer_count(.buffer_count, all_lens)) {
            .buffer_count <- common_fmap_buffer_count(.list = lapply(x, dim))
        }
        buffer_count_is_valid <- validate_fmap_buffer_count(.buffer_count, all_lens)
        if (!buffer_count_is_valid) {
            stop("Input & output arrays have inconsistent lengths. Cannot map functions")
        }
        buffer_nelems <- all_lens / .buffer_count
        
        res <- FARR_buffer_map2(
            input_filebases = fbases,
            map = mapping_func,
            buffer_nelems = buffer_nelems
        )
        if (.simplify) {
            res <- simplify2array(res)
        }
        return(res)
    }
    
    
}


#' @rdname fmap
#' @export
fmap <- function(x, fun, .y = NULL, .buffer_count = NA_integer_, .output_size = NA_integer_, ...) {
    if (!missing(.output_size)) {
        warning("`fmap`: .output_size is deprecated. Please specify `.y` instead")
    }
    filearray_map_internal(x = x, fun = fun, y = .y, .buffer_count = .buffer_count, ...)
}

#' @rdname fmap
#' @export
fmap2 <- function(x, fun, .buffer_count = NA, .simplify = TRUE, ...) {
    
    filearray_map_internal(x = x, fun = fun, .buffer_count = .buffer_count, .output_is_filearray = FALSE, .simplify = .simplify, ...)
}

# only use values on hard disk
fmap_element_wise_internal <- function(x, fun, .y, ..., .input_size = NA) {
    if (!is.list(x)) {
        if (!is_filearray(x)) { stop("`x` must be a list of file arrays") }
        x <- list(x)
    } else {
        lapply(x, function(x) {
            if (!is_filearray(x)) { stop("`x` must be a list of file arrays") }
        })
    }
    if (!length(x)) { stop("`x` length must be positive") }
    dims <- sapply(x, dim)
    dim <- dims[, 1, drop = TRUE]
    
    if (any(dims - dim != 0)) {
        stop("Input `x` array dimensions must match")
    }
    
    miss_y <- missing(.y) || is.null(.y)
    if (miss_y) {
        .y <- filearray_create(temp_path(), dim, type = typeof(x[[1]]))
    } else {
        if (length(dim) != length(.y$dimension()) ||
           any(.y$dimension() - dim != 0)) {
            stop("Dimensions of x[[1]] and .y mismatch")
        }
    }
    .y$initialize_partition()
    
    if (is.na(.input_size)) {
        .input_size <- guess_fmap_buffer_size(dim(.y), .y$element_size())
        # .input_size <- get_buffer_size() / .y$element_size()
    }
    if (.input_size <= 0) {
        stop("`.input_size` must be postive")
    }
    .input_size <- as.integer(.input_size)
    
    fbases <- sapply(x, function(el) {
        if ( !inherits(el, "FileArray") ) {
            stop("Input `x` must only contains file arrays")
        }
        el$initialize_partition()
        el$.filebase
    })
    
    args <- list(quote(input), ...)
    map <- function(input) {
        do.call(fun, args)
    }
    
    if (length(x[[1]]) > length(.y) ) {
        stop("Inconsistent input and output length")
    }
    
    FARR_buffer_map(
        input_filebases = fbases,
        output_filebase = .y$.filebase,
        map = map,
        buffer_nelems = rep(.input_size, length(x)),
        result_nelems = .input_size,
        ...
    )
    # fmap(xs, fun = fun, .y = .y, .input_size = buf_size, .output_size = buf_size, ...)
    if (miss_y) {
        return(.y)
    } else {
        return(invisible(.y))
    }
}

#' @rdname fmap
#' @export
fmap_element_wise <- function(x, fun, .y, ..., .input_size = NA) {
    if (!is_fileproxy(x)) {
        return(fmap_element_wise_internal(x = x, fun = fun, .y = .y, ..., .input_size = .input_size))
    } else {
        # proxy!
        if (missing(.y) || is.null(.y)) {
            filebase <- NULL
        } else {
            filebase <- .y$.filebase
        }
        fa_eval_ops(
            x = x, filebase = filebase, use_cache = FALSE,
            addon = function(env, dlist, uuid) {
                env[[ uuid ]] <- fun(dlist)
            }
        )
        
    }
}

# misc internal functions that are byte-compiled

# move here to byte-compile
.max_mapper1 <- function(data, size, idx) {
    if (!length(size)) {
        return(numeric(0))
    }
    max(data, na.rm = TRUE)
}
.max_mapper2 <- function(data, size, idx) {
    if (!length(size)) {
        return(numeric(0))
    }
    if (length(data) != size) {
        data <- data[seq_len(size)]
    }
    max(data, na.rm = FALSE)
}
.min_mapper1 <- function(data, size, idx) {
    if (!length(size)) {
        return(numeric(0))
    }
    min(data, na.rm = TRUE)
}
.min_mapper2 <- function(data, size, idx) {
    if (!length(size)) {
        return(numeric(0))
    }
    if (length(data) != size) {
        data <- data[seq_len(size)]
    }
    min(data, na.rm = FALSE)
}
.range_mapper1 <- function(data, size, idx) {
    if (!length(size)) {
        return(numeric(0))
    }
    range(data, na.rm = TRUE)
}
.range_mapper2 <- function(data, size, idx) {
    if (!length(size)) {
        return(numeric(0))
    }
    if (length(data) != size) {
        data <- data[seq_len(size)]
    }
    range(data, na.rm = FALSE)
}

.sum_mapper1 <- function(data, size, idx) {
    if (!length(size)) {
        return(0)
    }
    sum(data, na.rm = TRUE)
}
.sum_mapper2 <- function(data, size, idx) {
    if (!length(size)) {
        return(0)
    }
    if (length(data) != size) {
        data <- data[seq_len(size)]
    }
    sum(data, na.rm = FALSE)
}

---
title: "Performance Comparisons - (Numerical)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Performance Comparisons - (Numerical)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  eval = FALSE,
  comment = "#>"
)
```

This vignette profiles `FileArray` operations and compares with R native functions. The goal is to

1. Benchmark different ways to operate on file arrays (write, read, subset, transform, ...)
2. Compare with other methods to see what file array can and cannot do.

The simulation was performed on `MacBook Air 2020 (M1 Chip, ARM, 8GB RAM)`, with R `4.1.0`. To reproduce the results, please install `CRAN` packages `dipsaus` and `microbenchmark`.


## Setup

We mainly test the performance of `double` and `float` data type. The dimensions for both arrays are `100x100x100x100`. Both arrays are around `800MB` in native R. This is because R does not have float precision. However, while `double` array occupies `800MB` space on the hard disk, `float` array only uses half size (`400MB`).

```{r setup, eval = FALSE}
library(filearray)

options(digits = 3)
filearray_threads()
#> [1] 8

# Create file array and initialize partitions
set.seed(1)
file <- tempfile()
unlink(file, recursive = TRUE)
x_dbl <- filearray_create(file, rep(100, 4))
x_dbl$initialize_partition()

file <- tempfile()
unlink(file, recursive = TRUE)
x_flt <- filearray_create(file, rep(100, 4), type = "float")
x_flt$initialize_partition()

# 800 MB double array
y <- array(rnorm(length(x_dbl)), dim(x_dbl))
```

## Simulation

The simulation contains

* [Write speed](#write)
  - [Writing along margin](#write-along-margin)
  - [Writing chunks of data](#write-chunks-of-data)
* [Read speed](#read)
  - [Read all the data](#read-the-whole-array)
  - [Read along margins](#read-along-margins)
  - [Random access](#random-access)
* [Collapse](#collapse)

## Write

### 1. Write along margin

Writing along margins refer to something like `x[,,,i] <- v` (along the last margin), or `x[,i,,] <- v` (along the second margin). It is always recommended to write along the last margin, and always discouraged to write along the first margin to file arrays.

1.  partition margin

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    double = {
        for (i in 1:100) {
            x_dbl[, , , i] <- y[, , , i]
        }
    },
    float = {
        for (i in 1:100) {
            x_flt[, , , i] <- y[, , , i]
        }
    },
    unit = "s",
    times = 3
)

#> Unit: seconds
#>    expr   min    lq mean median   uq  max neval
#>  double 0.933 0.935 1.44  0.936 1.69 2.45     3
#>   float 1.027 1.057 1.07  1.086 1.10 1.11     3
```

2. Write along fast margin

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    double = {
        for (i in 1:100) {
            x_dbl[, , i, ] <- y[, , i, ]
        }
    },
    float = {
        for (i in 1:100) {
            x_flt[, , i, ] <- y[, , i, ]
        }
    },
    unit = "s",
    times = 3
)

#> Unit: seconds
#>   expr  min   lq mean median   uq  max neval
#> double 1.23 1.27 1.47   1.30 1.59 1.89     3
#>  float 1.23 1.24 1.41   1.24 1.50 1.76     3
```

3. Writing along slow margin

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    double = {
        for (i in 1:100) {
            x_dbl[i, , , ] <- y[i, , , ]
        }
    },
    float = {
        for (i in 1:100) {
            x_flt[i, , , ] <- y[i, , , ]
        }
    },
    unit = "s",
    times = 3
)

#> Unit: seconds
#>    expr   min    lq  mean median    uq   max neval
#>  double  3.18  3.22  3.28   3.27  3.32  3.38     3
#>   float 20.04 20.04 20.44  20.05 20.64 21.22     3
```

In the current version, converting from `double` to `float` introduces overhead that delays the procedure.

### 2. Write chunks of data

Instead of writing one slice at a time along each margin, we write `100x100x100x5` (10 slices) each time.

1.  Write blocks of data along the partition margin

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    double = {
        for (i in 1:10) {
            idx <- (i - 1) * 10 + 1:10
            x_dbl[, , , idx] <- y[, , , idx]
        }
    },
    float = {
        for (i in 1:10) {
            idx <- (i - 1) * 10 + 1:10
            x_flt[, , , idx] <- y[, , , idx]
        }
    },
    unit = "s",
    times = 3
)

#> Unit: seconds
#>    expr   min    lq  mean median    uq  max neval
#>  double 0.650 0.684 0.911  0.718 1.041 1.37     3
#>   float 0.626 0.662 0.783  0.698 0.861 1.02     3
```

2. Write blocks of data along the fast margin

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    double = {
        for (i in 1:10) {
            idx <- (i - 1) * 10 + 1:10
            x_dbl[, , idx, ] <- y[, , idx, ]
        }
    },
    float = {
        for (i in 1:10) {
            idx <- (i - 1) * 10 + 1:10
            x_flt[, , idx, ] <- y[, , idx, ]
        }
    },
    unit = "s",
    times = 3
)

#> Unit: seconds
#>    expr   min    lq  mean median    uq   max neval
#>  double 0.582 0.620 0.668  0.657 0.710 0.763     3
#>   float 0.625 0.652 0.732  0.679 0.786 0.893     3
```

3. Write blocks of data along slow margin

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    double = {
        for (i in 1:10) {
            idx <- (i - 1) * 10 + 1:10
            x_dbl[idx, , , ] <- y[idx, , , ]
        }
    },
    float = {
        for (i in 1:10) {
            idx <- (i - 1) * 10 + 1:10
            x_flt[idx, , , ] <- y[idx, , , ]
        }
    },
    unit = "s",
    times = 3
)

#> Unit: seconds
#>    expr  min   lq mean median   uq  max neval
#>  double 4.48 4.48 4.64   4.48 4.72 4.95     3
#>   float 2.64 2.70 2.73   2.77 2.78 2.79     3
```


## Read 

### 1. Read the whole array

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    double = {
        x_dbl[]
    },
    float = {
        x_flt[]
    },
    unit = "s",
    times = 3
)

#> Unit: seconds
#>    expr   min    lq  mean median    uq   max neval
#>  double 0.155 0.172 0.185  0.188 0.200 0.211     3
#>   float 0.104 0.106 0.144  0.107 0.164 0.220     3
```

### 2. Read along margins

```{r, eval = FALSE}
microbenchmark::microbenchmark(
    farr_double_partition_margin = {
        x_dbl[, , , 1]
    },
    farr_double_fast_margin = {
        x_dbl[, , 1, ]
    },
    farr_double_slow_margin = {
        x_dbl[1, , , ]
    },
    farr_float_partition_margin = {
        x_flt[, , , 1]
    },
    farr_float_fast_margin = {
        x_flt[, , 1, ]
    },
    farr_float_slow_margin = {
        x_flt[1, , , ]
    },
    native_partition_margin = {
        y[, , , 1]
    },
    native_fast_margin = {
        y[, , 1, ]
    },
    native_slow_margin = {
        y[1, , , ]
    },
    times = 100L,
    unit = "ms"
)

#> Unit: milliseconds
#>                          expr   min    lq  mean median    uq    max neval
#>  farr_double_partition_margin  2.01  2.66  4.02   2.85  3.64  71.06   100
#>       farr_double_fast_margin  1.35  1.99  3.16   2.35  3.79  25.88   100
#>       farr_double_slow_margin 33.25 36.52 44.11  37.32 38.76 125.61   100
#>   farr_float_partition_margin  1.77  2.40  3.96   2.61  3.66  58.17   100
#>        farr_float_fast_margin  1.33  1.85  2.80   2.08  3.43  11.01   100
#>        farr_float_slow_margin 14.98 18.86 23.42  19.54 20.47 160.90   100
#>       native_partition_margin  3.42  3.75  4.14   4.02  4.27   6.89   100
#>            native_fast_margin  3.42  3.96  4.86   4.09  4.64  54.74   100
#>            native_slow_margin 21.52 22.15 24.34  22.65 23.97  91.06   100
```

The file array indexing is close to handling in-memory arrays in R!

### 3. Random access

```{r, eval = FALSE}
# access 50 x 50 x 50 x 50 sub-array, with random indices
idx1 <- sample(1:100, 50)
idx2 <- sample(1:100, 50)
idx3 <- sample(1:100, 50)
idx4 <- sample(1:100, 50)

microbenchmark::microbenchmark(
    farr_double = {
        x_dbl[idx1, idx2, idx3, idx4]
    },
    farr_float = {
        x_flt[idx1, idx2, idx3, idx4]
    },
    native = {
        y[idx1, idx2, idx3, idx4]
    },
    times = 100L,
    unit = "ms"
)

#> Unit: milliseconds
#>         expr   min    lq mean median   uq   max neval
#>  farr_double 11.68 13.13 18.9  13.81 15.2 143.3   100
#>   farr_float  8.29  8.89 12.0   9.95 10.6  63.6   100
#>       native 30.86 31.94 34.0  32.62 33.1 103.0   100
```

Random access could be faster than base R (also much less memory!)

## Collapse

Collapse calculates the margin sum/mean. Collapse function in `filearray` uses single thread. This is because the bottle-neck often comes from hard-disk accessing speed. However, it is still faster than native R, and is more memory-efficient.

```{r, eval = FALSE}
keep <- c(2, 4)
output <- filearray_create(tempfile(), dim(x_dbl)[keep])
output$initialize_partition()
microbenchmark::microbenchmark(
    farr_double = {
        x_dbl$collapse(keep = keep, method = "sum")
    },
    farr_float = {
        x_flt$collapse(keep = keep, method = "sum")
    },
    native = {
        apply(y, keep, sum)
    },
    ravetools = {
        ravetools::collapse(y, keep, average = FALSE)
    },
    unit = "s",
    times = 5
)

#> Unit: seconds
#>         expr   min    lq  mean median    uq   max neval
#>  farr_double 0.651 0.666 0.867  0.716 0.718 1.583     5
#>   farr_float 0.628 0.637 0.737  0.642 0.652 1.124     5
#>       native 1.011 1.029 1.128  1.078 1.207 1.316     5
#>    ravetools 0.109 0.110 0.126  0.131 0.138 0.139     5
```

The `ravetools` package uses multiple threads to collapse arrays in-memory. It is `7~8x` as fast as base R. File array is `1.5~2x` as fast as base R. Both `ravetools` and `filearray` have little memory over-heads.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind.R
\name{filearray_bind}
\alias{filearray_bind}
\title{Merge and bind homogeneous file arrays}
\usage{
filearray_bind(
  ...,
  .list = list(),
  filebase = tempfile(),
  symlink = FALSE,
  overwrite = FALSE,
  cache_ok = FALSE
)
}
\arguments{
\item{..., .list}{file array instances}

\item{filebase}{where to create merged array}

\item{symlink}{whether to use \code{\link[base]{file.symlink}}; if true,
then partition files will be symbolic-linked to the original arrays,
otherwise the partition files will be copied over. If you want your data
to be portable, do not use symbolic-links. The default value is \code{FALSE}}

\item{overwrite}{whether to overwrite when \code{filebase} already exists;
default is false, which raises errors}

\item{cache_ok}{see 'Details', only used if \code{overwrite} is true.}
}
\value{
A bound array in \code{'FileArray'} class.
}
\description{
The file arrays to be merged must be homogeneous:
same data type, partition size, and partition length
}
\details{
The input arrays must share the same data type and partition size.
The dimension for each partition should also be the same. For example
an array \code{x1} has dimension \eqn{100x20x30} with partition size 
\code{1}, then each partition dimension is \eqn{100x20x1}, and there are 
\code{30} partitions. \code{x1} can bind with another array of the same 
partition size. This means if \code{x2} has dimension
\eqn{100x20x40} and each partition size is \code{1}, then \code{x1} and 
\code{x2} can be merged.

If \code{filebase} exists and \code{overwrite} is \code{FALSE}, an error will 
always raise. If \code{overwrite=TRUE} and \code{cache_ok=FALSE}, then
the existing \code{filebase} will be erased and any data stored within will
be lost. 
If both \code{overwrite} and \code{cache_ok} are \code{TRUE}, then 
, before erasing \code{filebase}, the function validates the existing
array header and compare the header signatures. If the existing header
signature is the same as the array to be created, then the existing array 
will be returned. This \code{cache_ok} could be extremely useful when
binding large arrays with \code{symlink=FALSE} as the cache might avoid
moving files around. However, \code{cache_ok} should be enabled with caution.
This is because only the header information will be compared, but the 
partition data will not be compared. If the existing array was generated from
an old versions of the source arrays, but the data from the source arrays
has been altered, then the \code{cache_ok=TRUE} is rarely proper as the cache
is outdated.

The \code{symlink} option should be used with extra caution. Creating 
symbolic links is definitely faster than copying partition files. However, 
since the partition files are simply linked to the original partition files, 
changing to the input arrays will also affect the merged arrays, and 
vice versa; see 'Examples'. Also for arrays created from symbolic links, if 
the original 
arrays are deleted, while the merged arrays will not be invalidated, 
the corresponding partitions will no longer be accessible. Attempts to 
set deleted partitions will likely result in failure. Therefore
\code{symlink} should be set to true when creating merged arrays are
temporary for read-only purpose, and when speed and disk space is in
consideration. For extended reading, please check \code{\link[base]{files}} 
for details.
}
\examples{
partition_size <- 1
type <- "double"
x1 <- filearray_create(
    tempfile(), c(2,2), type = type,
    partition_size = partition_size)
x1[] <- 1:4
x2 <- filearray_create(
    tempfile(), c(2,1), type = type,
    partition_size = partition_size)
x2[] <- 5:6

y1 <- filearray_bind(x1, x2, symlink = FALSE)
y2 <- filearray_bind(x1, x2)

# y1 copies partition files, and y2 simply creates links 
# if symlink is supported

y1[] - y2[]

# change x1
x1[1,1] <- NA

# y1 is not affected
y1[]

# y2 changes 
y2[]


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc-docs.R
\name{filearray_threads}
\alias{filearray_threads}
\title{Set or get file array threads}
\usage{
filearray_threads(n, ...)
}
\arguments{
\item{n}{number of threads to set. If \code{n} is negative,
then default to the number of cores that computer has.}

\item{...}{internally used}
}
\value{
An integer of current number of threads
}
\description{
Will enable/disable multi-threaded reading or writing
at \code{C++} level.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-filearray.R, R/class-proxy.R, R/load.R
\docType{class}
\name{as_filearray}
\alias{as_filearray}
\alias{FileArrayProxy-class}
\alias{FileArrayProxy}
\alias{as_filearrayproxy}
\alias{filearray}
\alias{filearray_create}
\alias{filearray_load}
\alias{filearray_checkload}
\alias{filearray_load_or_create}
\title{Create or load existing file arrays}
\usage{
as_filearray(x, ...)

as_filearrayproxy(x, ...)

filearray_create(
  filebase,
  dimension,
  type = c("double", "float", "integer", "logical", "raw", "complex"),
  partition_size = NA,
  initialize = FALSE,
  ...
)

filearray_load(filebase, mode = c("readwrite", "readonly"))

filearray_checkload(
  filebase,
  mode = c("readonly", "readwrite"),
  ...,
  symlink_ok = TRUE
)

filearray_load_or_create(
  filebase,
  dimension,
  on_missing = NULL,
  type = NA,
  ...,
  mode = c("readonly", "readwrite"),
  symlink_ok = TRUE,
  initialize = FALSE,
  partition_size = NA,
  verbose = FALSE,
  auto_set_headers = TRUE
)
}
\arguments{
\item{x}{R object such as array, file array proxy, or character that can be 
transformed into file array}

\item{...}{additional headers to check used by \code{filearray_checkload}
(see 'Details'). This argument is ignored by \code{filearray_create}, 
reserved for future compatibility.}

\item{filebase}{a directory path to store arrays in the local file 
system. When creating an array, the path must not exist.}

\item{dimension}{dimension of the array, at least length of 2}

\item{type}{storage type of the array; default is \code{'double'}. Other
options include \code{'integer'}, \code{'logical'}, and \code{'raw'}.}

\item{partition_size}{positive partition size for the last margin, or
\code{NA} to automatically guess; see 'Details'.}

\item{initialize}{whether to initialize partition files; default is false
for performance considerations. However, if the array is dense, it is 
recommended to set to true}

\item{mode}{whether allows writing to the file; choices are 
\code{'readwrite'} and \code{'readonly'}.}

\item{symlink_ok}{whether arrays with symbolic-link partitions can pass 
the test; this is usually used on bound arrays with symbolic-links; see 
\code{\link{filearray_bind}};}

\item{on_missing}{function to handle file array (such as initialization)
when a new array is created; must take only one argument, the array object}

\item{verbose}{whether to print out some debug messages}

\item{auto_set_headers}{whether to automatically set headers if array is 
missing or to be created; default is true}
}
\value{
A \code{\link{FileArray-class}} instance.
}
\description{
Create or load existing file arrays
}
\details{
The file arrays partition out-of-memory array objects and store them 
separately in local file systems. Since R stores matrices/arrays 
in column-major style, file array uses the slowest margin (the 
last margin) to slice the partitions. This helps to align the elements
within the files with the corresponding memory order. An array with 
dimension \code{100x200x300x400} has 4 margins. The length of the 
last margin is 400, which is also the maximum number of potential
partitions. The number of partitions are determined by the last margin
size divided by \code{partition_size}. For example, if the partition
size is 1, then there will be 400 partitions. If the partition size 
if 3, there will be 134 partitions. The default partition sizes 
are determined internally following these priorities:
\describe{
\item{1. }{the file size of each partition does not exceed \code{1GB}}
\item{2. }{the number of partitions do not exceed 100}
}
These two rules are not hard requirements. The goal is to reduce the
numbers of partitions as much as possible. 

The arguments \code{...} in \code{filearray_checkload} should be named
arguments that provide additional checks for the header information. 
The check will fail if at least one header is not identical. For example,
if an array contains header key-signature pair, one can use 
\code{filearray_checkload(..., key = signature)} to validate the signature.
Note the comparison will be rigid, meaning the storage type of the headers 
will be considered as well. If the signature stored in the array is an 
integer while provided is a double, then the check will result in failure.
}

\examples{


# Prepare
library(filearray)
filebase <- tempfile()
if(file.exists(filebase)) { unlink(filebase, TRUE) }

# create array
x <- filearray_create(filebase, dimension = c(200, 30, 8))
print(x)

# Assign values
x[] <- rnorm(48000)

# Subset
x[1,2,]

# load existing array
filearray_load(filebase)

x$set_header("signature", "tom")
filearray_checkload(filebase, signature = "tom")

\dontrun{
# Trying to load with wrong signature
filearray_checkload(filebase, signature = "jerry")
}


# check-load, and create a new array if fail
x <- filearray_load_or_create(
    filebase = filebase, dimension = c(200, 30, 8),
    verbose = FALSE, signature = "henry"
)
x$get_header("signature")

# check-load with initialization
x <- filearray_load_or_create(
    filebase = filebase,
    dimension = c(3, 4, 5),
    verbose = FALSE, mode = "readonly",
    on_missing = function(array) {
        array[] <- seq_len(60)
    }
)

x[1:3,1,1]

# Clean up
unlink(filebase, recursive = TRUE)

}
\author{
Zhengjia Wang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class-filearray.R
\name{FileArray-class}
\alias{FileArray-class}
\title{Definition of file array}
\description{
\code{S4} class definition of \code{FileArray}. Please 
use \code{\link{filearray_create}} and \code{\link{filearray_load}}
to create instances.
}
\section{Public Methods}{

\describe{
\item{\code{get_header(key, default = NULL)}}{Get header information; returns \code{default} if \code{key} is missing}
\item{\code{set_header(key, value)}}{Set header information; the extra headers will be stored in meta file. Please do not store large headers as they will be loaded into memory frequently.}
\item{\code{can_write()}}{Whether the array data can be altered}
\item{\code{create(filebase, dimension, type = "double", partition_size = 1)}}{Create a file array instance}
\item{\code{delete(force = FALSE)}}{Remove array from local file system and reset}
\item{\code{dimension()}}{Get dimension vector}
\item{\code{dimnames(v)}}{Set/get dimension names}
\item{\code{element_size()}}{Internal storage: bytes per element}
\item{\code{fill_partition(part, value)}}{Fill a partition with given scalar}
\item{\code{get_partition(part, reshape = NULL)}}{Get partition data, and reshape (if not null) to desired dimension}
\item{\code{expand(n)}}{Expand array along the last margin; returns true if expanded; if the \code{dimnames} have been assigned prior to expansion, the last dimension names will be filled with \code{NA}}
\item{\code{initialize_partition()}}{Make sure a partition file exists; if not, create one and fill with \code{NA}s or 0 (\code{type='raw'})}
\item{\code{load(filebase, mode = c("readwrite", "readonly"))}}{Load file array from existing directory}
\item{\code{partition_path(part)}}{Get partition file path}
\item{\code{partition_size()}}{Get partition size; see \code{\link{filearray}}}
\item{\code{set_partition(part, value, ..., strict = TRUE)}}{Set partition value}
\item{\code{sexp_type()}}{Get data \code{SEXP} type; see R internal manuals}
\item{\code{show()}}{Print information}
\item{\code{type()}}{Get data type}
\item{\code{valid()}}{Check if the array is valid.}
}
}

\seealso{
\code{\link{filearray}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-typeof.R
\name{typeof}
\alias{typeof}
\alias{typeof,FileArray-method}
\alias{typeof,FileArrayProxy-method}
\title{The type of a file array (extended)}
\usage{
typeof(x)

\S4method{typeof}{FileArray}(x)

\S4method{typeof}{FileArrayProxy}(x)
}
\arguments{
\item{x}{any file array}
}
\value{
A character string. The possible values are \code{"double"},
\code{"integer"}, \code{"logical"}, and \code{"raw"}
}
\description{
The type of a file array (extended)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/method-map.R
\name{fmap}
\alias{fmap}
\alias{fmap2}
\alias{fmap_element_wise}
\title{Map multiple file arrays and save results}
\usage{
fmap(
  x,
  fun,
  .y = NULL,
  .buffer_count = NA_integer_,
  .output_size = NA_integer_,
  ...
)

fmap2(x, fun, .buffer_count = NA, .simplify = TRUE, ...)

fmap_element_wise(x, fun, .y, ..., .input_size = NA)
}
\arguments{
\item{x}{a list of file arrays to map; each element of \code{x} must 
share the same dimensions.}

\item{fun}{function that takes one list}

\item{.y}{a file array object, used to save results}

\item{.buffer_count}{number of total buffers (chunks) to run}

\item{.output_size}{\code{fun} output vector length}

\item{...}{other arguments passing to \code{fun}}

\item{.simplify}{whether to apply \code{\link[base]{simplify2array}} to 
the result}

\item{.input_size}{number of elements to read from each array of \code{x}}
}
\value{
File array instance \code{.y}
}
\description{
Advanced mapping function for multiple file arrays. \code{fmap}
runs the mapping functions and stores the results in file arrays. 
\code{fmap2} stores results in memory. This 
feature is experimental. There are several constraints to the input. 
Failure to meet these constraints may result in undefined results, or 
even crashes. Please read Section 'Details' carefully before using 
this function.
}
\details{
Denote the first argument of \code{fun} as \code{input}, The length 
of \code{input} equals the length of \code{x}. The size of each
element of \code{input} is defined by \code{.input_size}, except for the
last loop. For example, given dimension of each input array as 
\eqn{10x10x10x10}, if \code{.input_size=100}, then 
\code{length(input[[1]])=100}. The total number of runs equals to
\code{length(x[[1]])/100}. If \code{.input_size=300}, then 
\code{length(input[[1]])} will be \code{300} except for the last run. 
This is because \eqn{10000} cannot be divided by \code{300}. 
The element length of the last run will be \code{100}. 

The returned variable length of \code{fun} will be checked by 
\code{.output_size}. If the output length exceed \code{.output_size}, 
an error will be raised.

Please make sure that \code{length(.y)/length(x[[1]])} equals to 
\code{.output_size/.input_size}.

For \code{fmap_element_wise}, the \code{input[[1]]} and output length 
must be the consistent.
}
\examples{


set.seed(1)
x1 <- filearray_create(tempfile(), dimension = c(100,20,3))
x1[] <- rnorm(6000)
x2 <- filearray_create(tempfile(), dimension = c(100,20,3))
x2[] <- rnorm(6000)

# Add two arrays
output <- filearray_create(tempfile(), dimension = c(100,20,3))
fmap(list(x1, x2), function(input) {
    input[[1]] + input[[2]]
}, output)

# check
range(output[] - (x1[] + x2[]))

output$delete()

# Calculate the maximum of x1/x2 for every 100 elements
# total 60 batches/loops (`.buffer_count`)
output <- filearray_create(tempfile(), dimension = c(20,3))
fmap(list(x1, x2), function(input) {
    max(input[[1]] / input[[2]])
}, .y = output, .buffer_count = 60)

# check
range(output[] - apply(x1[] / x2[], c(2,3), max))

output$delete()

# A large array example
if(interactive()) {
    x <- filearray_create(tempfile(), dimension = c(287, 100, 301, 4))
    dimnames(x) <- list(
        Trial = 1:287,
        Marker = 1:100,
        Time = 1:301,
        Location = 1:4
    )

    for (i in 1:4) {
        x[,,,i] <- runif(8638700)
    }
    # Step 1:
    # for each location, trial, and marker, calibrate (baseline)
    # according to first 50 time-points

    output <- filearray_create(tempfile(), dimension = dim(x))

    # baseline-percentage change
    fmap(
        list(x),
        function(input) {
            # get locational data
            location_data <- input[[1]]
            dim(location_data) <- c(287, 100, 301)

            # collapse over first 50 time points for
            # each trial, and marker
            baseline <- apply(location_data[,,1:50], c(1,2), mean)

            # calibrate
            calibrated <- sweep(location_data, c(1,2), baseline,
                                FUN = function(data, bl) {
                                    (data / bl - 1) * 100
                                })
            return(calibrated)
        },

        .y = output,

        # input dimension is 287 x 100 x 301 for each location
        # hence 4 loops in total
        .buffer_count = 4
    )

    # cleanup
    x$delete()

}

# cleanup
x1$delete()
x2$delete()
output$delete()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa-docs.R, R/methods-add.R,
%   R/methods-simple.R
\name{S4-filearray}
\alias{S4-filearray}
\alias{+,FileArray,FileArray-method}
\alias{+,FileArray,numeric-method}
\alias{+,numeric,FileArray-method}
\alias{+,FileArray,complex-method}
\alias{+,complex,FileArray-method}
\alias{+,FileArray,logical-method}
\alias{+,logical,FileArray-method}
\alias{+,FileArray,array-method}
\alias{+,array,FileArray-method}
\alias{-,FileArray,FileArray-method}
\alias{-,FileArray,numeric-method}
\alias{-,numeric,FileArray-method}
\alias{-,FileArray,complex-method}
\alias{-,complex,FileArray-method}
\alias{-,FileArray,logical-method}
\alias{-,logical,FileArray-method}
\alias{-,FileArray,array-method}
\alias{-,array,FileArray-method}
\alias{*,FileArray,FileArray-method}
\alias{*,FileArray,numeric-method}
\alias{*,numeric,FileArray-method}
\alias{*,FileArray,complex-method}
\alias{*,complex,FileArray-method}
\alias{*,FileArray,logical-method}
\alias{*,logical,FileArray-method}
\alias{*,FileArray,array-method}
\alias{*,array,FileArray-method}
\alias{/,FileArray,FileArray-method}
\alias{/,FileArray,numeric-method}
\alias{/,numeric,FileArray-method}
\alias{/,FileArray,complex-method}
\alias{/,complex,FileArray-method}
\alias{/,FileArray,logical-method}
\alias{/,logical,FileArray-method}
\alias{/,FileArray,array-method}
\alias{/,array,FileArray-method}
\alias{^,FileArray,FileArray-method}
\alias{^,FileArray,numeric-method}
\alias{^,numeric,FileArray-method}
\alias{^,FileArray,complex-method}
\alias{^,complex,FileArray-method}
\alias{^,FileArray,logical-method}
\alias{^,logical,FileArray-method}
\alias{^,FileArray,array-method}
\alias{^,array,FileArray-method}
\alias{\%\%,FileArray,FileArray-method}
\alias{\%\%,FileArray,numeric-method}
\alias{\%\%,numeric,FileArray-method}
\alias{\%\%,FileArray,complex-method}
\alias{\%\%,complex,FileArray-method}
\alias{\%\%,FileArray,logical-method}
\alias{\%\%,logical,FileArray-method}
\alias{\%\%,FileArray,array-method}
\alias{\%\%,array,FileArray-method}
\alias{\%/\%,FileArray,FileArray-method}
\alias{\%/\%,FileArray,numeric-method}
\alias{\%/\%,numeric,FileArray-method}
\alias{\%/\%,FileArray,complex-method}
\alias{\%/\%,complex,FileArray-method}
\alias{\%/\%,FileArray,logical-method}
\alias{\%/\%,logical,FileArray-method}
\alias{\%/\%,FileArray,array-method}
\alias{\%/\%,array,FileArray-method}
\alias{==,FileArray,FileArray-method}
\alias{==,FileArray,numeric-method}
\alias{==,numeric,FileArray-method}
\alias{==,FileArray,complex-method}
\alias{==,complex,FileArray-method}
\alias{==,FileArray,logical-method}
\alias{==,logical,FileArray-method}
\alias{==,FileArray,array-method}
\alias{==,array,FileArray-method}
\alias{>,FileArray,FileArray-method}
\alias{>,FileArray,numeric-method}
\alias{>,numeric,FileArray-method}
\alias{>,FileArray,complex-method}
\alias{>,complex,FileArray-method}
\alias{>,FileArray,logical-method}
\alias{>,logical,FileArray-method}
\alias{>,FileArray,array-method}
\alias{>,array,FileArray-method}
\alias{<,FileArray,FileArray-method}
\alias{<,FileArray,numeric-method}
\alias{<,numeric,FileArray-method}
\alias{<,FileArray,complex-method}
\alias{<,complex,FileArray-method}
\alias{<,FileArray,logical-method}
\alias{<,logical,FileArray-method}
\alias{<,FileArray,array-method}
\alias{<,array,FileArray-method}
\alias{!=,FileArray,FileArray-method}
\alias{!=,FileArray,numeric-method}
\alias{!=,numeric,FileArray-method}
\alias{!=,FileArray,complex-method}
\alias{!=,complex,FileArray-method}
\alias{!=,FileArray,logical-method}
\alias{!=,logical,FileArray-method}
\alias{!=,FileArray,array-method}
\alias{!=,array,FileArray-method}
\alias{>=,FileArray,FileArray-method}
\alias{>=,FileArray,numeric-method}
\alias{>=,numeric,FileArray-method}
\alias{>=,FileArray,complex-method}
\alias{>=,complex,FileArray-method}
\alias{>=,FileArray,logical-method}
\alias{>=,logical,FileArray-method}
\alias{>=,FileArray,array-method}
\alias{>=,array,FileArray-method}
\alias{<=,FileArray,FileArray-method}
\alias{<=,FileArray,numeric-method}
\alias{<=,numeric,FileArray-method}
\alias{<=,FileArray,complex-method}
\alias{<=,complex,FileArray-method}
\alias{<=,FileArray,logical-method}
\alias{<=,logical,FileArray-method}
\alias{<=,FileArray,array-method}
\alias{<=,array,FileArray-method}
\alias{&,FileArray,FileArray-method}
\alias{&,FileArray,numeric-method}
\alias{&,numeric,FileArray-method}
\alias{&,FileArray,complex-method}
\alias{&,complex,FileArray-method}
\alias{&,FileArray,logical-method}
\alias{&,logical,FileArray-method}
\alias{&,FileArray,array-method}
\alias{&,array,FileArray-method}
\alias{|,FileArray,FileArray-method}
\alias{|,FileArray,numeric-method}
\alias{|,numeric,FileArray-method}
\alias{|,FileArray,complex-method}
\alias{|,complex,FileArray-method}
\alias{|,FileArray,logical-method}
\alias{|,logical,FileArray-method}
\alias{|,FileArray,array-method}
\alias{|,array,FileArray-method}
\alias{!,FileArray-method}
\alias{exp,FileArray-method}
\alias{expm1,FileArray-method}
\alias{log,FileArray-method}
\alias{log10,FileArray-method}
\alias{log2,FileArray-method}
\alias{log1p,FileArray-method}
\alias{abs,FileArray-method}
\alias{sqrt,FileArray-method}
\alias{sign,FileArray-method}
\alias{signif,FileArray-method}
\alias{trunc,FileArray-method}
\alias{floor,FileArray-method}
\alias{ceiling,FileArray-method}
\alias{round,FileArray-method}
\alias{acos,FileArray-method}
\alias{acosh,FileArray-method}
\alias{asin,FileArray-method}
\alias{asinh,FileArray-method}
\alias{atan,FileArray-method}
\alias{atanh,FileArray-method}
\alias{cos,FileArray-method}
\alias{cosh,FileArray-method}
\alias{cospi,FileArray-method}
\alias{sin,FileArray-method}
\alias{sinh,FileArray-method}
\alias{sinpi,FileArray-method}
\alias{tan,FileArray-method}
\alias{tanh,FileArray-method}
\alias{tanpi,FileArray-method}
\alias{gamma,FileArray-method}
\alias{lgamma,FileArray-method}
\alias{digamma,FileArray-method}
\alias{trigamma,FileArray-method}
\alias{Arg,FileArray-method}
\alias{Conj,FileArray-method}
\alias{Im,FileArray-method}
\alias{Mod,FileArray-method}
\alias{Re,FileArray-method}
\alias{is.na,FileArray-method}
\title{'S4' methods for \code{FileArray}}
\usage{
\S4method{+}{FileArray,FileArray}(e1, e2)

\S4method{+}{FileArray,numeric}(e1, e2)

\S4method{+}{numeric,FileArray}(e1, e2)

\S4method{+}{FileArray,complex}(e1, e2)

\S4method{+}{complex,FileArray}(e1, e2)

\S4method{+}{FileArray,logical}(e1, e2)

\S4method{+}{logical,FileArray}(e1, e2)

\S4method{+}{FileArray,array}(e1, e2)

\S4method{+}{array,FileArray}(e1, e2)

\S4method{-}{FileArray,FileArray}(e1, e2)

\S4method{-}{FileArray,numeric}(e1, e2)

\S4method{-}{numeric,FileArray}(e1, e2)

\S4method{-}{FileArray,complex}(e1, e2)

\S4method{-}{complex,FileArray}(e1, e2)

\S4method{-}{FileArray,logical}(e1, e2)

\S4method{-}{logical,FileArray}(e1, e2)

\S4method{-}{FileArray,array}(e1, e2)

\S4method{-}{array,FileArray}(e1, e2)

\S4method{*}{FileArray,FileArray}(e1, e2)

\S4method{*}{FileArray,numeric}(e1, e2)

\S4method{*}{numeric,FileArray}(e1, e2)

\S4method{*}{FileArray,complex}(e1, e2)

\S4method{*}{complex,FileArray}(e1, e2)

\S4method{*}{FileArray,logical}(e1, e2)

\S4method{*}{logical,FileArray}(e1, e2)

\S4method{*}{FileArray,array}(e1, e2)

\S4method{*}{array,FileArray}(e1, e2)

\S4method{/}{FileArray,FileArray}(e1, e2)

\S4method{/}{FileArray,numeric}(e1, e2)

\S4method{/}{numeric,FileArray}(e1, e2)

\S4method{/}{FileArray,complex}(e1, e2)

\S4method{/}{complex,FileArray}(e1, e2)

\S4method{/}{FileArray,logical}(e1, e2)

\S4method{/}{logical,FileArray}(e1, e2)

\S4method{/}{FileArray,array}(e1, e2)

\S4method{/}{array,FileArray}(e1, e2)

\S4method{^}{FileArray,FileArray}(e1, e2)

\S4method{^}{FileArray,numeric}(e1, e2)

\S4method{^}{numeric,FileArray}(e1, e2)

\S4method{^}{FileArray,complex}(e1, e2)

\S4method{^}{complex,FileArray}(e1, e2)

\S4method{^}{FileArray,logical}(e1, e2)

\S4method{^}{logical,FileArray}(e1, e2)

\S4method{^}{FileArray,array}(e1, e2)

\S4method{^}{array,FileArray}(e1, e2)

\S4method{\%\%}{FileArray,FileArray}(e1, e2)

\S4method{\%\%}{FileArray,numeric}(e1, e2)

\S4method{\%\%}{numeric,FileArray}(e1, e2)

\S4method{\%\%}{FileArray,complex}(e1, e2)

\S4method{\%\%}{complex,FileArray}(e1, e2)

\S4method{\%\%}{FileArray,logical}(e1, e2)

\S4method{\%\%}{logical,FileArray}(e1, e2)

\S4method{\%\%}{FileArray,array}(e1, e2)

\S4method{\%\%}{array,FileArray}(e1, e2)

\S4method{\%/\%}{FileArray,FileArray}(e1, e2)

\S4method{\%/\%}{FileArray,numeric}(e1, e2)

\S4method{\%/\%}{numeric,FileArray}(e1, e2)

\S4method{\%/\%}{FileArray,complex}(e1, e2)

\S4method{\%/\%}{complex,FileArray}(e1, e2)

\S4method{\%/\%}{FileArray,logical}(e1, e2)

\S4method{\%/\%}{logical,FileArray}(e1, e2)

\S4method{\%/\%}{FileArray,array}(e1, e2)

\S4method{\%/\%}{array,FileArray}(e1, e2)

\S4method{==}{FileArray,FileArray}(e1, e2)

\S4method{==}{FileArray,numeric}(e1, e2)

\S4method{==}{numeric,FileArray}(e1, e2)

\S4method{==}{FileArray,complex}(e1, e2)

\S4method{==}{complex,FileArray}(e1, e2)

\S4method{==}{FileArray,logical}(e1, e2)

\S4method{==}{logical,FileArray}(e1, e2)

\S4method{==}{FileArray,array}(e1, e2)

\S4method{==}{array,FileArray}(e1, e2)

\S4method{>}{FileArray,FileArray}(e1, e2)

\S4method{>}{FileArray,numeric}(e1, e2)

\S4method{>}{numeric,FileArray}(e1, e2)

\S4method{>}{FileArray,complex}(e1, e2)

\S4method{>}{complex,FileArray}(e1, e2)

\S4method{>}{FileArray,logical}(e1, e2)

\S4method{>}{logical,FileArray}(e1, e2)

\S4method{>}{FileArray,array}(e1, e2)

\S4method{>}{array,FileArray}(e1, e2)

\S4method{<}{FileArray,FileArray}(e1, e2)

\S4method{<}{FileArray,numeric}(e1, e2)

\S4method{<}{numeric,FileArray}(e1, e2)

\S4method{<}{FileArray,complex}(e1, e2)

\S4method{<}{complex,FileArray}(e1, e2)

\S4method{<}{FileArray,logical}(e1, e2)

\S4method{<}{logical,FileArray}(e1, e2)

\S4method{<}{FileArray,array}(e1, e2)

\S4method{<}{array,FileArray}(e1, e2)

\S4method{!=}{FileArray,FileArray}(e1, e2)

\S4method{!=}{FileArray,numeric}(e1, e2)

\S4method{!=}{numeric,FileArray}(e1, e2)

\S4method{!=}{FileArray,complex}(e1, e2)

\S4method{!=}{complex,FileArray}(e1, e2)

\S4method{!=}{FileArray,logical}(e1, e2)

\S4method{!=}{logical,FileArray}(e1, e2)

\S4method{!=}{FileArray,array}(e1, e2)

\S4method{!=}{array,FileArray}(e1, e2)

\S4method{>=}{FileArray,FileArray}(e1, e2)

\S4method{>=}{FileArray,numeric}(e1, e2)

\S4method{>=}{numeric,FileArray}(e1, e2)

\S4method{>=}{FileArray,complex}(e1, e2)

\S4method{>=}{complex,FileArray}(e1, e2)

\S4method{>=}{FileArray,logical}(e1, e2)

\S4method{>=}{logical,FileArray}(e1, e2)

\S4method{>=}{FileArray,array}(e1, e2)

\S4method{>=}{array,FileArray}(e1, e2)

\S4method{<=}{FileArray,FileArray}(e1, e2)

\S4method{<=}{FileArray,numeric}(e1, e2)

\S4method{<=}{numeric,FileArray}(e1, e2)

\S4method{<=}{FileArray,complex}(e1, e2)

\S4method{<=}{complex,FileArray}(e1, e2)

\S4method{<=}{FileArray,logical}(e1, e2)

\S4method{<=}{logical,FileArray}(e1, e2)

\S4method{<=}{FileArray,array}(e1, e2)

\S4method{<=}{array,FileArray}(e1, e2)

\S4method{&}{FileArray,FileArray}(e1, e2)

\S4method{&}{FileArray,numeric}(e1, e2)

\S4method{&}{numeric,FileArray}(e1, e2)

\S4method{&}{FileArray,complex}(e1, e2)

\S4method{&}{complex,FileArray}(e1, e2)

\S4method{&}{FileArray,logical}(e1, e2)

\S4method{&}{logical,FileArray}(e1, e2)

\S4method{&}{FileArray,array}(e1, e2)

\S4method{&}{array,FileArray}(e1, e2)

\S4method{|}{FileArray,FileArray}(e1, e2)

\S4method{|}{FileArray,numeric}(e1, e2)

\S4method{|}{numeric,FileArray}(e1, e2)

\S4method{|}{FileArray,complex}(e1, e2)

\S4method{|}{complex,FileArray}(e1, e2)

\S4method{|}{FileArray,logical}(e1, e2)

\S4method{|}{logical,FileArray}(e1, e2)

\S4method{|}{FileArray,array}(e1, e2)

\S4method{|}{array,FileArray}(e1, e2)

\S4method{!}{FileArray}(x)

\S4method{exp}{FileArray}(x)

\S4method{expm1}{FileArray}(x)

\S4method{log}{FileArray}(x, base = exp(1))

\S4method{log10}{FileArray}(x)

\S4method{log2}{FileArray}(x)

\S4method{log1p}{FileArray}(x)

\S4method{abs}{FileArray}(x)

\S4method{sqrt}{FileArray}(x)

\S4method{sign}{FileArray}(x)

\S4method{signif}{FileArray}(x, digits = 6)

\S4method{trunc}{FileArray}(x, ...)

\S4method{floor}{FileArray}(x)

\S4method{ceiling}{FileArray}(x)

\S4method{round}{FileArray}(x, digits = 0)

\S4method{acos}{FileArray}(x)

\S4method{acosh}{FileArray}(x)

\S4method{asin}{FileArray}(x)

\S4method{asinh}{FileArray}(x)

\S4method{atan}{FileArray}(x)

\S4method{atanh}{FileArray}(x)

\S4method{cos}{FileArray}(x)

\S4method{cosh}{FileArray}(x)

\S4method{cospi}{FileArray}(x)

\S4method{sin}{FileArray}(x)

\S4method{sinh}{FileArray}(x)

\S4method{sinpi}{FileArray}(x)

\S4method{tan}{FileArray}(x)

\S4method{tanh}{FileArray}(x)

\S4method{tanpi}{FileArray}(x)

\S4method{gamma}{FileArray}(x)

\S4method{lgamma}{FileArray}(x)

\S4method{digamma}{FileArray}(x)

\S4method{trigamma}{FileArray}(x)

\S4method{Arg}{FileArray}(z)

\S4method{Conj}{FileArray}(z)

\S4method{Im}{FileArray}(z)

\S4method{Mod}{FileArray}(z)

\S4method{Re}{FileArray}(z)

\S4method{is.na}{FileArray}(x)
}
\arguments{
\item{x, z, e1, e2}{\code{FileArray} or compatible data}

\item{base, digits, ...}{passed to other methods}
}
\value{
See \code{\link[methods]{S4groupGeneric}}
}
\description{
'S4' methods for \code{FileArray}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapreduce.R
\name{mapreduce}
\alias{mapreduce}
\alias{mapreduce,FileArray,ANY,function-method}
\alias{mapreduce,FileArray,ANY,NULL-method}
\alias{mapreduce,FileArray,ANY,missing-method}
\title{A map-reduce method to iterate blocks of file-array data with little memory usage}
\usage{
mapreduce(x, map, reduce, ...)

\S4method{mapreduce}{FileArray,ANY,function}(x, map, reduce, buffer_size = NA, ...)

\S4method{mapreduce}{FileArray,ANY,NULL}(x, map, reduce, buffer_size = NA, ...)

\S4method{mapreduce}{FileArray,ANY,missing}(x, map, reduce, buffer_size = NA, ...)
}
\arguments{
\item{x}{a file array object}

\item{map}{mapping function that receives 3 arguments; see 'Details'}

\item{reduce}{\code{NULL}, or a function that takes a list as input}

\item{...}{passed to other methods}

\item{buffer_size}{control how we split the array; see 'Details'}
}
\value{
If \code{reduce} is \code{NULL}, return mapped results, otherwise
return reduced results from \code{reduce} function
}
\description{
A map-reduce method to iterate blocks of file-array data with little memory usage
}
\details{
When handling out-of-memory arrays, it is recommended to load
a block of array at a time and execute on block level. See 
\code{\link{apply}} for a implementation. When an array is too large,
and when there are too many blocks, this operation will become 
very slow if computer memory is low. 
This is because the R will perform garbage collection frequently. 
Implemented in \code{C++}, \code{mapreduce} creates a buffer to store
the block data. By reusing the memory over and over again, it is possible
to iterate through the array with minimal garbage collections. Many 
statistics, including \code{min}, \code{max}, \code{sum}, 
\code{mean}, ... These statistics can be calculated in this 
way efficiently.

The function \code{map} contains three arguments: \code{data} (mandate), 
\code{size} (optional), and \code{first_index} (optional). 
The \code{data} is the buffer,
whose length is consistent across iterations. \code{size} indicates
the effective size of the buffer. If the partition size
is not divisible by the buffer size, only first \code{size} elements of
the data are from array, and the rest elements will be \code{NA}. 
This situation could only occurs when \code{buffer_size} is manually 
specified. By default, all of \code{data} should belong to arrays.
The last argument \code{first_index} is the index of the first element
\code{data[1]} in the whole array. It is useful when positional data 
is needed. 

The buffer size, specified by \code{buffer_size} is an 
additional optional argument in \code{...}. Its default is \code{NA},
and will be calculated automatically. If manually specified, a
large buffer size would be desired to speed up the calculation.
The default buffer size will not exceed \eqn{nThreads x 2MB}, where 
\code{nThreads} is the number of threads set by \code{\link{filearray_threads}}.
When partition length cannot be divided by the buffer size, instead of
trimming the buffer, \code{NA}s will be filled to the buffer, 
passed to \code{map} function; see previous paragraph for treatments.

The function \code{mapreduce} ignores the missing partitions. That means
if a partition is missing, its data will not be read nor passed to 
\code{map} function. Please run \code{x$initialize_partition()} to make sure
partition files exist.
}
\examples{


x <- filearray_create(tempfile(), c(100, 100, 10))
x[] <- rnorm(1e5)

## calculate summation
# identical to sum(x[]), but is more feasible in large cases

mapreduce(x, map = function(data, size) {
    # make sure `data` is all from array
    if (length(data) != size) {
        data <- data[1:size]
    }
    sum(data)
}, reduce = function(mapped_list) {
    do.call(sum, mapped_list)
})


## Find elements are less than -3
positions <- mapreduce(
    x,
    map = function(data, size, first_index) {
        if (length(data) != size) {
            data <- data[1:size]
        }
        which(data < -3) + (first_index - 1)
    },
    reduce = function(mapped_list) {
        do.call(c, mapped_list)
    }
)

if(length(positions)) {
    x[[positions[1]]]
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa-docs.R, R/methods-subset.R,
%   R/methods-subsetAssign.R, R/methods.R
\name{S3-filearray}
\alias{S3-filearray}
\alias{[.FileArray}
\alias{[<-.FileArray}
\alias{[[.FileArray}
\alias{as.array.FileArray}
\alias{dim.FileArray}
\alias{dimnames.FileArray}
\alias{dimnames<-.FileArray}
\alias{length.FileArray}
\alias{max.FileArray}
\alias{min.FileArray}
\alias{range.FileArray}
\alias{sum.FileArray}
\alias{subset.FileArray}
\title{'S3' methods for 'FileArray'}
\usage{
\method{[}{FileArray}(
  x,
  i,
  ...,
  drop = TRUE,
  reshape = NULL,
  strict = TRUE,
  dimnames = TRUE,
  split_dim = 0
)

\method{[}{FileArray}(x, i, ..., lazy = FALSE) <- value

\method{[[}{FileArray}(x, i)

\method{as.array}{FileArray}(x, reshape = NULL, drop = FALSE, ...)

\method{dim}{FileArray}(x)

\method{dimnames}{FileArray}(x)

\method{dimnames}{FileArray}(x) <- value

\method{length}{FileArray}(x)

\method{max}{FileArray}(x, na.rm = FALSE, ...)

\method{min}{FileArray}(x, na.rm = FALSE, ...)

\method{range}{FileArray}(x, na.rm = FALSE, ...)

\method{sum}{FileArray}(x, na.rm = FALSE, ...)

\method{subset}{FileArray}(x, ..., drop = FALSE, .env = parent.frame())
}
\arguments{
\item{x}{a file array}

\item{i, ...}{index set, or passed to other methods}

\item{drop}{whether to drop dimensions; see topic \code{\link[base]{Extract}}}

\item{reshape}{a new dimension to set before returning subset results; default is \code{NULL} (use default dimensions)}

\item{strict}{whether to allow indices to exceed bound; currently only accept \code{TRUE}}

\item{dimnames}{whether to preserve \code{\link[base]{dimnames}}}

\item{split_dim}{internally used; split dimension and calculate indices to
manually speed up the subset; value ranged from 0 to size of dimension minus
one.}

\item{lazy}{whether to lazy-evaluate the method, only works when assigning
arrays with logical array index}

\item{value}{value to substitute or set}

\item{na.rm}{whether to remove \code{NA} values during the calculation}

\item{.env}{environment to evaluate formula when evaluating subset margin indices.}
}
\description{
These are 'S3' methods for 'FileArray'
}
\section{Functions}{
\itemize{
\item \code{[}: subset array

\item \code{`[`(FileArray) <- value}: subset assign array

\item \code{[[}: get element by index

\item \code{as.array(FileArray)}: converts file array to native array in R

\item \code{dim(FileArray)}: get dimensions

\item \code{dimnames(FileArray)}: get dimension names

\item \code{dimnames(FileArray) <- value}: set dimension names

\item \code{length(FileArray)}: get array length

\item \code{max(FileArray)}: get max value

\item \code{min(FileArray)}: get min value

\item \code{range(FileArray)}: get value range

\item \code{sum(FileArray)}: get summation

\item \code{subset(FileArray)}: get subset file array with formulae

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{fwhich}
\alias{fwhich}
\alias{fwhich.default}
\alias{fwhich.FileArray}
\title{A generic function of \code{which} that is \code{'FileArray'} compatible}
\usage{
fwhich(x, val, arr.ind = FALSE, ret.values = FALSE, ...)

\method{fwhich}{default}(x, val, arr.ind = FALSE, ret.values = FALSE, ...)

\method{fwhich}{FileArray}(x, val, arr.ind = FALSE, ret.values = FALSE, ...)
}
\arguments{
\item{x}{any R vector, matrix, array or file-array}

\item{val}{values to find, or a function taking one argument (a slice of 
data vector) and returns either logical vector with the same length as the 
slice or index of the slice; see 'Examples'}

\item{arr.ind}{logical; should array indices be 
returned when \code{x} is an array?}

\item{ret.values}{whether to return the values of corresponding indices as 
an attributes; default is false}

\item{...}{passed to \code{val} if \code{val} is a function}
}
\value{
The indices of \code{x} elements that are listed in \code{val}.
}
\description{
A generic function of \code{which} that is \code{'FileArray'} compatible
}
\examples{


# ---- Default case ------------------------------------
x <- array(1:27 + 2, rep(3,3))

# find index of `x` equal to either 4 or 5
fwhich(x, c(4,5))
res <- fwhich(x, c(4,5), ret.values = TRUE)
res
attr(res, "values")

# ---- file-array case --------------------------------
arr <- filearray_create(tempfile(), dim(x))
arr[] <- x
fwhich(arr, c(4,5))
fwhich(arr, c(4,5), arr.ind = TRUE, ret.values = TRUE)

arr[2:3, 1, 1]

# Clean up this example
arr$delete()

# ---- `val` is a function ----------------------------
x <- as_filearray(c(sample(15), 15), dimension = c(4,4))

ret <- fwhich(x, val = which.max, 
              ret.values = TRUE, arr.ind = FALSE)

# ret is the index
ret == which.max(x[])

# attr(ret, "values") is the max value
max(x[]) == attr(ret, "values")

# customize `val`
fwhich(x, ret.values = TRUE, arr.ind = FALSE,
       val = function( slice ) {
           slice > 10 # or which(slice > 10)
       })


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods-apply.R
\name{apply}
\alias{apply}
\alias{apply,FileArray-method}
\alias{apply,FileArrayProxy-method}
\title{Apply functions over file array margins (extended)}
\usage{
apply(X, MARGIN, FUN, ..., simplify = TRUE)

\S4method{apply}{FileArray}(X, MARGIN, FUN, ..., simplify = TRUE)

\S4method{apply}{FileArrayProxy}(X, MARGIN, FUN, ..., simplify = TRUE)
}
\arguments{
\item{X}{a file array}

\item{MARGIN}{scalar giving the subscripts which the function will be applied over. Current implementation only allows margin size to be one}

\item{FUN}{the function to be applied}

\item{...}{optional arguments to \code{FUN}}

\item{simplify}{a logical indicating whether results should be simplified if possible}
}
\value{
See Section 'Value' in \code{\link[base]{apply}};
}
\description{
Apply functions over file array margins (extended)
}
